#ifndef KHEPRI_HH
#define KHEPRI_HH

#include <gazebo.h>
#include <gazebo/Body.hh>
#include <gazebo/Model.hh>

#include <list>

// Forward declarations
class Body;
class RayProximity;

class LineSegment {
public:
	double x1;
	double y1;
	double x2;
	double y2;
	double dist;
	LineSegment(double _x1, double _y1, double _x2, double _y2) {x1 = _x1; y1 = _y1; x2 = _x2; y2 = _y2; ComputeDist();}
	void ComputeDist() { dist = sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1)); }
};

class Floorplan : public Model
{
public:
  // Constructor
  Floorplan( World *world );
  
  // Destructor
  virtual ~Floorplan();

  int Load( WorldFile *file, WorldFileNode *node );  
  int Init( WorldFile *file, WorldFileNode *node );
  int Fini();
  
  // Do updates
  public: virtual void Update( double step );

private:
  int OdeLoad( WorldFile *file, WorldFileNode *node );
	int ReadVertInput();
	int ReadPolyInput();
  
  void UpdateOdometry( double step );
  void GetVelocityCmd();
  
  // Update the data in the external interface
  void PutPositionData();

  Body *body;
  const char *texture_file;
  GzColor wall_color;
	double wall_thickness, wall_height;

	std::list<LineSegment> line_segments;

	const char* vert_input_filename;
	const char* poly_input_filename;

  // Update parameters
  double updatePeriod, updateTime;

  // External interface
  gz_position_t *position_iface;

  double odomPose[6], odomVel[6];
};

#endif
