#include	"config.h"
#include	<stdio.h>
#include	<signal.h>
#include	<string.h>
#include	<malloc.h>
#include        <stdlib.h>
#include	<errno.h>
#include        <memory.h>

#ifndef SEGV_PROT
/* for System V systems like Solaris */
#define SEGV_PROT SEGV_ACCERR
#endif

#ifdef USE_LWPS
/* Some SOLARIS versions use LWPs */
#include <sys/lwp.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/fcntl.h>
#include <sys/time.h>
#include <vm/faultcode.h>

#include        <sys/mman.h>
#include        <fcntl.h>
#include        <unistd.h>
#include        <sys/wait.h>
#include 	<sys/ipc.h>
#include 	<sys/shm.h>

#include        <sys/resource.h>


#include	"results.h"
#include	"general.h"
#include	"init.h"
#include	"lock.h"
#include	"queue.h"
#include	"internal.h"
#include	"memalloc.h"
#include	"cth_sched.h"
#include	"hwlib.h"
#include	"sched_utils.h"
#include	"idebug.h"
#include	"threads.h"
#include	"sparc_init.h"
#include        "sched_io.h"
#include "monitor.h"

extern shared short *monitor_status;
extern shared mutex_t logicalTimeStamp_lock;
extern shared unsigned long *logicalTimeStamp_counter;
extern shared LOCK *print_sema;

/* these symbols needed from init.c */
extern struct local_info *processor;
extern short threads_started;
extern int *num_of_procs;
extern atom_int *total_threads;
extern atom_int *total_active;
typedef long *process_table_t;
extern process_table_t processes;	/* An array of the processes    */
extern shared long dynamic_memory_per_processor;
extern struct configuration config;	/* configuration        */
#define		NO_SIZE		-1
#define		NO_NODE		-1
extern shared mem_free_list_t *memory_pool;

LOCK *ss_lock;
extern cthread_t current_thread_array[MAX_PROC];
extern shared struct sched *local_scheduler[MAX_PROC];
atom_int *sync_init;

/* externals not defined by sys/shm.h */
#if !defined(__svr4__) && !defined(__SVR4)
extern int shmdt ARGS((void *shmaddr));
extern void *shmat ARGS((int shmid, void *shmaddr, int shmflg));
extern int shmctl ARGS((int shmid, int cmd,...));
extern int shmget ARGS((key_t key, int size, int shmflg));
extern int wait3 ARGS((int *statusp, int options, struct rusage * rusage));
extern int sigblock ARGS((int mask));
extern int sigsetmask ARGS((int mask));
extern int sigpause ARGS((int sigmask));
extern int on_exit ARGS((void (*procp) (), caddr_t arg));
#else
extern int atexit(void (*func) (void));
#endif

/* externals not defined by sys/ipc.h */
extern key_t ftok ARGS((const char *path, int id));

/* other random unix externals */
extern void perror ARGS((const char *s));
extern void *memset();
extern int fflush ARGS((FILE *));
extern int select();

/* defined is $OS_signal.c */
extern void setup_child_sighandlers();
extern void setup_scheduler_sighandlers();
extern void init_sighandler_struct();
extern void setup_master_sighandlers();
extern void setup_term_sighandler();
extern void setup_chld_sighandler();

extern int getpagesize();
extern int page_aligned_size ARGS((int));
extern int generic_shared_size ARGS((int));

#ifdef WTERMSIG
#define WAIT_STAT_TYPE int
#define TERM_SIG(stat) WTERMSIG(stat)
#define RET_CODE(stat) WEXITSTATUS(stat)
#else
#define WAIT_STAT_TYPE union wait
#define TERM_SIG(stat) stat.w_termsig
#define RET_CODE(stat) stat.w_retcode
#endif

extern RESULT init_convert ARGS((int procs));
extern int ill_hndlr();
extern int seg_hndlr();

extern void monitorinit ARGS((int procs));
static void sparc_monitorinit ARGS((int procs));
void relinquish_processor();
void child_exit ARGS((int));
void clean_shut_down ARGS(());


#ifndef STATIC_MEMORY_SHARED
/* multi procs must use SYSV shared memory */
static int use_shared_memory = 1;

#else
static int use_shared_memory = 0;
/* should not use SYSV shared memory */
#endif

shared int real_num_of_procs;

char *shm_base, *monmem_base;
long shm_offset, monmem_offset, shm_limit, monmem_limit;
key_t shm_key, monmem_key;
int shm_fn, shm_id, monmem_fn, monmem_id;

static volatile int total_procs_exit = 0;
static int total_procs_forked = 0;
#ifndef USE_LWPS
static int parent_waiting_the_end = 0;
#endif
static int parent_in_exit = 0;

#ifdef HAVE_SIGINFO_H
int general_hndlr ARGS((int signo, siginfo_t siginfo, void *context));
#else
int general_hndlr ARGS((int signo, int code, void *scp, char *addr));
#endif

#ifndef USE_LWPS
private int private_current_proc = 0;
static void abnormal ARGS((WAIT_STAT_TYPE stat, int wid));
void child_exit_handler ARGS((int wid, WAIT_STAT_TYPE stat));

#ifdef HAVE_SIGINFO_H
int chld_hndlr ARGS((int signo, siginfo_t siginfo, void *context));
#else
int chld_hndlr ARGS((int signo, int code, void *scp, char *addr));
#endif
#endif


#if !defined(__svr4__) && !defined(__SVR4)

void
master_on_exit(int status)
{
    DBG3("In master on exit, VP=%d, total= %d, exit=%d\n",
	 virtual_processor(), total_procs_forked, total_procs_exit);
    parent_in_exit = 1;
    if (virtual_processor() == 0 &&
	total_procs_forked > 0 &&
	total_procs_exit > 0) {
	child_exit(status);
    }
}
#else
/* SOLARIS*/

void
master_on_exit()
{
    parent_in_exit = 1;
    DBG3("In 2master on exit, VP=%d, total= %d, exit=%d\n",
	 virtual_processor(), total_procs_forked, total_procs_exit);
    if (virtual_processor() == 0 &&
	total_procs_forked > 0 &&
	total_procs_exit != 0) {
	child_exit(0);
    }
}

#endif

int getpagesize()		/* not in libc before Solaris 2.5 */
{
    return sysconf(_SC_PAGESIZE);
}

int
current_processor()
{
    return virtual_processor();
}

void
child_init(node, start_func, arg)
int node;
int (*start_func) ();
any_t arg;
{
    setup_child_sighandlers(ill_hndlr, seg_hndlr, seg_hndlr);
    setup_term_sighandler(general_hndlr);

    if (use_shared_memory) {
	shmdt(shm_base);
	shm_base = (char *) shmat(shm_id, 0, 0);
	if (shm_base == (char *) -1) {
	    perror("Child Cannot attach shared memory");
	    child_exit(1);
	}
#ifdef WITH_MONITORING
	if (monitoring_enabled) {
	    monmem_base = (char *) shmat(monmem_id, 0, 0);
	    if (monmem_base == (char *) -1) {
		perror("Cannot create shared memory");
		child_exit(1);
	    }
	}
#endif
    }
    current_thread_array[node] = NULL;
    local_scheduler[node] = processor[node].scheduler;

    DBG3("In child %d, processor = %x, processes = %x\n",
	 node, (int) processor, (int) processes);
    DBG2("In child %d, local_scheduler = %lx\n",
	 node, (long) local_scheduler[node]);
    DBG4("In child %d no_of_procs = %x *no_of_procs = %d mem_pool = %x\n",
	 node, (int) num_of_procs, *num_of_procs,
	 (int) memory_pool);

    setup_scheduler_sighandlers();

    if (local_scheduler[node]->sched_vproc_init != 0) {
	(local_scheduler[node]->sched_vproc_init) (node);
    }
    threads_started = 1;


/*******************************************************************/
/* sync_init has to be equal to current_processor + 1 to make sure */
/* that all the child proceses wait till the child 0 is done       */
/* forking the init thread. This is necessary so that the monitor  */
/* thread doesn't get the chance to lock the list->lock before the */
/* init thread executes add_thread_to_lm_list                      */
/*******************************************************************/
    while (*sync_init != node + 1) {
	struct timeval delay_time;
	int delay_unit = (*num_of_procs - node);
	delay_time.tv_sec = 0;
	delay_time.tv_usec = delay_unit * 100000;	/* return *
							 * millisecs */
	DBG3("Sync init = %d  current proc = %d delaying for %d\n", *sync_init,
	     node, delay_unit);
	select(32, NULL, NULL, NULL, &delay_time);
    }

    start_proc_initial_thread(start_func);

    atomadd(sync_init, -1);
    while (*sync_init != 0) {
	struct timeval delay_time;
	delay_time.tv_sec = 0;
	delay_time.tv_usec = (*sync_init) * 100000;	/* return *
							 * millisecs */
	DBG2("Sync post init = %d  current proc = %d\n",
	     *sync_init, node);
	select(32, NULL, NULL, NULL, &delay_time);
    }

    DBG1("Children initialized. Starting our queue %d\n", node);
    if ((start_func == NULL) && (node == 0)) {
	current_thread_array[node] = *initial_thread;
	(*initial_thread)->already_started = 1;
	(*initial_thread)->usable_stack_base = (void*)0xffffffff;
	(*initial_thread)->usable_stack_limit = 0;
	swap_context_with_func((void (*)()) local_scheduler[node]->schedule,
			       NULL);
    } else {
	local_scheduler[node]->schedule();
    }

}				/* child init */


void
create_shared_arena(procs)
int procs;
{
    int size = generic_shared_size(procs);
    if ((procs > 1) && use_shared_memory) {
	if ((shm_fn = open(".shm_mem", O_RDWR | O_CREAT | O_TRUNC, 0777)) == -1)
	    perror(".shm_mem");
	shm_key = ftok(".shm_mem", '1');

	shm_id = shmget(shm_key, size, IPC_CREAT | 0600);
	if (shm_id == -1) {
	    perror("SHared Memory GET failed!");
	    fprintf(stderr, "Possible cause:  Requested thread/processor/memory (size: %d)\
configuration requires\n\tmore shared memory than is \
available.  Reduce requirements.\n", size);
	    shmctl(shm_id, IPC_RMID, 0);
	    close(shm_fn);
	    unlink(".shm_mem");
	    exit(1);
	}
	shm_base = (char *) shmat(shm_id, 0, 0);
	if (shm_base == (char *) -1) {
	    perror("Cannot create shared memory");
	    shmctl(shm_id, IPC_RMID, 0);
	    close(shm_fn);
	    unlink(".shm_mem");
	    exit(1);
	}
	shm_offset = 0;
	shm_limit = size;
    } else {
	use_shared_memory = 0;
	shm_base = (char *) malloc(size);
	shm_offset = 0;
	shm_limit = size;
    }
}

#ifdef USE_LWPS
/* THIS IS CODE FOR SOLARIS LWP VERSION!! */

/* 
 *  the thr_* and key declarations are copied from thread.h because typedefs 
 *  in that file conflict with some of Cthreads basic types (like mutex_t).
 *  It may be possible to resolve this at some time since we don't truely 
 *  need the conflicting Cthreads declarations in this file.  But until that 
 *  happens, this at least keeps the compiler warnings away.
 */
typedef unsigned int thread_t;
typedef unsigned int thread_key_t;
static thread_key_t key;

extern int thr_keycreate ARGS((thread_key_t *, void (*)(void *)));
extern int thr_setspecific ARGS((thread_key_t, void *));
extern int thr_getspecific ARGS((thread_key_t, void **));
extern int thr_create ARGS((void *, size_t, void *(*)(void *), void *, long, void *));
extern void thr_exit ARGS((void *));
extern thread_t thr_self();
extern int thr_kill ARGS((thread_t, int));
extern int thr_join ARGS((thread_t, thread_t *, void **));

static thread_t thrs[MAX_PROC];

int
virtual_processor()
{
    int *node_p;
    if (thr_getspecific(key, (void **) &node_p) != 0)
	perror("getspecific");
    if (node_p == NULL) {
	return 0;
    }
    if (total_procs_forked == 1) {
	return 0;
    }
    return (*node_p);
}

typedef struct child_params {
    int (*start_func) ();
    int node;
} *param_t;

void
lwp_init(pblock)
param_t pblock;
{
    thr_setspecific(key, (void *) &pblock->node);
    DBG2("starting node %d, addr pblock %lx\n", pblock->node, (long) pblock);
    child_init(pblock->node, pblock->start_func, NULL);
}

void
fork_children_and_wait(start_func, procs)
int (*start_func) ();
int procs;
{
    thread_t thrid = thr_self();
    int node;
    int *priv = malloc(sizeof(int));

    total_procs_forked = procs;
    total_procs_exit = procs;	/* for the overwritten "exit" */

    if (thr_keycreate(&key, NULL) != 0)
	perror("key create");
    for (node = 1; node < procs; node++) {
	int error;
	param_t param_block;
	param_block = (void *) allocate_and_share(sizeof(struct
						      child_params), -1);
	param_block->node = node;
	param_block->start_func = start_func;
	if ((error = thr_create(NULL, 0, (void *(*)(void *)) lwp_init, param_block,
				1, &thrs[node])) != 0) {
	    perror("thr_create failed");
	    if (error == -1) {
		fprintf(stderr, "Did you link with -lthread?\n");
	    } else {
		fprintf(stderr, "Can't support so many processes.  error = %d, Reduce -num_procs.\n", error);
	    }
	    goto kill_children;
	}
	DBG2("virtual processor %d forked at TID= %lx\n", node,
	     (long)thrs[node]);
    }
    *priv = 0;
    if (thr_setspecific(key, priv) != 0)
	perror("thr_set_specific");
    thrs[0] = thr_self();
    DBG3("before child_init, VP=%d, total= %d, exit=%d\n",
	 virtual_processor(), total_procs_forked, total_procs_exit);
    child_init(0, start_func, NULL);

    return;
  kill_children:		/* GOTO TARGET FROM JUST ABOVE ! */
    DBG("Begin killing the children\n");
    for (node = 0; node < procs; node++) {
	if (thrs[node] != thrid) {
	    DBG2("Killing child Thread oflogical processor %d, pid %d\n", node, thrs[node]);
	    thr_kill(thrs[node], SIGTERM);
	}
    }
    for (node = 1; node < procs; node++) {
	DBG1("waiting on child %d\n", node);
	_lwp_wait(0, &thrs[node]);
    }
    DBG("Done with wait\n");

    exit(0);
}


int
ill_hndlr(signo, code, scp, addr)
int signo, code;
void *scp;
char *addr;
{
    fflush(stdout);
    fflush(stderr);
    fprintf(stderr, "Illegal instruction signal.  Possible stack overflow on processor %d.\n", virtual_processor());
    fprintf(stderr, "Thread ");
    fprint_thread(stderr, current_thread_array[virtual_processor()]);
    fprintf(stderr, " was active.\nTerminating program.\n");

    exit(1);
}

#ifdef HAVE_SIGINFO_H
int
seg_hndlr(signo, siginfo, context)
int signo;
siginfo_t siginfo;
void *context;
#else
int
seg_hndlr(signo, code, scp, addr)
int signo, code;
void *scp;
char *addr;
#endif
{
    cthread_t current_thread = current_thread_array[virtual_processor()];

    fflush(stdout);
    fflush(stderr);
    if (signo == SIGSEGV) {
	fprintf(stderr, "Segmentation fault signal (11) on processor %d.\n",
		virtual_processor());
    } else if (signo == SIGBUS) {
	fprintf(stderr, "Bus error signal (10) on processor %d.\n",
		virtual_processor());
    } else {
	fprintf(stderr, "Signal %d on processor %d.\n", signo,
		virtual_processor());
    }
#ifdef HAVE_SIGINFO_H
    printf("siginfo.si_code is %lx, cur thread %lx\n", (long)siginfo.si_code,
	   (long)current_thread);
    if (siginfo.si_code == SEGV_ACCERR) {
	fprintf(stderr, "Signal code suggests protection violation and a possible stack overflow.\n");
    }
    fprintf(stderr, "signo %d, code %d, stacktop %lx\n",
	    signo, siginfo.si_code, (long) current_thread->stack_top);

#else				/* indef HAVE_SIGINFO_H */

    if (FC_CODE(code) == FC_OBJERR)
	code = FC_ERRNO(code);

    if (code == SEGV_PROT) {
	fprintf(stderr, "Signal code suggests protection violation and a possible stack overflow.\n");
    }
    fprintf(stderr, "signo %d, code %d, addr %lx, stacktop %lx\n",
	    signo, code, (long) addr, (long) current_thread->stack_top);
#endif				/* HAVE_SIGINFO_H */
    fprintf(stderr, "Thread ");
    fprint_thread(stderr, current_thread);
    fprintf(stderr, " was active.\nTerminating program.\n");


    child_exit(1);
    return 0;
}

#ifdef HAVE_SIGINFO_H
int
general_hndlr(signo, siginfo, context)
int signo;
siginfo_t siginfo;
void *context;
#else
int
general_hndlr(signo, code, scp, addr)
int signo, code;
void *scp;
char *addr;
#endif
{
    DBG1("General handler code %d\n", signo);
    child_exit(1);
    return 0;
}


void
child_exit(status)
int status;
{
    int i;
    DBG1("in LWP child_exit %d\n", virtual_processor());
    for (i = 0; i < total_procs_forked; i++) {
        processor[i].terminate = TERMINATE;
    }
    if (!parent_in_exit) {
        DBG("Nobody called exit yet.  I will.\n");
        exit(status);
    }
    if (virtual_processor() == 0) {
	/* 
	 * the loop is mainly meant to shut down the steering and 
	 * local monitor processes 
	 */
	DBG("Master waiting\n");

	for (i=1; i< total_procs_forked; i++) {
	    void *status;
	    DBG1("Master joining %d\n", i);
	    thr_join(thrs[i], NULL, &status);
	}
	DBG("Master exitting\n");
    } else {
	DBG("in LWP child_exit exitting\n");
	total_procs_exit--;
	thr_exit((void*)status);
    }
}

#else				/* ifndef USE_LWPS */
/* SUNOS and Solaris non-LWP code */


void
fork_children_and_wait(start_func, procs)
int (*start_func) ();
int procs;
{
    int node, i, chid;
    WAIT_STAT_TYPE stat;

    setup_chld_sighandler(chld_hndlr);

    total_procs_forked = total_procs_exit = procs;

    stat = SIGINT;
    processes[0] = getpid();
    processes[procs] = PROC_SENTINEL;
    private_current_proc = 0;
    if (procs > 1) {
	for (node = 1; node < procs; node++) {
	    if ((chid = fork()) == -1) {
		perror("fork failed");
		fprintf(stderr, "Can't support so many processes.  Reduce -num_procs.\n");
		total_procs_forked = total_procs_exit = node;
		goto kill_children;
	    }
	    if (chid == 0) {
		private_current_proc = node;
		child_init(node, start_func, NULL);
		printf("(AQ) Internal Error:Returned from child init !!!!!!\n");
		exit(1);
	    } else {
		processes[node] = chid;
	    }
	}
	DBG("Process table:\n");
	for (i = 0, processes; i < procs; i++) {
	    DBG2("%d         %ld\n", i, processes[i]);
	}
    }
    child_init(0, start_func, NULL);
    return;
  kill_children:		/* GOTO TARGET FROM JUST ABOVE ! */
    child_exit_handler(processes[0], stat);
    return;
}


static void
abnormal(stat, wid)
WAIT_STAT_TYPE stat;
int wid;
{
    process_table_t pptr;
    long int proc;

    fflush(stdout);
    fflush(stderr);

    for (proc = 0, pptr = processes; *pptr != PROC_SENTINEL && *pptr != wid;
	 pptr++, proc++);

    if (*pptr == PROC_SENTINEL) {
	printf("Received a termination of an unknown child (%d) termsig %d retcode %d\n",
	       wid, TERM_SIG(stat), RET_CODE(stat));
    } else {
	if ((RET_CODE(stat) != 0) || (TERM_SIG(stat) != 0)) {
	    switch (TERM_SIG(stat)) {
	    case SIGINT:
		printf("Child %d died with INTERRUPT SIGNAL %d on logical proc %ld\n",
		       wid, SIGINT, proc);
		break;
	    case SIGQUIT:
		printf("Child %d died with QUIT SIGNAL %d on logical proc %ld\n",
		       wid, SIGQUIT, proc);
		break;
	    case SIGBUS:
		printf("Child %d died with BUS ERROR signal %d on logical proc %ld\n",
		       wid, SIGBUS, proc);
		break;
	    case SIGSEGV:
		printf("Child %d died with segmentation violation signal %d on logical proc %ld\n",
		       wid, SIGSEGV, proc);
		break;
	    case SIGTERM:
		printf("Child %d died from kill %d on logical proc %ld\n",
		       wid, SIGTERM, proc);
		break;
	    case SIGILL:
		printf("Child %d died with illegal instruction signal %d on logical proc %ld\n",
		       wid, SIGILL, proc);
		break;
	    default:
		printf("Child %d died with termination signal %d on logical proc %ld\n",
		       wid, TERM_SIG(stat), proc);
	    }
	}
    }
}

#ifdef HAVE_SIGINFO_H
int
ill_hndlr(signo, siginfo, context)
int signo;
siginfo_t siginfo;
void *context;
#else
int
ill_hndlr(signo, code, scp, addr)
int signo, code;
void *scp;
char *addr;
#endif
{
    fflush(stdout);
    fflush(stderr);
    fprintf(stderr, "Illegal instruction signal.  Possible stack overflow on processor %d.\n", virtual_processor());
    fprintf(stderr, "Thread ");
    fprint_thread(stderr, current_thread_array[virtual_processor()]);
    fprintf(stderr, " was active.\nTerminating program.\n");


#ifdef HAVE_SIGINFO_H
    general_hndlr(signo, siginfo, context);
#else
    general_hndlr(signo, code, scp, addr);
#endif
    return 0;
}

#ifdef HAVE_SIGINFO_H
int
seg_hndlr(signo, siginfo, context)
int signo;
siginfo_t siginfo;
void *context;
#else
int
seg_hndlr(signo, code, scp, addr)
int signo, code;
void *scp;
char *addr;
#endif
{
    cthread_t current_thread = current_thread_array[virtual_processor()];

    fflush(stdout);
    fflush(stderr);
    if (signo == SIGSEGV) {
	fprintf(stderr, "Segmentation fault signal (11) on processor %d.\n",
		virtual_processor());
    } else if (signo == SIGBUS) {
	fprintf(stderr, "Bus error signal (10) on processor %d.\n",
		virtual_processor());
    } else {
	fprintf(stderr, "Signal %d on processor %d.\n", signo,
		virtual_processor());
    }
#ifdef HAVE_SIGINFO_H
    if (siginfo.si_code == SEGV_ACCERR) {
	fprintf(stderr, "Signal code suggests protection violation and a possible stack overflow.\n");
    }
    fprintf(stderr, "signo %d, code %d, stacktop %lx\n",
	    signo, siginfo.si_code, (long) current_thread->stack_top);

#else				/* indef  HAVE_SIGINFO_H */

    if (FC_CODE(code) == FC_OBJERR)
	code = FC_ERRNO(code);

    if (code == SEGV_PROT) {
	fprintf(stderr, "Signal code suggests protection violation and a possible stack overflow.\n");
    }
    fprintf(stderr, "signo %d, code %d, addr %lx, stacktop %lx\n",
	    signo, code, (long) addr, (long) current_thread->stack_top);
#endif				/* HAVE_SIGINFO_H */

    fprintf(stderr, "Thread ");
    fprint_thread(stderr, current_thread);
    fprintf(stderr, " was active.\nTerminating program.\n");

#ifdef HAVE_SIGINFO_H
    general_hndlr(signo, siginfo, context);
#else
    general_hndlr(signo, code, scp, addr);
#endif
    return 0;
}

#ifdef HAVE_SIGINFO_H
int
general_hndlr(signo, siginfo, context)
int signo;
siginfo_t siginfo;
void *context;
#else
int
general_hndlr(signo, code, scp, addr)
int signo, code;
void *scp;
char *addr;
#endif

{
#if !defined(__svr4__) && !defined(__SVR4)
    int oldmask;
    oldmask = sigblock(sigmask(SIGCHLD));
#else
    sighold(SIGCHLD);
#endif

    DBG1("In general_hndlr, signo = %d\n", signo);
    if (virtual_processor() != 0) {
	abnormal(signo, getpid());
	exit(1);
    } else {
	/* the parent shutdown everything */
	child_exit_handler(processes[0], signo);

#if !defined(__svr4__) && !defined(__SVR4)
	sigsetmask(oldmask);
#else
	sigrelse(SIGCHLD);
#endif
    }
    return 0;
}


#ifdef HAVE_SIGINFO_H
int
chld_hndlr(signo, siginfo, context)
int signo;
siginfo_t siginfo;
void *context;
#else
int
chld_hndlr(signo, code, scp, addr)
int signo, code;
void *scp;
char *addr;
#endif
{
    WAIT_STAT_TYPE stat;
    int wid;

    DBG("In chld_hndlr\n");
#if !defined(__svr4__) && !defined(__SVR4)
    while ((wid = wait3((int *) &stat, WNOHANG | WUNTRACED, NULL)) != 0) {
	DBG2("Wait3 returns %ld, stat = %x\n", (long) wid, stat);
	if (wid == -1) {
	    DBG("Returning from chld_hndlr\n");
	    return 0;
	}
	child_exit_handler(wid, stat);
    }
#else
    while ((wid = waitpid((pid_t) - 1, (int *) &stat, WNOHANG | WUNTRACED)) != 0) {
	DBG2("Waitpid returns %ld, stat = %lx\n", (long) wid, (long) stat);
	if (wid == -1) {
	    DBG("Returning from chld_hndlr\n");
	    return 0;
	}
	child_exit_handler(wid, stat);
    }
#endif
    DBG("Returning from chld_hndlr\n");
    return 0;
}


void
child_exit_handler(int wid, WAIT_STAT_TYPE stat)
{
    int i, force_parent_exit = 0;

    DBG("In Child Exit Handler\n");

    total_procs_exit--;
    if (TERM_SIG(stat) != 0 || RET_CODE(stat) != 0) {
	DBG3("Child %d terminated with termsig %d, retcode %d\n",
	     wid, TERM_SIG(stat), RET_CODE(stat));

	if (parent_waiting_the_end < 0) {
	    parent_waiting_the_end = 1;

	    /* 
	     * the parent is already ended and waiting for the children
	     * signal all the remaining children to exit 
	     */
	    for (i = 1; i < total_procs_forked; i++) {
		if (processes[i] != wid) {
		    DBG1("Killing child %lx\n", processes[i]);
		    kill(processes[i], SIGKILL);
		}
	    }
	} else if (parent_waiting_the_end == 0) {
	    parent_waiting_the_end = 1;
	    force_parent_exit = 1;
	    if (wid != processes[0])
		total_procs_exit--;

	    for (i = 1; i < total_procs_forked; i++) {
		processor[i].terminate = TERMINATE;
	    }
	    sleep(1);
	    for (i = 1; i < total_procs_forked; i++) {
		if (processes[i] != wid) {
		    DBG1("Killing child %lx\n", processes[i]);
		    kill(processes[i], SIGKILL);
		}
	    }
	} else {
	    /* the parent is already signaled to finish */
	}
	abnormal(stat, wid);
    } else {
	DBG("Normal exit\n");
	if (wid == processes[0]) {
	    force_parent_exit = 1;
	}
    }

    if (total_procs_exit <= 0) {
	clean_shut_down();
	if (!parent_in_exit)
	    exit(0);
    }
    if (force_parent_exit != 0) {
	while (total_procs_exit > 0) {

#if !defined(__svr4__) && !defined(__SVR4)
	    sigpause(0);	/* wait for the children to shut down */
#else
	    /* SVR4 looses the SIGCHLD while held */
	    wid = waitpid((pid_t) - 1, (int *) &stat, WUNTRACED);
	    if (wid == -1) {
		if (errno == ECHILD) {
		    total_procs_exit = 0;
		} else if (errno != EINTR) {
		    perror("Waitpid error");
		}
	    } else {
		total_procs_exit--;
		abnormal(stat, wid);
	    }
#endif
	}
	clean_shut_down();
	if (!parent_in_exit)
	    exit(0);
    }
}

void
child_exit(status)
int status;
{
    DBG1("in nonLWP child_exit %d\n", virtual_processor());
    if (virtual_processor() == 0) {
	int i;
	if (parent_waiting_the_end == 0) {
	    /* 
	     * the loop is mainly meant to shut down the steering and 
	     * local monitor processes 
	     */

	    for (i = 1; i < total_procs_forked; i++) {
		processor[i].terminate = TERMINATE;
	    }
	}
	child_exit_handler(processes[0], 0);
    }
    if (!parent_in_exit) {
	DBG("in nonLWP child_exit exitting\n");
	exit(0);
    }
}

#endif				/* USE_LWPS */

void
clean_shut_down()
{
    close_global_io();
    shmdt(shm_base);
    shmctl(shm_id, IPC_RMID, 0);
    close(shm_fn);
    unlink(".shm_mem");

    shmdt(monmem_base);
    shmctl(monmem_id, IPC_RMID, 0);
    close(monmem_fn);
    unlink(".mon_mem");
}

void
doinit(start_func, procs)
int (*start_func) ();
int procs;
{
    int node;
    long *shared_shorts;
    int total_procs = procs;

    if (monitoring_enabled) {
	total_procs += number_of_lms;
	if (steering_enabled)
	    total_procs++;
    }
    /* 
     * Initialize the signal handler data structure.
     */
    init_sighandler_struct();

    create_shared_arena(total_procs);
    if (monitoring_enabled)
	sparc_monitorinit(total_procs);

    shared_shorts = (long *) allocate_and_share(SHARED_SHORT_SPACE, -1);

    num_of_procs = &real_num_of_procs;
    *num_of_procs = procs;
    total_threads = (short *) (shared_shorts++);
    *total_threads = 0;
    total_active = (short *) (shared_shorts++);
    *total_active = 0;
    sync_init = (short *) (shared_shorts++);
    *sync_init = total_procs;
    monitor_status = (short *) (shared_shorts++);
    *monitor_status = (short) MONITOR_NOT_INITIALIZED;
    ss_lock = (LOCK *) (shared_shorts++);
    LOCK_INIT(*ss_lock);

    print_sema = (LOCK *) (shared_shorts++);
    *print_sema = 0;

    init_shared_data(total_procs);

    DBG1("Before procinit total_procs %d \n", total_procs);
    for (node = 0; node < total_procs; node++) {
	procinit(node);
	DBG1("procinit for node %d done \n", node);
    }

    DBG("Per processor structures inited\n");

    /* prevent the master to exit without shutting down the children */
#if !defined(__svr4__) && !defined(__SVR4)

    on_exit(master_on_exit, 0);
#else
    atexit(master_on_exit);
#endif
    setup_master_sighandlers(general_hndlr, general_hndlr);

    if (init_global_io() < 0)
	exit(1);

    fork_children_and_wait(start_func, total_procs);

    /* never returns */
}


any_t
allocate_and_share(size, node)
unsigned size;
int node;
{
    char *address;
    address = (char *) ((int) shm_base + shm_offset);
    shm_offset = shm_offset + size;
    if (shm_offset >= shm_limit) {
	printf("WARNING: Shared Memory Exhausted offset %ld, limit %ld, size %d\n",
	       shm_offset, shm_limit, size);
    }
    return ((any_t) address);
}

any_t
allocate_and_copy(size)
unsigned size;
{
    char *address;

    if ((address = (char *) malloc(size)) == NULL) {
	perror("Internal allocate_and_copy");
	exit(1);
    }
    return ((any_t) address);
}

char *
get_stack_memory(size)
int size;
{
    return (char *) allocate_and_copy(size);
}


extern RESULT
init_convert(procs)
int procs;
{
    int i, j;

    /* if the processor is online, enter in the conversion table */
    j = 0;
    for (i = 0; i < procs; i++) {
	processor[j++].logical_to_phys = i;
    }
    return (T_SUCCEED);
}				/* init_convert */

void
check_num_procs(procs)
int procs;
{
    int total_procs = procs;
    if (monitoring_enabled) {
	total_procs += number_of_lms;	/* procs reserved for monitoring */
	if (steering_enabled)
	    total_procs++;
    }
    if (total_procs > MAX_PROC) {
	fprintf(stderr, "Requested number of processors exceeds Cthreads internal limits\n");
	exit(1);
    }
}

int
check_stack_size(size)
int size;
{
    size += getpagesize();	/* add one page for protection */

    /* round up to the nearest integral multiple of page size */
    if ((size % getpagesize()) != 0) {
	size = ((size / getpagesize()) + 1) * getpagesize();
    }
    return size;
}

void *
allocate_monmem(size, node)
int size;
int node;
{
    long n;
    char *address;

    if (!use_shared_memory) {
	return malloc(size);
    }
    address = (char *) monmem_base + monmem_offset;
    if ((n = (unsigned long) address % sizeof(long))) {
	address += sizeof(long) - n;
	monmem_offset += sizeof(long) - n;
    }
    monmem_offset = monmem_offset + size;
    if (monmem_offset >= monmem_limit) {
	printf("WARNING: Shared Monitoring Memory Exhausted offset %ld, limit %ld, size %d\n",
	       monmem_offset, monmem_limit, size);
    }
    return ((void *) address);
}

static void
sparc_monitorinit(procs)
int procs;
{
    int size;

    DBG2("monitor init, procs is %d, %d\n", procs,
	 config.threads_per_proc);
    size = generic_monmem_size(procs);

    if (use_shared_memory) {
	if ((monmem_fn = open(".mon_mem", O_RDWR | O_CREAT | O_TRUNC, 0777)) == -1) {
	    perror(".mon_mem");
	}
	monmem_key = ftok(".mon_mem", '1');

	DBG1("getting shared monitor memory, size is %d\n", size);
	monmem_id = shmget(monmem_key, size, IPC_CREAT | 0600);
	if (monmem_id == -1) {
	    perror("SHared Memory GET failed  Monmem!");
	    shmctl(monmem_id, IPC_RMID, 0);
	    close(monmem_fn);
	    unlink(".mon_mem");
	    exit(1);
	}
	monmem_base = (char *) shmat(monmem_id, 0, 0);
	if (monmem_base == (char *) -1) {
	    perror("Cannot create monitor shared memory  monmem");
	    shmctl(monmem_id, IPC_RMID, 0);
	    close(monmem_fn);
	    unlink(".monmem_mem");
	    exit(1);
	}
	monmem_offset = 0;
	monmem_limit = size;
    }
    monitorinit(procs);
}

void
relinquish_processor()
{
    struct timeval delay_time;
    delay_time.tv_sec = 0;
    delay_time.tv_usec = 10000;	/* return millisecs */
    select(32, NULL, NULL, NULL, &delay_time);
}

unsigned long start_time;

unsigned long
cthread_timestamp()
{
    struct timeval tp0;
    unsigned long timestamp;

    extern int gettimeofday();

    if (logicalTimeStamp) {
	internal_mutex_lock(&logicalTimeStamp_lock->mlock);
	timestamp = (*logicalTimeStamp_counter)++;
	internal_mutex_unlock(&logicalTimeStamp_lock->mlock);
    } else {
	gettimeofday(&tp0, NULL);
	timestamp = (unsigned long) tp0.tv_sec * 1000000 + tp0.tv_usec;
	timestamp -= start_time;
    }
    return timestamp;
}

void
get_start_timestamp_and_resolution(start_time_p, resolution_p)
unsigned long *start_time_p;
unsigned long *resolution_p;
{
    struct timeval tp0;

    extern int gettimeofday();

    if (logicalTimeStamp) {
	*resolution_p = 1;
	start_time = 0;
    } else {
	gettimeofday(&tp0, NULL);
	*resolution_p = 1000000;
	start_time = (unsigned long) tp0.tv_sec * 1000000 + tp0.tv_usec;
    }
    *start_time_p = start_time;
}
