
#ifndef _LIFE_TIME_POLKA_H_
#define _LIFE_TIME_POLKA_H_

#include "tuxian.H"
#include "palette.H"
#include <polka.H>

class TimedView: public View {
  private:
    double  h_zoom_ratio, v_zoom_ratio;
    double  h_move_ratio, v_move_ratio;
    ResizeMode resize_mode;
    int     screen_ticks;	/* ticks to draw on each screen */
    Widget  timelinewid;	/* a drawing area widget for time stamp line */
    Window  timelinewin;        /* the drawing window */
    Widget  clrmapwid;		/* color to thread state mapping */
    Window  clrmapwin;
    Font    timevalFont, tslabelFont, legendFont;
    XFontStruct *timevalFontStruct, *tslabelFontStruct, *legendFontStruct;

  protected:
    void setupWindow(char*);

  public:
    TimedView(double h_zoom = 0.5, double v_zoom = 0.5,
	      double h_move = 0.5, double v_move = 0.5,
	      ResizeMode rm=CoordStretch, double start_time=0,
	      double screen_time=2, int ticks=20);
    int Create(const char *title="Threads Display", int wid=0, int hei=0,
	       const char *viewlabel="Threads Lifetime View",
	       const char *timelabel="Time Stamp");
    ~TimedView() {};
    double  start_ts;		/* timestamp value at the left edge */
    double  screen_length;	/* timestamp across a screen */
    void DrawTimeLineTicks();
    void DrawTimeValues();
    void DrawTSLabel();
    void DrawLegend();
    void ClearTimeValueArea();
};

#define MAX_SECONDS_INIT     60
#define MAX_SECONDS_INCR     60

class ThreadLifeTime_polka: public ThreadLifeTime {
  private:
    int global_time;
    int max_num_threads;
    double screen_width;		// screen width in term of secs
    double screen_width_units;		// screen_width * timestamp_resolution
    double screen_width_units_inv;      // inverse of screen_width_units

    int  max_seconds;			// max seconds of time line
    double origx, origy;		// position of the origin
    double *xpos, *ypos;		// position of las state
    int  *position_index;		// relative vertical postion of thrd
    int  next_position, max_num_positions;
    int  *position_array;		// inverse of position_index
    int  num_threads_in_unit_window;    // num threads in 1.0 height of win
    double yincr;			// y increment for each new thread
    double thickness;			// thickness of threads lines
    double h_zoom_ratio, v_zoom_ratio;
    double h_move_ratio, v_move_ratio;
    ResizeMode resize_mode;
    Palette  *color_depot;		// color names

  protected:
    TimedView *polkaview;		// polka view
    void lifetimeInitialize();
    void lifetimeNewThread(int, unsigned long, int, char*);
    void lifetimeJoinThread(int, unsigned long, int);
    void lifetimeChangeState(int, unsigned long, ThreadStateType);
    void lifetimeHScroll(double, double);

  public:
    ThreadLifeTime_polka(int threads_in_win = 16,
			 double screen_width = 2.0,
			 double h_zoom = 0.5, double v_zoom = 0.5,
			 double h_move = 0.5, double v_move = 0.5,
			 int resize_const = False,
			 int num_of_colors = 0,
			 char **names_of_colors = NULL);
    ~ThreadLifeTime_polka();
    void draw() {
	global_time = polkaview->FastAnimate(global_time, 1); }
};

#endif
