/***************************************************************/
/*							       */
/*			     POLKAGROUP.C		       */
/*							       */
/***************************************************************/
/*       Copyright 1993 Georgia Institute of Technology 
                     -- John T. Stasko                         */

#include "polka_local.H"

/***************************************************************/
/*							       */
/*							       */
/***************************************************************/


Loc1Desc::Loc1Desc()
{
   num = 0;
}


int
Loc1Desc::Make(Loc *l[])
{
   double x,y,dx,dy;
   int i;

   if (num <= 0) return(0);
   if (num == 1)
       { l[0] = new Loc(fx,fy);
         return(1);
     }

   x=fx; y=fy;
   dx = (tx - fx) / (num-1);
   dy = (ty - fy) / (num-1);
   for (i=0; i<num; i++)
       { l[i] = new Loc(x,y);
         x += dx;
         y += dy;
     }
   return(1);
}



/***************************************************************/
/*							       */
/*							       */
/***************************************************************/

Loc2Desc::Loc2Desc()
{
   rows = cols = 0;
}

int
Loc2Desc::Make(Loc *l[])
{
   double x,y,dx,dy;
   int i,r,c;

   if (rows <= 0) return(0);
   if (cols <= 0) return(0);

   y=lly;
   dx = (cols==1 ? 0.0 : (urx - llx) / (cols - 1));
   dy = (rows==1 ? 0.0 : (ury - lly) / (rows - 1));
   for (i=0,r=0; r<rows; r++) { 
     for (c=0,x=llx; c<cols; c++) { 
         l[i] = new Loc(x,y);
         i++;
         x += dx;
     }
     y += dy;
   }
   return(1);
}




/***************************************************************/
/*							       */
/*							       */
/***************************************************************/

AnimObjectGroup::AnimObjectGroup()
{
   horiz = 1;
   align = AlignBottom;
   spacetype = SpacingSame;
   useints = 0;
   usedoubles = 0;
   vis = 1;
   strcpy(color,"black");
}




/***************************************************************/
/*							       */
/*							       */
/***************************************************************/

LineGroup::LineGroup()
{
   linewidth = 0.0;
   linestyle = 1.0;
   arrow = 0;
}




/***************************************************************/
/*							       */
/*							       */
/***************************************************************/

RectangleGroup::RectangleGroup()
{
   fill = 0.0;
}




/***************************************************************/
/*							       */
/*							       */
/***************************************************************/

CircleGroup::CircleGroup()
{
   fill = 0.0;
}




/***************************************************************/
/*							       */
/*   Make - create a row or column of AnimObjects              */
/*      subject to the given parameters.        	       */
/*							       */
/***************************************************************/

int
LineGroup::Make(View *v, Line *o[], int n, 
                double lx, double by, double rx, double ty)
{
   double s,locx,locy,sx,sy;
   int i;

   locx = lx;
   locy = by;
   for (i=0; i<n; i++) {
      if (horiz) {
         if (spacetype == SpacingNone) s = (rx-lx)/(n-1);
         else if (spacetype == SpacingSame) s = (rx-lx) / (n-1);
         else if (spacetype == SpacingAbsolute) s = spacing;

         if (useints)
            sy = (double)(intvals[i]-intmin) / 
                       (double)(intmax-intmin) * (ty-by);
         else if (usedoubles)
            sy = (doublevals[i]-doublemin) / 
                      (doublemax-doublemin) * (ty-by);
         else
            sy = ty-by;

         if (align == AlignBottom)
            o[i] = new Line(v,vis,locx,locy,0.0,sy,color,
                              linewidth,linestyle,arrow);
         else if (align == AlignMiddle)
            o[i] = new Line(v,vis,locx,by+(ty-by)/2.0-sy/2.0,
                             0.0,sy,color,linewidth,
                             linestyle,arrow);
         else if (align == AlignTop)
            o[i] = new Line(v,vis,locx,ty-sy,0.0,sy,color,
                             linewidth,linestyle,arrow);
         else /* bad */
            o[i] = new Line(v,vis,locx,locy,0.0,sy,color,
                              linewidth,linestyle,arrow);

         locx += s;
      }
      else /* vertical */ {
         if (spacetype == SpacingNone) s = (ty-by)/(n-1);
         else if (spacetype == SpacingSame) s = (ty-by) / (n-1);
         else if (spacetype == SpacingAbsolute) s = spacing;

         if (useints)
            sx = (double)(intvals[i]-intmin) / 
                               (double)(intmax-intmin) * (rx-lx);
         else if (usedoubles)
            sx = (doublevals[i]-doublemin) / (doublemax-doublemin) * (rx-lx);
         else
            sx = rx-lx;

         if (align == AlignLeft)
            o[i] = new Line(v,vis,locx,locy,sx,0.0,color,
                             linewidth,linestyle,arrow);
         else if (align == AlignMiddle)
            o[i] = new Line(v,vis,lx+(rx-lx)/2.0-sx/2.0,locy,
                             sx,0.0,color,linewidth,linestyle,arrow);
         else if (align == AlignRight)
            o[i] = new Line(v,vis,rx-sx,locy,sx,0.0,color,
                             linewidth,linestyle,arrow);
         else /* bad */
            o[i] = new Line(v,vis,locx,locy,sx,0.0,color,
                              linewidth,linestyle,arrow);

         locy += s;
      }
   }
   return (1);
}  





int
RectangleGroup::Make(View *v, Rectangle *o[], int n, 
                        double lx, double by, double rx, double ty)
{
   double s,locx,locy,sx,sy;
   int i;

   locx = lx;
   locy = by;
   for (i=0; i<n; i++) {
      if (horiz) {
         if (spacetype == SpacingNone) s = 0.0;
         else if (spacetype == SpacingSame) s = (rx-lx) / (2.0 * n - 1.0);
         else if (spacetype == SpacingAbsolute) s = spacing;

         sx = ((rx-lx)-(s*(n-1))) / (double)n;

         if (useints)
            sy = (double)(intvals[i]-intmin) / 
                     (double)(intmax-intmin) * (ty-by);
         else if (usedoubles)
            sy = (doublevals[i]-doublemin) / 
                      (doublemax-doublemin) * (ty-by);
         else
            sy = ty-by;

         if (align == AlignBottom)
            o[i] = new Rectangle(v,vis,locx,locy,sx,sy,color,fill);
         else if (align == AlignMiddle)
            o[i] = new Rectangle(v,vis,locx,by+(ty-by)/2.0-sy/2.0,
                             sx,sy,color,fill);
         else if (align == AlignTop)
            o[i] = new Rectangle(v,vis,locx,ty-sy,sx,sy,color,fill);
         else /* bad */
            o[i] = new Rectangle(v,vis,locx,locy,sx,sy,color,fill);

         locx += sx + s;
      }
      else /* vertical */ {
         if (spacetype == SpacingNone) s = 0.0;
         else if (spacetype == SpacingSame) s = (ty-by) / (2.0 * n - 1.0);
         else if (spacetype == SpacingAbsolute) s = spacing;

         sy = ((ty-by)-(s*(n-1))) / (double)n;

         if (useints)
            sx = (double)(intvals[i]-intmin) / 
                       (double)(intmax-intmin) * (rx-lx);
         else if (usedoubles)
            sx = (doublevals[i]-doublemin) / 
                      (doublemax-doublemin) * (rx-lx);
         else
            sx = rx-lx;

         if (align == AlignLeft)
            o[i] = new Rectangle(v,vis,locx,locy,sx,sy,color,fill);
         else if (align == AlignMiddle)
            o[i] = new Rectangle(v,vis,lx+(rx-lx)/2.0-sx/2.0,locy,
                             sx,sy,color,fill);
         else if (align == AlignRight)
            o[i] = new Rectangle(v,vis,rx-sx,locy,sx,sy,color,fill);
         else /* bad */
            o[i] = new Rectangle(v,vis,locx,locy,sx,sy,color,fill);

         locy += sy + s;
      }
   }
   return (1);
}  







int
CircleGroup::Make(View *v, Circle *o[], int n, 
                   double lx, double by, double rx, double ty)
{
   double s,locx,locy,rad;
   int i;

   locx = lx;
   locy = by;
   for (i=0; i<n; i++) {
      if (horiz) {
         if (spacetype == SpacingNone) s = 0.0;
         else if (spacetype == SpacingSame) s = (rx-lx) / (2.0 * n - 1.0);
         else if (spacetype == SpacingAbsolute) s = spacing;

         if (useints)
            rad = (double)(intvals[i]-intmin) / 
                     (double)(intmax-intmin) * (ty-by) * 0.5;
         else if (usedoubles)
            rad = (doublevals[i]-doublemin) / 
                      (doublemax-doublemin) * (ty-by) * 0.5;
         else
            rad = ((rx-lx)-(s*(n-1))) / (2.0 * n);
   
         if (align == AlignBottom)
            o[i] = new Circle(v,vis,locx+rad,locy+rad,rad,color,fill);
         else if (align == AlignMiddle)
            o[i] = new Circle(v,vis,locx+rad,by+(ty-by)/2.0,rad,color,fill);
         else if (align == AlignTop)
            o[i] = new Circle(v,vis,locx+rad,ty-rad,rad,color,fill);
         else /* bad */
            o[i] = new Circle(v,vis,locx+rad,locy+rad,rad,color,fill);

         locx += 2.0 * rad + s;
      }
      else /* vertical */ {
         if (spacetype == SpacingNone) s = 0.0;
         else if (spacetype == SpacingSame) s = (ty-by) / (2.0 * n - 1.0);
         else if (spacetype == SpacingAbsolute) s = spacing;

         if (useints)
            rad = (double)(intvals[i]-intmin) / 
                        (double)(intmax-intmin) * (rx-lx);
         else if (usedoubles)
            rad = (doublevals[i]-doublemin) / 
                      (doublemax-doublemin) * (rx-lx);
         else
            rad = ((ty-by)-(s*(n-1))) / (2.0 * n);

         if (align == AlignLeft)
            o[i] = new Circle(v,vis,locx+rad,locy+rad,rad,color,fill);
         else if (align == AlignMiddle)
            o[i] = new Circle(v,vis,lx+(rx-lx)/2.0,locy+rad,rad,color,fill);
         else if (align == AlignRight)
            o[i] = new Circle(v,vis,rx-rad,locy+rad,rad,color,fill);
         else /* bad */
            o[i] = new Circle(v,vis,locx+rad,locy+rad,rad,color,fill);

         locy += 2.0 * rad + s;
      }
   }
   return (1);
}  



