/**********************************************************************
 **                                                                  **
 **                     database_sharing.h                           **
 **                                                                  **
 **                                                                  **
 **  Written by:  Zsolt Kira                                         **
 **                                                                  **
 **  This file contains functions for sharing of information between **
 **  robots regarding their positions, objects, etc.  This is done   **
 **  via the database broadcast mechanism.                           **
 **                                                                  **
 **  Copyright 2009, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef DATABASE_SHARING_H
#define DATABASE_SHARING_H

#include "gt_simulation.h"

#define DATABASE_SHARING_MAX_NUM_OBJECTS 1000
#define DATABASE_SHARING_MAX_NUM_ROBOTS 100

#define DATABASE_SHARING_MAX_STRING_LENGTH 10000


/////// 
// Sharing of positions between robots
/////// 

// Number of robot positions
int database_get_num_robot_positions();

// Fill in robot positions
// new_num_robots contains number of robot positions successfully obtained
// (some may fail due to clobbering or other problems)
int database_get_num_world_robot_positions();
void database_get_robot_positions();
void database_fill_robot_positions(RobotReadings &rtn, int &new_num_robots);

// Broadcasts robot position to others
void database_put_robot_position();


/////// 
///Sharing of objects between robots
//////

// Gets number of objects from a robot
int database_get_num_objects(int robot_id);
void database_get_objects(int robot_id);

// Fills in objects from the world objects structure.
int database_get_num_world_objects();
void database_fill_objects(ObjectList &objs, int &new_num_objects, int robot_id, int objects_start_index = 0);

// Broadcast object to others
void database_put_objects(ObjectList &objs, int objects_start_index, int rid = -1);

#endif
