/**********************************************************************
 **                                                                  **
 **                              hclient.c                           **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Hardware client. Connects to hserver                            **
 **                                                                  **
 **  Copyright 2000 - 2008, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: hclient.c,v 1.13 2010-05-03 22:08:34 zkira Exp $ */

#include <stdio.h>
#include <termio.h>
#include <math.h>
#include <sys/file.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <linux/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "ipt/ipt.h"
#include "ipt/message.h"

#include "hserver/hserver.h"
#include "hserver/hserver_ipt.h"
#include "hclient.h"
#include "robot_side_com.h"

const int NUM_GPS_READINGS = 7;

//variable added to disable printing sensor data to terminal
const bool PRINT_READING = false;

extern int     robot_id;

robot_setup_t *rs = NULL;
const int MAX_SENSORS = 29;        // NOTE: This must be changed according to hserver/hserver.h SENSOR types
                                   // i.e. this must be greater than the largest SENSOR constant value + 1
int sensor_setup_link[MAX_SENSORS];
int num_sensors;
sensor_t sensor[MAX_SENSORS];

char hclient_ipt_name[100];
char hserver_ipt_name[100];
char *ipt_server_name;
char ipt_name[100];

IPCommunicator *hc_communicator;
IPConnection *hserver_connection;

/* Print robot setup to stdout */
int robotsetupPrintStdout(REC_RobotSetupInfo *rs){
    int i;
// All the sensors!
    info_pantilt_camera_t *info_pantilt;
    info_laser_t *info_laser;
    info_sonar_t *info_sonar;
    info_rfid_t *info_rfid;
    info_comm_t *info_comm;
    info_chemical_t *info_chemical;
    info_enemy_t *info_enemy;
    info_ir_t *info_ir;
    info_cog_t *info_cog;
    info_video_t *info_video;
    info_xyt_t *info_xyt;

    printf("Robot Setup\n");
    printf("name: %s\n",rs->name);
    printf("number of parts: %d\n",rs->num_parts);
    for(i=0;i<rs->num_parts;i++){
        switch(rs->part[i].type){
        case SENSOR_PANTILT_CAMERA:
            info_pantilt = (info_pantilt_camera_t*)rs->part[i].info;
            printf("%d.  pan/tilt camera   angle: %.2f  fov x: %.2f y: %.2f\n",i,
                   info_pantilt->angle,info_pantilt->fov_x,info_pantilt->fov_y);
            break;
        case SENSOR_XYT:
            info_xyt = (info_xyt_t*)rs->part[i].info;
            printf("%d.  xyt  scale: %f\n",i,info_xyt->scale);
            break;
        case SENSOR_ARM:
            printf("%d.  arm\n",i);
            break;
        case SENSOR_SICKLMS200_1:
        case SENSOR_SICKLMS200_2:
        case SENSOR_SICKLMS200_3:
        case SENSOR_SICKLMS200_4:
            info_laser = (info_laser_t*)rs->part[i].info;
            printf("%d.  laser  num readings: %d angle: %.2f  fov %.2f\n",i,
                   info_laser->num_readings,info_laser->angle,info_laser->fov);
            break;
        case SENSOR_SONAR:
            info_sonar = (info_sonar_t*)rs->part[i].info;
            printf("%d.  sonar  number: %d  fov: %.2f  range: %.2f\n",i,
                   info_sonar->number,info_sonar->fov,info_sonar->range);
            break;
        case SENSOR_RFID:
            info_rfid = (info_rfid_t*)rs->part[i].info;
            printf("%d.  RFID  Max Number: %d (size of array:%d)\n",i,
                   info_rfid->number/4, info_rfid->number);
            break;
        case SENSOR_COMM:
            info_comm = (info_comm_t*)rs->part[i].info;
            printf("%d.  Comm  Max Number: %d (size of array:%d)\n",i,
                   (info_comm->number - 1)/2, info_comm->number);
            break;
        case SENSOR_CHEMICAL:
            info_chemical = (info_chemical_t*)rs->part[i].info;
            printf("%d.  Chemical  Max Number: %d (size of array:%d)\n",i,
                   (info_chemical->number -1)/6, info_chemical->number);
            break;
        case SENSOR_ENEMY:
            info_enemy = (info_enemy_t*)rs->part[i].info;
            printf("%d.  Enemy  Max Number: %d (size of array:%d)\n",i,
                   (info_enemy->number - 1)/7, info_enemy->number);
            break;
        case SENSOR_IR:
            info_ir = (info_ir_t*)rs->part[i].info;
            printf("%d.  ir  number: %d  fov: %.2f  range: %.2f\n",i,
                   info_ir->number,info_ir->fov,info_ir->range);
            break;
        case SENSOR_COGNACHROME:
            info_cog = (info_cog_t*)rs->part[i].info;
            printf("%d.  cognachrome  num channels: %d  max per channel: %d\n",i,
                   info_cog->num_channels,info_cog->max_per_channel);
            break;
        case SENSOR_VIDEO:
            info_video = (info_video_t*)rs->part[i].info;
            printf("%d.  video  num channels: %d\n",i,info_video->num_channels);
            break;
        case SENSOR_COMPASS:
            printf("%d.  compass\n",i);
            break;
        case SENSOR_GPS:
            printf("%d.  gps\n",i);
            break;
	case SENSOR_CRICKET:
            printf("%d.  cricket\n",i);
	    break;
        default:
            printf("%d.  Unknown sensor type\n",i);
        }
    }
    return(0);
}

/* Send simple query to hserver */
int hclient_send_simple_query(int command){
    REC_SimpleCommand rec,*rtn;
    int rtn_cmd;

    rec.command = command;
    IPMessage *msg = hc_communicator->Query( hserver_connection, MSG_SimpleCommand, 
                                             &rec, MSG_SimpleCommand );
    rtn = (REC_SimpleCommand*)msg->FormattedData();
    rtn_cmd = rtn->command;
    delete msg;
    return(rtn_cmd);
}

/* Send simple command to hserver */
void hclient_send_simple_command(int command){
    REC_SimpleCommand rec;

    rec.command = command;
    hc_communicator->SendMessage( hserver_connection, MSG_SimpleCommand, &rec );
}

int hclient_frame_upload(int channel,char *name){
    REC_FrameUpload rec;
    REC_SimpleCommand *rtn;
    REC_VideoFrame *vrtn;
    int rtn_cmd,i,j;
    FILE *ofile;
    rec.channel = channel;
    rec.name = name;
    rec.send = 1;
    if (!rec.send) {
        IPMessage *msg = hc_communicator->Query( hserver_connection, MSG_FrameUpload, 
                                                 &rec, MSG_SimpleCommand);
        rtn = (REC_SimpleCommand*)msg->FormattedData();
        rtn_cmd = rtn->command;
        delete msg;
    } else {
        IPMessage *msg = hc_communicator->Query( hserver_connection, MSG_FrameUpload, 
                                                 &rec, MSG_VideoFrame );
        vrtn = (REC_VideoFrame*)msg->FormattedData();
        printf("got video frame h %d w %d\n",vrtn->height,vrtn->width);
        ofile = fopen(name,"w");
        fprintf(ofile,"P6\n %d %d\n 255\n",vrtn->width,vrtn->height);
        fflush(ofile);
	
        for(i=0;i<vrtn->height;i++)
            for(j=0;j<vrtn->width;j++){
                fwrite(vrtn->frame + (i * vrtn->width + j)*3+2, 1, 1, ofile);
                fwrite(vrtn->frame + (i * vrtn->width + j)*3+1, 1, 1, ofile);
                fwrite(vrtn->frame + (i * vrtn->width + j)*3  , 1, 1, ofile);
            }
        fclose(ofile);
        delete msg;
    }
    return(rtn_cmd);
}


FrameData hclient_frame_grab(void) {

    REC_FrameUpload rec;
    //    REC_SimpleCommand *rtn;
    REC_VideoFrame *vrtn;
    int rtn_cmd,i,j;
    FILE *ofile;
    rec.channel = 0;
    rec.name = 0;
    rec.send = 1;
 
    IPMessage *msg = hc_communicator->Query( hserver_connection, MSG_FrameUpload, &rec, MSG_VideoFrame );
    vrtn = (REC_VideoFrame*)msg->FormattedData();
    //printf("got video frame h %d w %d\n",vrtn->height,vrtn->width);

    FrameData fd;
    fd.Data = NULL;
    fd.DataSize = 0;

    fd.Data = vrtn->frame;
    fd.DataSize = vrtn->size;
    printf("Got video frame of size: %d\n", fd.DataSize);

    return fd;

    //ofile = fopen(name,"w");
    //fprintf(ofile,"P6\n %d %d\n 255\n",vrtn->width,vrtn->height);
    // fflush(ofile);
	
    //for(i=0;i<vrtn->height;i++)
    //for(j=0;j<vrtn->width;j++){
    //          fwrite(vrtn->frame + (i * vrtn->width + j)*3+2, 1, 1, ofile);
    //          fwrite(vrtn->frame + (i * vrtn->width + j)*3+1, 1, 1, ofile);
    //          fwrite(vrtn->frame + (i * vrtn->width + j)*3  , 1, 1, ofile);
    //      }
    //  fclose(ofile);
    //  delete msg;
    //}
    //return(rtn_cmd);

}



/* Retrieve robot setup from hserver */
void hclient_get_robot_setup(){
    REC_SimpleCommand rec;
    int i,type;
    info_cog_t *info_cog;
    info_laser_t *info_laser;
    info_rfid_t *info_rfid;
    info_comm_t *info_comm;
    info_chemical_t *info_chemical;
    info_enemy_t *info_enemy;
    info_sonar_t *info_sonar;
    info_ir_t *info_ir;
    
    fprintf(stderr,"get robot setup\n");

    if (rs) delete rs;
    rec.command = HCLIENT_GET_ROBOT_SETUP_CMD;
    IPMessage *msg = hc_communicator->Query( hserver_connection, MSG_SimpleCommand, 
                                             &rec, MSG_RobotSetupInfo );
    REC_RobotSetupInfo *robot_setup = (REC_RobotSetupInfo*)msg->FormattedData(1);
    robotsetupPrintStdout(robot_setup);
    rs = (robot_setup_t*) robot_setup;
    for(i=0;i<rs->num_parts;i++){
        type = rs->part[i].type;
        sensor_setup_link[type] = i;
        if (sensor[type].type == -1) {
            sensor[type].type = type;
            sensor[type].len_extra = 0;
            sensor[type].num_angles = 0;
            sensor[type].num_readings = 0;
            switch(type){
            case SENSOR_SONAR:
                info_sonar = (info_sonar_t*)rs->part[i].info;
                sensor[type].num_readings = info_sonar->number;
                sensor[type].num_angles = info_sonar->number;
                sensor[type].len_extra = sizeof(float)*info_sonar->number*3;
                break;
            case SENSOR_IR:
                info_ir = (info_ir_t*)rs->part[i].info;
                sensor[type].num_readings = info_ir->number;
                sensor[type].num_angles = info_ir->number;
                break;
            case SENSOR_PANTILT_CAMERA:
                sensor[type].num_angles = 4;
                break;
            case SENSOR_SICKLMS200_1:
            case SENSOR_SICKLMS200_2:
            case SENSOR_SICKLMS200_3:
            case SENSOR_SICKLMS200_4:
                info_laser = (info_laser_t*)rs->part[i].info;
                sensor[type].len_extra = sizeof(extra_laser_t);
                sensor[type].num_angles = info_laser->num_readings;
                sensor[type].num_readings = info_laser->num_readings;
                break;
            case SENSOR_XYT:
                sensor[type].num_readings = 6;
                break;
            case SENSOR_COGNACHROME:
                info_cog = (info_cog_t*)rs->part[i].info;
                sensor[type].len_extra = sizeof(extra_sensor_cog_t);
                sensor[type].num_angles = info_cog->num_channels*info_cog->max_per_channel;
                sensor[type].num_readings = info_cog->num_channels*info_cog->max_per_channel;
                break;
            case SENSOR_ARM:
                sensor[type].num_angles = 1;
                break;
            case SENSOR_GPS:
                sensor[type].num_readings = NUM_GPS_READINGS;
                break;
            case SENSOR_RFID:
                info_rfid = (info_rfid_t*)rs->part[i].info;
                sensor[type].len_extra = 0;
                sensor[type].num_angles = 0;
                sensor[type].num_readings = info_rfid->number;
                break;
            case SENSOR_COMM:
                info_comm = (info_comm_t*)rs->part[i].info;
                sensor[type].len_extra = 0;
                sensor[type].num_angles = 0;
                sensor[type].num_readings = info_comm->number;
                break;
            case SENSOR_CHEMICAL:
                info_chemical = (info_chemical_t*)rs->part[i].info;
                sensor[type].len_extra = 0;
                sensor[type].num_angles = 0;
                sensor[type].num_readings = info_chemical->number;
    
                // REMOVE: This is because there is a bug right now in the map and/or ipt communication for these sensors
                //sensor[SENSOR_CHEMICAL].reading[10*6+1] = 0;

                break;
            case SENSOR_ENEMY:
                info_enemy = (info_enemy_t*)rs->part[i].info;
                sensor[type].len_extra = 0;
                sensor[type].num_angles = 0; 
                sensor[type].num_readings = info_enemy->number;
 
                // REMOVE: This is because there is a bug right now in the map and/or ipt communication for these sensors
                //sensor[SENSOR_ENEMY].reading[46*7+1] = 0;

                break;
            case SENSOR_COMPASS:
                sensor[type].num_angles = 1;
                break;
            case SENSOR_CRICKET:
                sensor[type].len_extra = 0;
                sensor[type].num_readings = 0;
                sensor[type].num_angles = 0;
		break;
            default:
                fprintf(stderr,"Unknown robot setup sensor type: %d\n",type);
            }
            if (sensor[type].len_extra)
            {
                sensor[type].extra = (char*) malloc( sensor[type].len_extra );
            }
            if (sensor[type].num_angles)
            {
                sensor[type].angle = (float*)malloc( sizeof(float) * sensor[type].num_angles );
            }
            if (sensor[type].num_readings)
            {
                sensor[type].reading = (float*)malloc( sizeof(float) * sensor[type].num_readings );
            }
        } else fprintf(stderr,"***** error in get robot setup\n");
    }
    num_sensors = rs->num_parts;
    delete msg;
}

/* Initialize ipt connection with hserver */
void hclient_init_ipt_connection(void)
{
    REC_InitConnection rec;
    rec.name = hclient_ipt_name;
    hc_communicator->SendMessage(hserver_connection,MSG_InitConnection,&rec);
}

void hclientUpdateRobotSetup(IPCommunicator *comm, IPMessage *msg, void *hndData)
{
    fprintf(stderr,"*** !@# robot setup\n");
}

void hclientUpdateSensorsHandler(IPCommunicator *comm, IPMessage *msg, void *hndData)
{
    fprintf(stderr,"*** !@# sensors\n");
}

/* Setup ipt */
int hclientSetupIpt(void)
{
    unsigned int i;
    IPMessageType* msg_type;
    
    if (ipt_server_name == NULL){
        ipt_server_name = getenv("IPTHOST");
    }
    if (ipt_server_name == NULL)
        ipt_server_name = getenv("HOST");
    if (ipt_server_name == NULL){
        fprintf(stderr,"Unable to determine iptserver host");
        return(-1);
    }

    hc_communicator = IPCommunicator::Instance(hclient_ipt_name,ipt_server_name);
    hc_communicator->RegisterNamedFormatters(iptHclientFormatArray);
    hc_communicator->RegisterMessages(iptHclientMessageArray);
    
    IPMsgHandlerSpec hndArray[] = {
        { MSG_RobotSetupInfo, hclientUpdateRobotSetup, IPT_HNDL_ALL, NULL},
        { MSG_SensorUpdate, hclientUpdateSensorsHandler, IPT_HNDL_ALL,NULL}
    };
    
    for(i=0;i<(sizeof(hndArray)/sizeof(IPMsgHandlerSpec));i++) {
        msg_type = hc_communicator->LookupMessage( hndArray[i].msg_name );
        hc_communicator->RegisterHandler( msg_type, hndArray[i].callback, 
                                       hndArray[i].data, hndArray[i].context );
    }

    return(0);
}


/* Return side ranges from laser. Used for door trigger */
int hclient_get_side_ranges(float *left,float *right){

    if (sensor[SENSOR_SICKLMS200_1].type != SENSOR_SICKLMS200_1) {
        fprintf(stderr,"No laser information\n");
        return(-1);
    } else if (sensor[SENSOR_SICKLMS200_1].num_readings < 361) {
      fprintf(stderr,"Laser has less than 361 readings! (%d readings)\n", sensor[SENSOR_SICKLMS200_1].num_readings);
      return(-1);     
    }
    
    *right = sensor[SENSOR_SICKLMS200_1].reading[0];
    *left = sensor[SENSOR_SICKLMS200_1].reading[360];

    return(0);
}

/* Set robot speed */
int hclient_drive(float speed)  // m per sec
{
    REC_DriveCommand rec;

    rec.vel = speed;
    hc_communicator->SendMessage(hserver_connection,MSG_DriveCommand,&rec);

    return(0);
}

/* Set robot angular speed */
int hclient_steer(float avel)
{
    REC_SteerCommand rec;
    
    rec.avel = avel; 
    hc_communicator->SendMessage(hserver_connection,MSG_SteerCommand,&rec);

    return(0);
}


/* Steer toward theta and when within drive_wait_angle start to drive at speed */
int hclient_steer_toward_and_drive(
    float speed,
    float theta,
    float altitudeSpeed,
    float drive_wait_angle, 
    int use_reverse)
{
    REC_SteerTowardDriveCommand rec;

    rec.theta = theta;
    rec.speed = speed;
    rec.altitudeSpeed = altitudeSpeed;
    rec.drive_wait_angle = drive_wait_angle;
    rec.use_reverse = use_reverse;
    hc_communicator->SendMessage(hserver_connection,MSG_SteerTowardDriveCommand,&rec);
    return(0);
}


/* Stop robot motion */
int hclient_stop(void)
{
    hclient_send_simple_command(HCLIENT_STOP_CMD);
    return(0);
}


/* Start sonar */
int hclient_range_start()
{
    hclient_send_simple_command(HCLIENT_RANGE_START_CMD);
    return(0);
}

/* Stop sonar */
int hclient_range_stop(void)
{
    hclient_send_simple_command(HCLIENT_RANGE_STOP_CMD);
    return(0);
}

/* Set robot location */
int hclient_setxyt(float x, float y, float theta){
    REC_SetXYT rec;

    rec.x = x;
    rec.y = y;
    rec.t = theta;
    hc_communicator->SendMessage(hserver_connection,MSG_SetXYT,&rec);
    
    return 0;
}


int hclient_move_camera( int pandegree, int tiltdegree ) {
    REC_MoveCamera rec;
    
    rec.pan = pandegree;
    rec.tilt = tiltdegree;
    hc_communicator->SendMessage(hserver_connection,MSG_MoveCamera,&rec);
    return 0;
}

/* Pan/tilt camera to location */
int hclient_set_camera_tracker_mode(int mode){
    static int last_mode = 0;

    if (mode == last_mode) return 0;
    last_mode = mode;
    
    REC_SetCameraTrackerMode rec;

    rec.mode = mode;
    hc_communicator->SendMessage(hserver_connection,MSG_SetCameraTrackerMode,&rec);
    return 0;
}  

void hclientInitVariables(){
    int i;
    for (i=0;i<MAX_SENSORS;i++){
        sensor[i].type = -1;
        sensor_setup_link[i] = -1;
    }
}

/* Close robot connection */
int hclient_close_robot(void)
{
    hclient_send_simple_command(HCLIENT_CLOSE_ROBOT_CMD);
    return 0;
}

/* Return robot location */
int hclient_getxyt(double *x,double *y,double *z, double *yaw, double *roll, double *pitch){
    if (sensor[SENSOR_XYT].type != SENSOR_XYT) {
        fprintf(stderr,"Hclient Error: no xyt\n");
        return(-1);
    }
    if (sensor[SENSOR_XYT].num_readings != 6) {
        fprintf(stderr,"Hclient Error: wrong number of xyt readings\n");
        return(-1);
    }
    *x = sensor[SENSOR_XYT].reading[0];
    *y = sensor[SENSOR_XYT].reading[1];
    *z = sensor[SENSOR_XYT].reading[2];
    *yaw = sensor[SENSOR_XYT].reading[3];
    *roll = sensor[SENSOR_XYT].reading[4];
    *pitch = sensor[SENSOR_XYT].reading[5];

    return(0);
}

/* Return current battery readings */
int hclient_get_battery(int *level) {
    REC_SimpleCommand rec;
    REC_BatteryUpdate *current_battery;

    rec.command = HCLIENT_GET_BATTERY_READINGS_CMD;
    IPMessage *msg = hc_communicator->Query(hserver_connection,MSG_SimpleCommand,&rec,MSG_BatteryUpdate);
    current_battery = (REC_BatteryUpdate*)msg->FormattedData();

    *level = current_battery->battery_level;

    delete msg;
    return(0);
}

// Return current Jbox info
int hclient_get_jbox_info(info_jbox_list_t *jboxInfoList)
{
    info_jbox_t jboxInfo;
    REC_SimpleCommand rec;
    REC_JboxUpdate *current_jbox;
    int i, j, numNeighbors;
    const bool DUMP_READINGS = false;
                                                                                                                                                             
    rec.command = HCLIENT_GET_JBOX_READINGS_CMD;
    IPMessage *msg = hc_communicator->Query(
        hserver_connection,
        MSG_SimpleCommand,
        &rec,
        MSG_JboxUpdate);

    current_jbox = (REC_JboxUpdate*)msg->FormattedData();
                                                                                                                                                             
    numNeighbors = current_jbox->neighbor_num;

    for (i = 0; i < numNeighbors; i++)
    {
        jboxInfo.id = current_jbox->jbox_readingt[i].id;
        jboxInfo.numHops = current_jbox->jbox_readingt[i].num_hops;
        jboxInfo.signalStrength = current_jbox->jbox_readingt[i].signal_strength;

        memcpy(
            jboxInfo.gpsReading,
            current_jbox->jbox_readingt[i].gps_reading,
            sizeof(double)*GPS_NUM_ELEMENTS);

        jboxInfoList->push_back(jboxInfo);

        if (DUMP_READINGS)
        {
            printf(
                "\n%d. ID: %d, Hops: %d, Signal strength: %d\n",
                i,
                jboxInfo.id,
                jboxInfo.numHops,
                jboxInfo.signalStrength);

            printf("GPS:");
            for (j = 0; j < GPS_NUM_ELEMENTS; j++)
            {
                printf(" %f ", jboxInfo.gpsReading[j]);
            }
            printf("\n");
        }
    }

    delete msg;

    return (0);
}


/* Return current GPS position*/
int hclient_get_gps(double *lat,double *lon,double *dir){

    // Get data from sensor array, make sure it's valid
    if (sensor[SENSOR_GPS].type != SENSOR_GPS) {

        *lat = 0.0;
        *lon = 0.0;
        *dir = 0.0;
        return(0);
    }
    if (sensor[SENSOR_GPS].num_readings != NUM_GPS_READINGS) {
        fprintf(stderr,"Hclient Error: wrong number of gps readings\n");
        return(-1);
    }

    *lat = sensor[SENSOR_GPS].reading[0];
    *lon = sensor[SENSOR_GPS].reading[1];
    *dir = sensor[SENSOR_GPS].reading[3];
    
    // jh
    // return   < 0 -> error
    //          ==0 -> simulated value.`
    //          > 0 -> real gps
    return(1);
}

/* Return current RFID position*/
int hclient_get_rfid(float *rfidArray){

    int numRFIDs = 0, rfidIndex = -1, found = 0;
    double x=0.0, y=0.0, z=0.0;
 
    // Get data from sensor array, make sure it's valid
    if (sensor[SENSOR_RFID].type != SENSOR_RFID) {
	
        return(0);
    }

    if (rfidArray != NULL) {
      // Just copy the readings array (NOTE: caller must preallocate the array)
      // This function could also sort through the RFIDs, make a nice list of 
      // ones that have been found (stored in nice structures, not just a packed array),
      // or whatever else the robot executable needs
      memcpy(rfidArray, sensor[SENSOR_RFID].reading, sizeof(float)*sensor[SENSOR_RFID].num_readings);
    }

    // Iterate through the RFIDs just for fun!
    if (PRINT_READING) {
    numRFIDs = sensor[SENSOR_RFID].num_readings / 4;
    for (int i = 0; i < numRFIDs; i++) {
      rfidIndex = i*4;
      found = (int) sensor[SENSOR_RFID].reading[rfidIndex];
      x = sensor[SENSOR_RFID].reading[rfidIndex+1];
      y = sensor[SENSOR_RFID].reading[rfidIndex+2];
      z = sensor[SENSOR_RFID].reading[rfidIndex+3];
      if (found) printf("RFID %d: %s (%f,%f,%f)\n", i, found?"FOUND!":"NOT FOUND!", x, y, z);
    }
    printf("\n");
    }

    // return   < 0 -> error
    //          > 0 -> real values
    return(1);
}

/* Get sensor data array sizes to use when copying the arrays with hclient_get_XYZ(float*) functions below */
int hclient_enemy_size(){
    return sensor[SENSOR_ENEMY].num_readings;
}

int hclient_chemical_size(){
    return sensor[SENSOR_CHEMICAL].num_readings;
}

int hclient_comm_size(){
    return sensor[SENSOR_COMM].num_readings;
}

int hclient_get_num_enemy_detected() {
    int numenemies = 0;
    int total_readings = 0;
    int enemyIndex;
    int found = 0;

    if (sensor[SENSOR_ENEMY].type != SENSOR_ENEMY)
        return 0;
    numenemies = (sensor[SENSOR_ENEMY].num_readings - 1) / 7;
    for (int i = 0; i < numenemies; i++) {
        enemyIndex = i*7 + 1;
        found = (int) sensor[SENSOR_ENEMY].reading[enemyIndex];
        if (found) total_readings++;
    }
    return total_readings;
}

/* Return current enemy readings*/
int hclient_get_enemy(float *enemyArray){

    int numenemies = 0, enemyIndex = -1, found = 0;
    double x=0.0, y=0.0, z=0.0, dist=0.0, podE=0.0, podS=0.0;
 
    // Get data from sensor array, make sure it's valid
    if (sensor[SENSOR_ENEMY].type != SENSOR_ENEMY) {
	
        return(0);
    }

    if (enemyArray != NULL) {
      // Just copy the readings array (NOTE: caller must preallocate the array)
      // This function could also sort through the enemies, make a nice list of 
      // ones that have been found (stored in nice structures, not just a packed array),
      // or whatever else the robot executable needs
      memcpy(enemyArray, sensor[SENSOR_ENEMY].reading, sizeof(float)*sensor[SENSOR_ENEMY].num_readings);
    }

    // Iterate through the enemies just for fun!
    if (PRINT_READING) {
    numenemies = (sensor[SENSOR_ENEMY].num_readings - 1) / 7;
    for (int i = 0; i < numenemies; i++) {
      enemyIndex = i*7 + 1;
      found = (int) sensor[SENSOR_ENEMY].reading[enemyIndex];
      x = sensor[SENSOR_ENEMY].reading[enemyIndex+2];
      y = sensor[SENSOR_ENEMY].reading[enemyIndex+3];
      z = sensor[SENSOR_ENEMY].reading[enemyIndex+4];
      dist = sensor[SENSOR_ENEMY].reading[enemyIndex+1];
      podE = sensor[SENSOR_ENEMY].reading[enemyIndex+5];
      podS = sensor[SENSOR_ENEMY].reading[enemyIndex+6];

      //printf("%f %f %f %f %f %f %f", sensor[SENSOR_ENEMY].reading[enemyIndex], sensor[SENSOR_ENEMY].reading[enemyIndex+1], sensor[SENSOR_ENEMY].reading[enemyIndex+2], sensor[SENSOR_ENEMY].reading[enemyIndex+3], sensor[SENSOR_ENEMY].reading[enemyIndex+4], sensor[SENSOR_ENEMY].reading[enemyIndex+5], sensor[SENSOR_ENEMY].reading[enemyIndex+6]); 

      if (found) {

        printf("Enemy %d: %s (%f,%f,%f)\n", i, found?"FOUND!":"NOT FOUND!", x, y, z);
        printf("\tDistance: %f, PODSelf: %f, PODEnemy: %f\n", dist, podS, podE);
      }
    }
    printf("\n");
    }

    // return   < 0 -> error
    //          > 0 -> real values
    return(1);
}

int hclient_get_num_chemical_detected() {
    int numchemical = 0;
    int total_readings = 0;
    int chemicalIndex;
    int found = 0;
 
    // Check for valid chemical sensor
    if (sensor[SENSOR_CHEMICAL].type != SENSOR_CHEMICAL)
        return 0;

    numchemical = (sensor[SENSOR_CHEMICAL].num_readings - 1) / 6;
    for (int i = 0; i < numchemical; i++) {
        chemicalIndex = i*6 + 1;
        found = (int) sensor[SENSOR_CHEMICAL].reading[chemicalIndex];
        if (found) total_readings++;
    }
    return total_readings;
}

/* Return current chemical readings*/
int hclient_get_chemical(float *chemicalArray){
    int numchemicals = 0, chemicalIndex = -1, found = 0;
    double x=0.0, y=0.0, z=0.0, dist=0.0, podC=0.0;
 
    // Get data from sensor array, make sure it's valid
    if (sensor[SENSOR_CHEMICAL].type != SENSOR_CHEMICAL) {
	
        return(0);
    }

    if (chemicalArray != NULL) {
      // Just copy the readings array (NOTE: caller must preallocate the array)
      // This function could also sort through the chemicals, make a nice list of 
      // ones that have been found (stored in nice structures, not just a packed array),
      // or whatever else the robot executable needs
      memcpy(chemicalArray, sensor[SENSOR_CHEMICAL].reading, sizeof(float)*sensor[SENSOR_CHEMICAL].num_readings);
    }

    // Iterate through the chemicals just for fun!
    if (PRINT_READING) {
    numchemicals = (sensor[SENSOR_CHEMICAL].num_readings - 1) / 6;
    for (int i = 0; i < numchemicals; i++) {
      chemicalIndex = i*6 + 1;
      found = (int) sensor[SENSOR_CHEMICAL].reading[chemicalIndex];
      x = sensor[SENSOR_CHEMICAL].reading[chemicalIndex+2];
      y = sensor[SENSOR_CHEMICAL].reading[chemicalIndex+3];
      z = sensor[SENSOR_CHEMICAL].reading[chemicalIndex+4];
      dist = sensor[SENSOR_CHEMICAL].reading[chemicalIndex+1];
      podC = sensor[SENSOR_CHEMICAL].reading[chemicalIndex+5];
      if (found) {

        printf("Chemical %d: %s (%f,%f,%f)\n", i, found?"FOUND!":"NOT FOUND!", x, y, z);
        printf("\tDistance: %f, PODChemical: %f\n", dist, podC);
      }
    }
    printf("\n");
    }

    // return   < 0 -> error
    //          > 0 -> real values
    return(1);
}

/* Return current communication readings*/
int hclient_get_comm(float *commArray){

    int numcomms = 0, commIndex = -1, found = 0;
    double pd=0.0;
 
    // Get data from sensor array, make sure it's valid
    if (sensor[SENSOR_COMM].type != SENSOR_COMM) {
	
        return(0);
    }

    if (commArray != NULL) {
      // Just copy the readings array (NOTE: caller must preallocate the array)
      // This function could also sort through the comms, make a nice list of 
      // ones that have been found (stored in nice structures, not just a packed array),
      // or whatever else the robot executable needs
      memcpy(commArray, sensor[SENSOR_COMM].reading, sizeof(float)*sensor[SENSOR_COMM].num_readings);
    }

    // Iterate through the chemicals just for fun!
    if (PRINT_READING) {
    numcomms = (sensor[SENSOR_COMM].num_readings - 1) / 2;
    for (int i = 0; i < numcomms; i++) {
      commIndex = i*2 + 1;
      found = (int) sensor[SENSOR_COMM].reading[commIndex];
      pd = sensor[SENSOR_COMM].reading[commIndex+1];
      if (found) {
        printf("Bug Comm %d: %s, Pathloss: %f\n", i, found?"FOUND!":"NOT FOUND!", pd);
      }
    }
    printf("\n");
    }
    
    // return   < 0 -> error
    //          > 0 -> real values
    return(1);
}


/* Copy sensor packet into current sensors */
void hclientUpdateSensors(REC_SensorUpdate *cs){
    int i;
    int type;

    for(i=0;i<cs->num_sensors;i++){
        type = cs->sensor[i].type;
        if (sensor[type].type == -1) printf("error in update sensors %d\n", type);
        else {
            sensor[type].type = cs->sensor[i].type;
            sensor[type].len_extra = cs->sensor[i].len_extra;
            if (sensor[type].len_extra) {
                memcpy( sensor[type].extra, cs->sensor[i].extra, 
                        sensor[type].len_extra );
            }
            sensor[type].num_angles = cs->sensor[i].num_angles;
            sensor[type].pitch_angle = cs->sensor[i].pitch_angle;
            if (sensor[type].num_angles) {
                memcpy( sensor[type].angle, cs->sensor[i].angle, 
                        sizeof(float) * sensor[type].num_angles);
            }
            sensor[type].num_readings = cs->sensor[i].num_readings;
            if (sensor[type].num_readings) {
                memcpy( sensor[type].reading, cs->sensor[i].reading, 
                        sizeof(float) * sensor[type].num_readings );
            }
        }
    }
}

/* Return new sensor readings from hserver */
int hclient_get_sensor_readings(int mask,int *num,sensor_t **sensorPtr)
{
    REC_SimpleCommand rec;
    REC_SensorUpdate *current_sensors;

    rec.command = HCLIENT_GET_SENSOR_READINGS_CMD;
    IPMessage *msg = hc_communicator->Query(hserver_connection,MSG_SimpleCommand,&rec,MSG_SensorUpdate);
    current_sensors = (REC_SensorUpdate*)msg->FormattedData();
    hclientUpdateSensors(current_sensors);
    delete msg;
    *num = num_sensors;
    *sensorPtr = sensor; 
    return(0);
}

// ENDO
/*
// Initialize robot location. Will home nomads
int hclient_init_xyt(void)
{
    REC_SimpleCommand rec;

    rec.command = HCLIENT_INIT_XYT_CMD;
    IPMessage *msg = hc_communicator->Query(hserver_connection,MSG_SimpleCommand,&rec,MSG_SimpleCommand);
    delete msg;
    return(0);
}
*/  
// This function will initializes robot, and sets its current position to
// be the specified value.
int hclient_init_robot(double x, double y, double heading)
{
    REC_InitRobot rec;

    rec.x = x;
    rec.y = y;
    rec.heading = heading;

    // Use Query (instead of SendMessage) as to make sure that
    // the robot is initialized.
    IPMessage *msg = hc_communicator->Query(
        hserver_connection,
        MSG_InitRobot,
        &rec,
        MSG_SimpleCommand);

    delete msg;
    return(0);
}

/* Update sensor readings from hserver */
int hclient_update_sensors(){
    sensor_t *sensor;
    int num_sensors;
    hclient_get_sensor_readings(0xFFFF,&num_sensors,&sensor);
    return(0);
}

/* Open robot connection. Name is the hserver name */
int hclient_open_robot(char *name,bool multipleHservers)
{
    int rtn;

    fprintf(stderr,"Connecting to hserver %s\n",name);

    if (multipleHservers) {
        sprintf(hserver_ipt_name,"%s:control",name);
    } else {
        sprintf(hserver_ipt_name,"hserver_%s",name);
    }
    sprintf(hclient_ipt_name,"hclient_%s",name);
    hclientInitVariables();
    rtn = hclientSetupIpt();
    if (rtn) return(-1);
    hserver_connection = hc_communicator->Connect(hserver_ipt_name,IPT_REQUIRED);
    hclient_init_ipt_connection();
    hclient_get_robot_setup();
    hclient_update_sensors();
    hclient_send_simple_command(HCLIENT_RANGE_STOP_CMD);
    // ENDO
    //hclient_init_xyt();
    return 0;
}

/* Return obstacles from latest sensor reading */
int hclient_get_obstacles(
    int mask,
    float danger_range, 
    sensor_reading_t** reading,
    bool adjust)
{
    int i,j,idx,total_readings,rsidx;
    double x,y,z,roll,pitch,yaw;
    float *sp,st,sd;

    hclient_getxyt(&x,&y,&z,&yaw,&pitch,&roll);

    total_readings = 0;

    for(i=0;i<MAX_SENSORS;i++){
        switch(sensor[i].type){
        case SENSOR_SICKLMS200_1:
            if (!(mask&SENSOR_SICKLMS200_1_MASK)) break;
            total_readings += sensor[i].num_readings;
            break;
        case SENSOR_SICKLMS200_2:
            if (!(mask&SENSOR_SICKLMS200_2_MASK)) break;
            total_readings += sensor[i].num_readings;
            break;
        case SENSOR_SICKLMS200_3:
            if (!(mask&SENSOR_SICKLMS200_3_MASK)) break;
            total_readings += sensor[i].num_readings;
            break;
        case SENSOR_SICKLMS200_4:
            if (!(mask&SENSOR_SICKLMS200_4_MASK)) break;
            total_readings += sensor[i].num_readings;
            break;
         case SENSOR_COGNACHROME:
            if (!(mask&SENSOR_COGNACHROME_MASK)) break;
            total_readings += sensor[i].num_readings;
            break;
        case SENSOR_SONAR:
            if (!(mask&SENSOR_SONAR_MASK)) break;
            for (j = 0; j<sensor[i].num_readings; j++)
            {
                if (sensor[i].reading[j] != (float)SONAR_READING_OUTSIDE_RANGE) ++total_readings;
            }
            break;
        }
    }
    
    if (total_readings) {
        *reading = (sensor_reading_t*)malloc(sizeof(sensor_reading_t)*total_readings);
    } else {
        *reading = NULL;
    }

    idx = 0;
    for(i=0;i<MAX_SENSORS;i++){
        switch(sensor[i].type){
        case SENSOR_SICKLMS200_1:
        case SENSOR_SICKLMS200_2:
        case SENSOR_SICKLMS200_3:
        case SENSOR_SICKLMS200_4:
            if ((sensor[i].type == SENSOR_SICKLMS200_1) && !(mask&SENSOR_SICKLMS200_1_MASK)) break;
            if ((sensor[i].type == SENSOR_SICKLMS200_2) && !(mask&SENSOR_SICKLMS200_2_MASK)) break;
            if ((sensor[i].type == SENSOR_SICKLMS200_3) && !(mask&SENSOR_SICKLMS200_3_MASK)) break;
            if ((sensor[i].type == SENSOR_SICKLMS200_4) && !(mask&SENSOR_SICKLMS200_4_MASK)) break;
            
            for(j=0;j<sensor[i].num_readings;j++){
                (*reading)[idx].type = i;
                (*reading)[idx].angle = sensor[i].angle[j];
                (*reading)[idx].pitch_angle = sensor[i].pitch_angle;
                (*reading)[idx].reading = sensor[i].reading[j];
                if ((*reading)[idx].reading < danger_range) {
                    (*reading)[idx].color = SENSOR_DANGER_COLOR;
                } else { 
                    (*reading)[idx].color = SENSOR_SICKLMS200_COLOR;
                }
                idx++;
            }
            break;
        case SENSOR_COGNACHROME:
            if (!(mask&SENSOR_COGNACHROME_MASK)) break;
            extra_sensor_cog_t *ex_cog;
            ex_cog =  (extra_sensor_cog_t*)(sensor[SENSOR_COGNACHROME].extra);
            for(j=0;j<sensor[SENSOR_COGNACHROME].num_readings;j++){
                (*reading)[idx].type = SENSOR_COGNACHROME;
                (*reading)[idx].angle = sensor[SENSOR_COGNACHROME].angle[j];
                (*reading)[idx].reading = sensor[SENSOR_COGNACHROME].reading[j];
                if (j<(ex_cog->num_per_channel[0])) 
                    (*reading)[idx].color = 0;
                else if (j<(ex_cog->num_per_channel[0]+ex_cog->num_per_channel[1]))
                    (*reading)[idx].color = 1;
                else (*reading)[idx].color = 2;
                idx++; 
            }
            break;
        case SENSOR_SONAR:
            if (!(mask&SENSOR_SONAR_MASK)) break;
            rsidx = sensor_setup_link[SENSOR_SONAR];
            for(j=0;j<sensor[SENSOR_SONAR].num_readings;j++)
            {
                if (sensor[SENSOR_SONAR].reading[j] != (float)SONAR_READING_OUTSIDE_RANGE)
                {
                    // If "adjust" is true, the obstacle location will be adjusted for each time
                    // cycle by considering how much the robot moved. However, this part should
                    // be reviewed since some values are hard coded -- endo
                    //if (adjust)
                    if (adjust && sensor[SENSOR_SONAR].len_extra)
                    {
                        sp = (float*)(sensor[SENSOR_SONAR].extra);
                        st = *(sp+j*3+2);
                        sd = st-yaw;
                    }
                    else sd = 0;

                    (*reading)[idx].type = SENSOR_SONAR;
                    (*reading)[idx].angle = sensor[SENSOR_SONAR].angle[j]+sd;
                    (*reading)[idx].reading = sensor[SENSOR_SONAR].reading[j];
                    idx++;
                }
            }
            break;
        }
    }
    return(total_readings);
}

/* Return array of latest laser reading */
int hclient_laser_read(float *reading){
    int i;
    int num_readings;

    if (sensor[SENSOR_SICKLMS200_1].type != SENSOR_SICKLMS200_1) {
        fprintf(stderr,"No laser sensor");
        return(-1);
    }

    num_readings = sensor[SENSOR_SICKLMS200_1].num_readings;
    for(i=0;i<sensor[SENSOR_SICKLMS200_1].num_readings;i++){
        reading[i] = sensor[SENSOR_SICKLMS200_1].reading[i];
    }
    return(num_readings);
}

/**********************************************************************
# $Log: hclient.c,v $
# Revision 1.13  2010-05-03 22:08:34  zkira
# Cricket sensor and blob tracking
#
# Revision 1.12  2009/09/29 17:00:14  zkira
# Blob detection stuff
#
# Revision 1.11  2009/04/13 22:23:29  zkira
# *** empty log message ***
#
# Revision 1.9  2009/03/02 18:45:13  zkira
# Added 3D pose stuff (for 3D logging)
#
# Revision 1.8  2008/11/20 01:46:29  ts2883
# Finished adding the pitch information for the laser.
#
# Revision 1.7  2008/11/05 22:14:17  zkira
# In the process of adding pitch information for laser
#
# Revision 1.6  2008/11/04 04:53:55  zkira
# Added pitch for 3D laser readings
#
# Revision 1.5  2008/10/30 22:50:48  zkira
# Added support for up to 4 laser sensors
#
# Revision 1.4  2008/10/27 21:22:43  zkira
# Minor change: Removed print
#
# Revision 1.3  2008/09/09 20:06:10  zkira
# Crude altitude control: Changed hclient_getxyt & steer_toward_and_drive functions to include altitude (z)
#
# Revision 1.2  2008/07/15 22:38:33  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.1.1.1  2008/07/14 16:44:18  endo
# MAST Project (based on MissionLab-MINOS-20071018.tar.gz)
#
# Revision 1.1.1.1  2006/07/20 17:17:47  endo
# MINOS Project (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:55  endo
# MissionLab 7.0
#
# Revision 1.2  2006/06/06 00:55:59  endo
# PoseCalculator from MARS 2020 migrated
#
# Revision 1.4  2004/09/10 19:30:57  endo
# New PoseCalc integrated.
#
# Revision 1.3  2004/04/17 12:21:47  endo
# Intalling JBox Sensor Update.
#
# Revision 1.2  2004/04/13 19:36:53  yangchen
# JBox feature added.
#
# Revision 1.1.1.1  2004/02/27 22:30:50  endo
# Initial MARS 2020 revision
#
# Revision 1.13  2003/06/19 21:18:53  endo
# *** empty log message ***
#
# Revision 1.12  2003/06/19 20:51:31  endo
# *** empty log message ***
#
# Revision 1.11  2003/04/06 13:20:20  endo
# gcc 3.1.1
#
# Revision 1.10  2003/04/02 22:14:17  zkira
# GPS and battery information for telop GUI
#
# Revision 1.9  2000/12/12 22:45:09  blee
# Changed the case for SENSOR_SICKLMS200 in hclient_get_robot_setup()
# to include angle counts, and changed the case for SENSOR_SICKLMS200
# in hclient_get_obstacles() to get the angles.
#
# Revision 1.8  2000/11/13 21:07:45  endo
# *** empty log message ***
#
# Revision 1.7  2000/11/13 20:16:04  endo
# If the sonar reading is outside its range, it will be filtered.
#
# Revision 1.6  2000/09/19 03:52:33  endo
# Sonar bug fixed.
#
#
#**********************************************************************/
