/**********************************************************************
 **                                                                  **
 **                        ContrilInterfaces.c                       **
 **                                                                  **
 **                                                                  **
 **  Written by:  Brian Lee                                          **
 **                                                                  **
 **  Copyright 2002 - 2008, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: ControlInterfaces.c,v 1.4 2009-03-02 18:52:09 zkira Exp $ */

#include <string>

#include "ControlInterfaces.h"
#include "message.h"
#include "fred.h"
#include "nomad.h"
#include "pioneer.h"
#include "pioneer2.h"
#include "amigobot.h"
#include "compass.h"
#include "compass_3DM-G.h"
#include "gyro.h"
#include "video.h"
#include "frame_grabber.h"
#ifndef NO_GAZEBO
#include "gazebo-robot.h"
#include "gazebo-video.h"
#endif
#include "usarsim-robot.h"
#include "usarsim-video.h"

#ifndef NO_MOBILITY
#include "urban.h"
#include "atrvjr.h"
#include "compass_kvh-c100.h"
#include "gyro_dmu-vgx.h"
#endif

#include "PoseCalculatorInterface.h"

using std::string;

void robotConnect(Robot*& robot)
{
    char c, robot_type;
    string strHost;
    string strPortString;
    int poseCalcID;
    const unsigned int uPIONEER_PORT_ECHO_LINE = 4;
    const unsigned int uPIONEER2_PORT_ECHO_LINE = 4;
    const unsigned int uAMIGOBOT_PORT_ECHO_LINE = 4;

    robot_type = messageGetChar(
        EnMessageFilterType_ROBOT_TYPE, 
        EnMessageType_ROBOT_TYPE,
        EnMessageErrType_ROBOT_TYPE);

    if (robot_type != 'x')
    {
        switch(robot_type) {

        case '1': // Emulated Robot (fred)
            new Fred(&robot);
            break;

        case '2': // Pioneer AT
            strPortString = messageGetString(EnMessageType_PIONEER_PORT, uPIONEER_PORT_ECHO_LINE);
            new Pioneer(&robot, strPortString);
            break;

	    case '3': // Pioneer2 DXE
            strPortString = messageGetString(EnMessageType_PIONEER2_PORT, uPIONEER2_PORT_ECHO_LINE);
            new Pioneer2(&robot, strPortString);
            break;

        case '4': // AmigoBot
            strPortString = messageGetString(EnMessageType_AMIGOBOT_PORT, uAMIGOBOT_PORT_ECHO_LINE);
            new AmigoBot(&robot, strPortString);
            break;

        case '5': // Nomad 150/200
            c = messageGetChar(EnMessageFilterType_NOMAD, EnMessageType_NOMAD, 
                               EnMessageErrType_NOMAD_TYPE);
            if (c != 'x')
            {
                const unsigned int uNOMAD_150_PORT_ECHO_LINE = 4;
                const unsigned int uNOMAD_200_HOST_NAME_ECHO_LINE = 3;

                Nomad::EnNomadType enType;

                switch(c - '0')
                {
                case 1:
                    enType = Nomad::EnNomadType_150;
                    strPortString = messageGetString(EnMessageType_NOMAD_150_PORT, 
                                                     uNOMAD_150_PORT_ECHO_LINE);
                    strHost = "";
                    break;
                case 2:
                    enType = Nomad::EnNomadType_200;
                    strPortString = "";
                    strHost = messageGetString(EnMessageType_NOMAD_200_HOST_NAME, 
                                               uNOMAD_200_HOST_NAME_ECHO_LINE);
                    break;
                default:
                    printTextWindow("Bad Nomad type in robotConnect()");
                    return;
                }
		        new Nomad(&robot, enType, strPortString, strHost);
            }
            break;

        case '6': // Urban Robot
#ifndef NO_MOBILITY	
            new Urban(&robot);
#else 
            printTextWindow("No Mobility support");
#endif
            break;

        case '7': // ATRV-Jr Robot
#ifndef NO_MOBILITY	
	        new ATRVJr(&robot);
#else 
            printTextWindow("No Mobility support");
#endif
            break;

        case '8': // Gazebo - Use default robot name defined in gazebo-robot.c.
#ifndef NO_GAZEBO
	  new GazeboRobot(&robot, 
			  GazeboRobot::DEFAULT_GAZEBO_ROBOT_NAME_.c_str(),
			  GazeboRobot::DEFAULT_GAZEBO_ROBOT_CLASS_.c_str(),
			  GazeboRobot::DEFAULT_GAZEBO_SERVER_ID_,
			  GazeboRobot::DEFAULT_GAZEBO_CLIENT_ID_);
#else
            printTextWindow("Gazebo disabled.");
#endif
            break;

/*
        case '9': 
                new USARSimRobot(
                    &gRobot,
		    USARSimRobot::DEFAULT_USARSIM_SERVER_HOST_MACHINE_.c_str(),
		    USARSimRobot::DEFAULT_USARSIM_ROBOT_NAME_.c_str(),
		    USARSimRobot::DEFAULT_USARSIM_ROBOT_CLASS_.c_str(),
		    USARSimRobot::DEFAULT_USARSIM_SERVER_PORT_.c_str(),
		    0.0, 0.0, 0.0, 0.0, 0.0, 0.0 ); // Default 3D init pose (TODO: Make constants in USARSimRobot::
	        break;
*/
        }

        if (robot != NULL)
        {
            poseCalcID = gPoseCalc->addRobot(robot);
            robot->savePoseCalcID(poseCalcID);
        }
    }
}

void videoConnect(Video*& video)
{
    string portString;
    char videoType;
    unsigned const int uPORT_ECHO_LINE = 4;
    const bool SHOW_IMAGE = true;

    videoType = messageGetChar(
        EnMessageFilterType_VIDEO_TYPE, 
        EnMessageType_VIDEO_TYPE,
        EnMessageErrType_VIDEO_TYPE);

    if (videoType != 'x')
    {
        switch (videoType) {

        case '1': // Frame Grabber
            // Get the connection port from the user
            portString = messageGetString(
                EnMessageType_FRAMEGRABBER_PORT,
                uPORT_ECHO_LINE);
            new FrameGrabber(&video, portString, SHOW_IMAGE);
            break;

        case '2': // Gazebo
#ifndef NO_GAZEBO
            new GazeboVideo(
                &video,
                NULL,
                NULL,
                SHOW_IMAGE);
#else
            printTextWindow("Gazebo disabled.");
#endif
            break;
        }
    }
}

void compassConnect(Compass*& compass)
{
    char compassType;
    string portString;
    const unsigned int uCOMPASS_3DMG_PORT_ECHO_LINE = 4;
    int poseCalcID;

    compassType = messageGetChar(
        EnMessageFilterType_COMPASS_TYPE, 
        EnMessageType_COMPASS_TYPE,
        EnMessageErrType_COMPASS_TYPE);

    if (compassType != 'x')
    {
        switch(compassType) {

        case '1': // KVH-C100
#ifndef NO_MOBILITY	
            new CompassKVHC100(&compass);
#else 
            printTextWindow("No Mobility support");
#endif
            break;

        case '2': // 3DM-G
            portString = messageGetString(
                EnMessageType_COMPASS_3DMG_PORT,
                uCOMPASS_3DMG_PORT_ECHO_LINE);
            new Compass3DMG(&compass, portString);
            break;
        }

        if (compass != NULL)
        {
            poseCalcID = gPoseCalc->addCompass(compass);
            compass->savePoseCalcID(poseCalcID);
        }
    }
}

void gyroConnect(Gyro*& gyro)
{
    char gyroType;
    int poseCalcID;

    gyroType = messageGetChar(
        EnMessageFilterType_GYRO_TYPE, 
        EnMessageType_GYRO_TYPE,
        EnMessageErrType_GYRO_TYPE);

    if (gyroType != 'x')
    {
        switch(gyroType) {

        case '1': // DMU-VGX
#ifndef NO_MOBILITY	
            new GyroDMUVGX(&gyro);
#else 
            printTextWindow("No Mobility support");
#endif
            break;

        default:
            break;
        }

        if (gyro != NULL)
        {
            poseCalcID = gPoseCalc->addGyro(gyro);
            gyro->savePoseCalcID(poseCalcID);
        }
    }
}

void gpsConnect(
    Gps*& gps,
    double base_lat,
    double base_lon,
    double base_x,
    double base_y, 
    double mperlat,
    double mperlon)
{
    int c, poseCalcID;
    bool base = false;
    unsigned const int uPORT_ECHO_LINE = 4;

    printfTextWindow("\n");
    printfTextWindow("GPS can only be started through the command line option.\n");
    printfTextWindow("Restart hserver with '-g <gps_section in .hserverrc>'.\n");

    return;

    // Note: Below is an obsolete code.

    c = messageGetChar(EnMessageFilterType_GPS, EnMessageType_GPS_SELECT, EnMessageErrType_GPS);

    if (c != 'x')
    {
        base = (c == '0');

        // get the connection port from the user
        string strPort = messageGetString(EnMessageType_GPS_PORT, uPORT_ECHO_LINE);

        new Gps(&gps, strPort, base_lat, base_lon, base_x, base_y, mperlat, mperlon, base);

        if (gps != NULL)
        {
            poseCalcID = gPoseCalc->addGps(gps);
            gps->savePoseCalcID(poseCalcID);
        }
    }
}

void controlGps(Gps* gps)
{
    bool bDone = false;
    int msgData[4];

    msgData[0] = gps->GetSavingPackets();
    msgData[1] = (int) gps->GetBaseX();
    msgData[2] = (int) gps->GetBaseY();

    messageDrawWindow(EnMessageType_GPS_CONTROL, EnMessageErrType_NONE, msgData);
    do
    {
        int c = getch();
        switch(c)
        {
        case 'd':
            messageHide();
            delete gps;
            return;
            break;

        case 's':
            gps->SetSavePackets(!gps->GetSavingPackets());
            msgData[0] = gps->GetSavingPackets();
            messageDrawWindow(EnMessageType_GPS_CONTROL, EnMessageErrType_NONE, msgData);
            break;
        case 'q':
            gps->AlterBaseYBy(1.0);
            msgData[2] = (int) gps->GetBaseY();
            messageDrawWindow(EnMessageType_GPS_CONTROL, EnMessageErrType_NONE, msgData);
            break; 
        case 'a':
            gps->AlterBaseYBy(-1.0);
            msgData[2] = (int) gps->GetBaseY();
            messageDrawWindow(EnMessageType_GPS_CONTROL, EnMessageErrType_NONE, msgData);
            break; 
        case 'w':
            gps->AlterBaseXBy(-1.0);
            msgData[1] = (int) gps->GetBaseX();
            messageDrawWindow(EnMessageType_GPS_CONTROL, EnMessageErrType_NONE, msgData);
            break; 
        case 'e':
            gps->AlterBaseXBy(1.0);
            msgData[1] = (int) gps->GetBaseX();
            messageDrawWindow(EnMessageType_GPS_CONTROL, EnMessageErrType_NONE, msgData);
            break; 
            gps->information();
            break;
        case 'x':
        case 'Q':
        case KEY_ESC:
            bDone = true;
            break;
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;
        }
    } while (!bDone);
    messageHide();
}

void controlJbox(Jbox* jbox)
{
    int c;
    int i,j;                                                                                                                                                             
    bool bDone = false;
    const unsigned int JBOX_ECHO_LINE = 4;
    int j_pos[7];
    double j_gps[7];
    
    string strJboxString;
    messageDrawWindow(EnMessageType_JBOX_CONTROL, EnMessageErrType_NONE);
    redrawWindows();
                                                                                                                                                             
    do
    {
        c = getch();
        switch(c)
        {
            case 'd':
                messageHide();
                delete jbox;
                //useJbox = false;
                bDone = true;
                break;
            
            case 'q':
                messageHide();
                messageDrawWindow (EnMessageType_JBOX_SELECT, EnMessageErrType_NONE);
                redrawWindows();
                strJboxString = messageGetString(EnMessageType_JBOX_SELECT, JBOX_ECHO_LINE);
                i=atoi(strJboxString.c_str());
                messageHide();
                                                                                                                                                             
                if (jbox != NULL)
                {
                    j = jbox->quality_to_neighbor(i);
                                                                                                                                                             
                    if (j == -1)
                    {
                        messageDrawWindow (EnMessageType_NONE, EnMessageErrType_JBOX_NOTFOUND);
                    }
                    else
                    {
                        messageDrawWindow (EnMessageType_JBOX_DISPLAY_NSS, EnMessageErrType_NONE, &j);
                    }
                }
                else
                {
                    messageDrawWindow (EnMessageType_NONE, EnMessageErrType_JBOX_NOTCONNECT);
                }
                                                                                                                                                             
                redrawWindows();
                                                                                                                                                             
                break;

            case 'h':
                messageHide();
                messageDrawWindow (EnMessageType_JBOX_SELECT, EnMessageErrType_NONE);
                redrawWindows();
                strJboxString = messageGetString(EnMessageType_JBOX_SELECT, JBOX_ECHO_LINE);
                i=atoi(strJboxString.c_str());
                messageHide();
                if (jbox != NULL)
                {
                    j = jbox->HOPS_of_node(i);
                    if (j == -1)
                    {
                        messageDrawWindow (EnMessageType_NONE, EnMessageErrType_JBOX_NOTFOUND);
                    }
                    else
                    {
                        messageDrawWindow (EnMessageType_JBOX_DISPLAY_HOPS, EnMessageErrType_NONE, &j);
                    }
                }
                else
                {
                    messageDrawWindow (EnMessageType_NONE, EnMessageErrType_JBOX_NOTCONNECT);
                }
                redrawWindows();
                break;

            case 'p':
                messageHide();
                messageDrawWindow (EnMessageType_JBOX_SELECT, EnMessageErrType_NONE);
                redrawWindows();
                strJboxString = messageGetString(EnMessageType_JBOX_SELECT, JBOX_ECHO_LINE);
                i=atoi(strJboxString.c_str());
                messageHide();
                                                                                                                                                             
                if (jbox != NULL)
                {
                    j = jbox->POS_of_node(j_pos, i);

                    if (j == -1)
                    {
                        messageDrawWindow (EnMessageType_NONE, EnMessageErrType_JBOX_NOTFOUND);
                    }
                    else
                    {
                        messageDrawWindow (EnMessageType_JBOX_DISPLAY_POS, EnMessageErrType_NONE, j_pos);
                    }
                }
                else
                {
                    messageDrawWindow (EnMessageType_NONE, EnMessageErrType_JBOX_NOTCONNECT);
                }
                                                                                                                                                             
                redrawWindows();
                                                                                                                                                             
                break;
            
            case 'g':
                messageHide();
                messageDrawWindow (EnMessageType_JBOX_SELECT, EnMessageErrType_NONE);
                redrawWindows();
                strJboxString = messageGetString(EnMessageType_JBOX_SELECT, JBOX_ECHO_LINE);
                i=atoi(strJboxString.c_str());
                messageHide();
                                                                                                                                                             
                if (jbox != NULL)
                {
                    j = jbox->GPS_of_node(j_gps, i);
                                                                                                                                                             
                    if (j == -1)
                    {
                        messageDrawWindow (EnMessageType_NONE, EnMessageErrType_JBOX_NOTFOUND);
                    }
                    else
                    {
                        messageDrawWindow (EnMessageType_JBOX_DISPLAY_GPS, EnMessageErrType_NONE, j_gps);
                    }
                }
                else
                {
                    messageDrawWindow (EnMessageType_NONE, EnMessageErrType_JBOX_NOTCONNECT);
                }
                                                                                                                                                             
                redrawWindows();
                break;
            
            case 'x':
            case 'Q':
            case KEY_ESC:
                bDone = true;
                break;
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
            case KEY_UP:
            case KEY_DOWN:
            case KEY_LEFT:
            case KEY_RIGHT:
                messageMovePanelKey(c);
                break;
        }
    } while (!bDone);
                                                                                                                                                             
    messageHide();
}

/**********************************************************************
# $Log: ControlInterfaces.c,v $
# Revision 1.4  2009-03-02 18:52:09  zkira
# Removed USARSim from HServer menu because it is not supported yet
#
# Revision 1.3  2008/09/18 04:15:44  ts2883
# Fixed gazebo code to compile/run
# Still does not work with scarab, but works with multiple robots/hservers otherwise
#
# Revision 1.2  2008/07/15 22:33:46  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.5  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.4  2006/11/15 23:45:58  endo
# GazeboLaser can now accept the second laser.
#
# Revision 1.3  2006/07/31 22:01:14  endo
# Gazebo -> GazeboRobot
#
# Revision 1.2  2006/07/21 23:35:01  endo
# Gazebo integrated into HServer
#
# Revision 1.1.1.1  2006/07/21 17:42:08  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.2  2006/06/05 21:35:51  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.6  2004/09/22 03:59:34  endo
# *** empty log message ***
#
# Revision 1.5  2004/09/10 19:41:06  endo
# New PoseCalc integrated.
#
# Revision 1.4  2004/04/24 11:20:53  endo
# 3DM-G added.
#
# Revision 1.3  2004/04/17 12:20:34  endo
# Intalling JBox Sensor Update.
#
# Revision 1.2  2004/04/13 19:25:33  yangchen
# JBox module integrated.
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.5  2003/06/19 20:52:45  endo
# *** empty log message ***
#
# Revision 1.4  2003/04/26 06:48:31  ebeowulf
# Added the Pioneer2-dxe controller.
#
# Revision 1.3  2002/02/18 13:48:21  endo
# AmigoBot added.
#
#
#**********************************************************************/
