/**********************************************************************
 **                                                                  **
 **                        RComDefaultSetups.h                       **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  The some global variables that are used by RCom.                **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef RComDefaultSetups_H
#define RComDefaultSetups_H

/* $Id: RComDefaultSetups.h,v 1.1 2008-07-15 22:33:57 endo Exp $ */

#include <string>
#include <vector>

using std::string;
using std::vector;

//---------------------------------------------------------------------
// RCom-to-RCom connection related
//---------------------------------------------------------------------
// The size of a char-array that stores an IP number.
#define IP_ARRAY_SIZE 16

// The number of paths that RCom can take to reach the other RCom.
#define RCOM_NUM_PATHS 3

// The maximum number of applications that can be using reliable communication
// simultaneously
#define RCOM_MAX_APPS 15

// The maximum number of payload (data) size that RCom can handle.
#define RCOM_MAX_PAYLOAD_SIZE 200000

// IP addresses of the other RCom (peer). 
const char DEFAULT_PEER_RCOM_IP[RCOM_NUM_PATHS][IP_ARRAY_SIZE] = {
  //"199.77.206.72", "199.77.206.72", "199.77.206.72" // Robot1 eth0 ---- do not use any more please(unless wireless is not working)
  //"199.77.206.249", "199.77.206.249", "199.77.206.249" // Robot2 eth0 ---- do not use any more please(unless wireless is not working)
  //"130.207.109.52", "130.207.109.52", "130.207.109.52" // OCS eth1 ---- do not use any more please(unless wireless is not working)
  //"192.168.1.60","192.168.2.60","192.168.3.60" //OCS eth0 (local/wireless)
  //"192.168.1.70","192.168.2.70","192.168.3.70" //Robot  eth1 (local/wireless)
  //"192.168.1.71","192.168.2.71","192.168.3.71" //Robot2 eth1 (local/wireless)
  "127.0.0.1","127.0.0.1","127.0.0.1" // Localhost
};

// Port number that RCom uses to connect the other RCom.
const int DEFAULT_INTER_RCOM_PORT = 6000; // OCS <--> Robot1 connection
//const int DEFAULT_INTER_RCOM_PORT = 6100; // OCS <--> Robot2 connection


//---------------------------------------------------------------------
// RComServer-to-RComClient connection related
//---------------------------------------------------------------------
// The name or IP of the machine rcomserver will be running on.
const string DEFAULT_RCOM_SERVER_HOST_MACHINE = "localhost";

// Port number that clients can use to talk to rcomserver.
const int DEFAULT_RCOM_SERVER_PORT = 7000; // OCS <--> Robot1 connection
//const int DEFAULT_RCOM_SERVER_PORT = 7001; // OCS <--> Robot2 connection

#endif
/**********************************************************************
 * $Log: RComDefaultSetups.h,v $
 * Revision 1.1  2008-07-15 22:33:57  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.2  2006/08/09 21:35:20  endo
 * DEFAULT_PEER_RCOM_IP changed to use localhost.
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.12  2006/04/20 17:25:14  endo
 * HServerController fixed.
 *
 * Revision 1.11  2006/04/19 17:12:42  endo
 * *** empty log message ***
 *
 * Revision 1.10  2006/04/13 14:37:36  endo
 * Connection list cleaned up.
 *
 * Revision 1.9  2006/04/13 07:31:53  endo
 * DP-SLAM implemented
 *
 * Revision 1.8  2006/04/06 18:42:06  envarli
 * Labeled all the IP's for different devices, just uncomment the one you want to use. Note: There is no wireless bridge for the second link so don't freak out if it says line 2 is down
 *
 * Revision 1.7  2006/04/05 18:01:55  envarli
 * *** empty log message ***
 *
 * Revision 1.6  2006/03/31 21:01:47  endo
 * *** empty log message ***
 *
 * Revision 1.5  2006/03/31 20:58:23  endo
 * *** empty log message ***
 *
 * Revision 1.4  2006/03/08 22:51:28  gtg979f
 * *** empty log message ***
 *
 * Revision 1.3  2006/03/08 01:19:41  jinhlee
 * original directories for image compress
 *
 * Revision 1.2  2006/02/27 20:41:54  endo
 * rcomserver bug fixed.
 *
 * Revision 1.1  2006/02/25 22:10:22  endo
 * The user can now specify ports to be used.
 *
 **********************************************************************/
