/**********************************************************************
 **                                                                  **
 **                      RComm_Controller.h                          **
 **                                                                  **
 **  Written by:  I.C. Envarli                                       **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
//Usage: Declare a Comm_Thread_Controller variable,
//       call .init with the ipnumbers and the port number
//To use communication functions, use variablename.r_comm.x
//where x is any of the public members in the Reliable_Comm class

#ifndef __RComm_Controller_h
#define __RComm_Controller_h
#include "Reliable_Comm.h"
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

typedef struct comm_thread_info{
	int pathnumber;
	Reliable_Comm * r_pointer;
} comm_thread_info_t;

class RComm_Controller {
   public:
       RComm_Controller(void);
	   bool init(char ipnumbers[RCOM_NUM_PATHS][IP_ARRAY_SIZE], int port);

       //Functions that simply call the corresponding function of the r_comm object:
  	   void Send_Packet_Reliable(byte payload[], const unsigned int payload_len, const unsigned int recipient_id);
	   void Send_Packet_Unreliable(byte payload[],const unsigned int payload_len, const unsigned int recipient_id);
	   void Print_Stats();
       void Unregister_App(int app_id);
       bool Register_App(Receiver_Callback_p callback_func, int app_id,byte * local_buffer, void *app_instance);
       bool initialized(void);
       const char *Get_Stats();
       void Modify_Inter_Packet_Time(int path,int newspeed_usec);
   
   private:
	   Reliable_Comm r_comm;
	   pthread_t recv_threads[RCOM_NUM_PATHS];
	   pthread_t heart_beat_thread;
	   pthread_t thruput_thread;
       bool initialized_;
};

inline bool RComm_Controller::initialized(void)
{
    return initialized_;
}

extern RComm_Controller *gRComm_Controller;

#endif //__RComm_Controller_h

/**********************************************************************
 * $Log: RComm_Controller.h,v $
 * Revision 1.1  2008-07-15 22:33:58  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.2  2006/08/09 21:46:53  endo
 * -ansi -pedantic added to Makefile
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.7  2006/04/14 11:24:35  endo
 * RComInterthread::getRComStatus() added.
 *
 * Revision 1.6  2006/04/14 04:39:35  envarli
 * *** empty log message ***
 *
 * Revision 1.5  2006/04/08 17:01:30  envarli
 * Added a new thread for thruput calculations, not currently used but will be soon
 *
 * Revision 1.4  2006/04/02 22:46:29  endo
 * Unregister_App() added to Reliable_Comm.
 *
 * Revision 1.3  2006/02/27 20:41:54  endo
 * rcomserver bug fixed.
 *
 * Revision 1.2  2006/02/25 22:10:22  endo
 * The user can now specify ports to be used.
 *
 * Revision 1.1  2006/02/24 11:50:33  endo
 * RComm_Controller checked in.
 *
 **********************************************************************/
