/*  

    rovio.cpp
    Copyright (C) 2009 Georgia Institute of Technology

    Author
    John Rogers
    jgrogers@cc.gatech.edu


    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/




#include "rovio.h"
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 
#include <unistd.h>

#ifdef NO_FREEIMAGE
#else 
#include <FreeImage.h>
#endif

#include <iostream>

#ifndef NO_OPENCV
#include <highgui.h>
#include <cv.h>
#endif

#include "../robot.h"

bool donerotate = true;
int rotatetimeout = 10;
int RTO_VAL = 4;

using namespace std;
Rovio::
Rovio(const char* ip_addr) {

#ifndef NO_OPENCV
  m_pLastImage = NULL;
#endif

  X = Y = Theta = 0.0;
  dX = dY = dTheta = 0.0;

  host = strdup(ip_addr);
  errorBuffer = new char[CURL_ERROR_SIZE];
  pthread_mutex_init(&velocity_mutex, NULL);
  pthread_mutex_init(&curl_mutex, NULL);
  // Create our curl handle  
  curl_setup(general_curl);
  curl_setup(mutexed_curl);

  char urlstr [100];

  //Raise the neck
  //sprintf(urlstr,"%s/rev.cgi?Cmd=nav&action=18&drive=13&s_value=5",host);
  //libcurl_sendcommand_noresp(urlstr);


  // we want the resolution set on the rovio
  //Set image resolution
  //  sprintf(urlstr,"%s/ChangeResolution.cgi?ResType=3",host);
  //  libcurl_sendcommand_noresp(urlstr);

  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  
  pthread_create(&mySubThread, &attr, 
		 thunk<Rovio, &Rovio::PeriodicUpdate>, 
		 this);



  // Initialize variables for grabbing camera images
  CurFrame.DataSize = 0;
  CurFrame.Data = NULL;
  NextFrameTime = 0;

}

void Rovio::
sockerror(const char *msg)
{
    perror(msg);
    exit(0);
}
static int writer(char *data, size_t size, size_t nmemb,  
                  std::string *buffer)  
{  
  // What we will return  
  int result = 0;  
  
  // Is there anything in the buffer?  
  if (buffer != NULL)  
  {  
    // Append the data to the buffer  
    buffer->append(data, size * nmemb);  
  
    // How much did we write?  
    result = size * nmemb;  
  }  
  //  printf("Wrote %d bytes\n", result);
  return result;  
}  

#ifndef NO_OPENCV
IplImage* Rovio::
PrepareImage(const IplImage* imgin, 
	     IplImage* imgout) {

  cvConvertImage(imgin, imgout);
  return imgout;
}
#endif

void Rovio::
curl_setup(CURL*& curl) {
  curl = curl_easy_init();  
  
  if (curl) {  
    // Now set up all of the curl options  
    curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, errorBuffer);  
    curl_easy_setopt(curl, CURLOPT_HEADER, 0);  
    curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1);  
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writer);  
  }else {
    printf("CURL initialization failed\n");
    exit(0);
  }
  
}

void Rovio::
PeriodicUpdate() {
  while (true) {
    //    cout <<"Called periodic update!\n";
    
    libcurl_getenc(host);
    usleep(100);

  }
}
Rovio::~Rovio() {
  // Always clean up  
  curl_easy_cleanup(general_curl);  
  curl_easy_cleanup(mutexed_curl);
  pthread_cancel(mySubThread);
  delete errorBuffer;
}

CURLcode Rovio::
libcurl_sendcommand(CURL* curl, 
		    char* urlstr, 
		    string& buffer) {
  string url(urlstr);  
  //  cout << "Retrieving " << url << endl;  
  
  curl_easy_setopt(curl, CURLOPT_URL, urlstr);  

  CURLcode result = CURLE_FAILED_INIT;  
  
  curl_easy_setopt(curl, CURLOPT_URL, urlstr);  
  curl_easy_setopt(curl, CURLOPT_WRITEDATA, &buffer);  
    
  // Attempt to retrieve the remote page  
  result = curl_easy_perform(curl);  
    
  return result;
}		    

bool Rovio::getIR() {
   string buffer;
  char urlstr[100];
  sprintf(urlstr,"%s/rev.cgi?Cmd=nav&action=1", host);
  CURLcode result = libcurl_sendcommand(general_curl,urlstr, buffer);
  // Did we succeed?
  if (result == CURLE_OK)
    {
      char * offset = strstr(buffer.c_str(), "flags=");
      int flag = 0;
      int toknum = sscanf(offset, "flags=%d", &flag);
      return (flag & 2);
    }
  else
    {
      cout << "Error: [" << result << "] - " << errorBuffer<<endl;
      exit(0);
    }

}

void Rovio::
libcurl_getstate(char* base) {
  string buffer;
  char urlstr[100];
  sprintf(urlstr,"%s/rev.cgi?Cmd=nav&action=1", base);
  CURLcode result = libcurl_sendcommand(general_curl,urlstr, buffer);
  // Did we succeed?  
  if (result == CURLE_OK)  
    {  
      char * offset = strstr(buffer.c_str(), "responses = ");
      unsigned int stat;
      int x;
      int y;
      float theta;
      int toknum = sscanf(offset, "responses = %d|x=%d|y=%d|theta=%f|%*s", &stat, &x, &y, &theta);
      printf("X: %d, Y: %d, theta: %f\n",
	     x,y,theta);
    }  
  else  
    {  
      cout << "Error: [" << result << "] - " << errorBuffer<<endl;  
      exit(0);
    }  
}

unsigned int char_to_hd(unsigned char in) {
  if (in >= '0' && in <= '9') return (unsigned int)( in - '0');
  else if (in >= 'A' && in <= 'F') return (unsigned int) (in - 'A' + 10);
  else return 0;
}

unsigned int get_byte(unsigned char* coded_resp, unsigned int& offset) {
  unsigned int n1 = char_to_hd(coded_resp[offset]);
  unsigned int n2 = char_to_hd(coded_resp[offset+1]);
  unsigned int result = (n1<<4) + n2;
  offset+=2;
  return result;
}

unsigned int get_double_byte(unsigned char* coded_resp, unsigned int& offset) {
  unsigned int n1 = get_byte(coded_resp, offset);
  unsigned int n2 = get_byte(coded_resp, offset);
  unsigned int result = (n1<<8) + n2;
  return result;
}

void Rovio::
ComputeForwardKinematics(double V_l, double V_r, double V_c,
			 double& Vx, double& Vy, double& omega) {
  static double front_factor_avg = 0.0;
  static double front_factor_count = 0.0;
  static const double phi = 0.34907; // 20 degrees
  static const double Lf = 0.74;//0.5;//0.1875;
  static const double Lb = 1.0 - Lf;

  static const double Wx = 270; //375; //375 encoder ticks per meter forwards
  static const double Wy = 100; //140; //encoder ticks per meter sideways?
  static const double Wt = 90; //135; //205 encoder ticks per radian of rotation?

  Vx = (-V_l + -V_r)/2*cos(phi)/Wx;
  double front_vel = (V_l - V_r)/2.0 * sin(phi)*Lf;

  double rear_vel = (V_c * Lb);
  Vy = (front_vel - rear_vel)/Wy;
  omega = (-V_r + V_l + V_c) / Wt;
  //  cout <<"Vx: "<<Vx<<" Vy: "<<Vy <<" omega: "<<omega<<endl;
}

void Rovio::
libcurl_getenc(char* base) {
  static double max_encoder_val = 0.0;
  string buffer;
  char urlstr[100];
  sprintf(urlstr,"%s/rev.cgi?Cmd=nav&action=20",base);
  pthread_mutex_lock(&curl_mutex);
  CURLcode result = libcurl_sendcommand(mutexed_curl,urlstr, buffer);
  pthread_mutex_unlock(&curl_mutex);
  if (result == CURLE_OK)  
    {  
      //printf("Status message : %s\n", buffer.c_str());
      char * offset = strstr(buffer.c_str(), "responses = ");
      unsigned char coded_resp[40];
      int toknum = sscanf(offset, "responses = %s", coded_resp);
           // printf("Coded response: %s\n", 
      	    // coded_resp);
      unsigned int ind = 0;
      unsigned int packet_length = get_byte(coded_resp,ind);
      get_byte(coded_resp,ind); //unused

      //left wheel:
      unsigned int lw_dir = get_byte(coded_resp,ind);
      unsigned int lw_count = get_double_byte(coded_resp,ind);
      
      //right wheel
      unsigned int rw_dir = get_byte(coded_resp,ind);
      unsigned int rw_count = get_double_byte(coded_resp,ind);
      
      //rear-centre wheel
      unsigned int cw_dir = get_byte(coded_resp,ind);
      unsigned int cw_count = get_double_byte(coded_resp,ind);

      get_byte(coded_resp,ind); //unused

      unsigned int headpos = get_byte(coded_resp,ind);
      unsigned int battery = get_byte(coded_resp,ind);
      unsigned int stat = get_byte(coded_resp,ind);
           // printf ("Encoder values l: dir %u val %u r: dir %u val %u, c: dir %u val %u.  Batt: %u\n",
      	   //   lw_dir, lw_count, rw_dir, rw_count, cw_dir, cw_count, battery); 
      if (lw_count > max_encoder_val) max_encoder_val = lw_count;
      if (rw_count > max_encoder_val) max_encoder_val = rw_count;
      if (cw_count > max_encoder_val) max_encoder_val = cw_count;
      //      cout <<"MAX ENCODER VAL: "<<max_encoder_val<<endl;
      double Vx, Vy, omega;
      double V_l = ((lw_dir & 4)? -1.0*lw_count:lw_count);
      double V_r = ((rw_dir & 4)? -1.0*rw_count:rw_count);
      double V_c = ((cw_dir & 4)? -1.0*cw_count:cw_count); 

      ComputeForwardKinematics(V_l, V_r, V_c,
			       Vx, Vy, omega);
      //      printf("Kinematics result: Vx: %lf, Vy: %lf, omega: %lf\n",
      //	     Vx, Vy, omega);
      pthread_mutex_lock(&velocity_mutex);
      dX += Vx;
      dY += Vy;
      dTheta += omega;

      // we want to determine if we have finished rotating if we have been
      //  commanded to.  
      if (!donerotate) {
//printfTextWindow("Not done rotating: %lf", dTheta*180/M_PI);
        // a rotation of 15 degrees or more is considered a complete rotation
        //  as each rotation should turn the rovio around 20 degrees
        if (fabs(dTheta * 180.0f / M_PI) > 15.0f) {
          donerotate = true;
        }

        // this is to time out the rotation command in the event that 
        //  something unusual happens and odometry never reads a 15 deg
        //  or more rotation
        else {
          rotatetimeout++;
          if (rotatetimeout == RTO_VAL) {
            rotatetimeout = 0;
            donerotate = true;
          }
       }
     }

      pthread_mutex_unlock(&velocity_mutex);

    }  
  else  
    {  
      cout << "Error: [" << result << "] - " << errorBuffer<<endl;  
      exit(0);
    }    
}

void Rovio::
libcurl_sendcommand_noresp(char * urlstr) {
  string buffer;
  pthread_mutex_lock(&curl_mutex);
  libcurl_sendcommand(general_curl, urlstr, buffer);
  pthread_mutex_unlock(&curl_mutex);
}



void Rovio::libcurl_grabframe( char* base) {
  string buffer;
  char urlstr [100];
  static int frame = 0;

#ifdef NO_FREEIMAGE
      fprintf(stderr, "Error: Rovio camera not implemented! Install FreeImage and change make.include\n");
      return;
#else

 pthread_mutex_lock(&curl_mutex);

  // reset the curframe data
  CurFrame.Data = NULL;
  CurFrame.DataSize = 0;

  sprintf(urlstr,"%s/Jpeg/CamImg%04u.jpg", base,frame++);
  
  CURLcode result = libcurl_sendcommand(general_curl, urlstr, buffer);

  if (result == CURLE_OK) {  

    // make a buffer to hold the image data
    BYTE* membuff = new BYTE[buffer.size()];
    memcpy(membuff, buffer.c_str(), buffer.size());
     
    CurFrame.Data = (char *)membuff;
    CurFrame.DataSize = buffer.size();

      /*

      FIMEMORY * hmem = FreeImage_OpenMemory(membuff, buffer.size());
      
      FIBITMAP* bmp = FreeImage_LoadFromMemory(FIF_JPEG, hmem, 0);
      unsigned int width = FreeImage_GetWidth(bmp);
      unsigned int height = FreeImage_GetHeight(bmp);
      unsigned int pitch = FreeImage_GetPitch(bmp);
      IplImage* out_img = cvCreateImage(cvSize(width,height), IPL_DEPTH_8U, 3);
      	FreeImage_Save(FIF_JPEG, bmp, "test.jpg", 0);
      FREE_IMAGE_TYPE image_type = FreeImage_GetImageType(bmp);
      if (image_type == FIT_RGBF) {
	BYTE* bits = (BYTE*) FreeImage_GetBits(bmp);
	for (unsigned int y = 0;y<height;y++) {
	  FIRGBF* pixel = (FIRGBF*)bits;
	  for (unsigned int x = 0;x<width;x++) {
	    cvSet2D(out_img, height-y-1,x, CV_RGB(pixel[x].red,
						  pixel[x].green,
						  pixel[x].blue));
	  }
	  bits += pitch;
	}
      }
      else if ((image_type == FIT_BITMAP) && (FreeImage_GetBPP(bmp) == 24)) {
	BYTE* bits = (BYTE*) FreeImage_GetBits(bmp);
	for (unsigned int y = 0;y<height;y++) {
	  BYTE* pixel = (BYTE*)bits;
	  for (unsigned int x = 0;x<width;x++) {
	    cvSet2D(out_img, height-y-1,x, CV_RGB((double)pixel[FI_RGBA_RED],
						  (double)pixel[FI_RGBA_GREEN],
						  (double)pixel[FI_RGBA_BLUE]));
	    pixel += 3;
	  }
	  bits += pitch;
	}
      }
      delete membuff;
      FreeImage_Unload(bmp);
      FreeImage_CloseMemory(hmem);
      return out_img;

      */


    }  
  else  
    {  
      cout << "Error: [" << result << "] - " << errorBuffer<<endl;  
    }
 
  pthread_mutex_unlock(&curl_mutex);
#endif
 
}

FrameData Rovio::GetCurFrame(void) {
  FrameData fd;

  pthread_mutex_lock(&curl_mutex);
  fd = CurFrame;
  pthread_mutex_unlock(&curl_mutex);

  return fd;
}


// we want to grab a frame every 1 sec otherwise just return the current frame
void Rovio::GetImage() {

  timeval curtime;

  // grab the frame once a sec	
  gettimeofday(&curtime, NULL);
  if (curtime.tv_sec >= NextFrameTime) {
    NextFrameTime = curtime.tv_sec + 1;

    if (CurFrame.Data != NULL) {
      delete CurFrame.Data;
    }
    libcurl_grabframe(host);
  }

  return;

}

void Rovio::
SetControl(double vx, double vy, double vtheta) {
  string buffer;
  char urlstr[100];
  unsigned int drive = 0;
  unsigned int speed = 5;


#ifndef NO_MLAB_ROVIO
  if (desiredang >= 30.0) {
    if (donerotate) {
      drive = 17;
      speed = 4;
      donerotate = false;
      rotatetimeout = 0;
    }

  }
  else if (desiredang <= -30.0) {
    if (donerotate) {
      drive = 18;
      speed = 4;
      donerotate = false;
      rotatetimeout = 0;
    }
    //printfTextWindow("Rotate\n");
  }
#else
  if (vtheta > 0.04) {
    drive = 5;
  } else if (vtheta < -0.04) {
    drive = 6;
  }
#endif 

  else {
    if (vx > 0) {
      if (vy>0) {
        if (vx > 2*vy) drive = 1; //straight forward
        //      else if (vy < 2*vx) drive = 7; //diagonal forward left
        else drive = 4; //straight left
      }
      else {
        if (vx > -2*vy) drive = 1; //straight forward
        //      else if (vy < -2*vx) drive = 8; //diagonal forward right


        else drive = 3; //straight right
      }
    }
    else if (vx < 0) {
      if (vy>0) {
        if (-vx > 2*vy) drive = 2; //straight backward
        //      else if (vy < -2*vx) drive = 9; //diagonal backward left
        else drive = 4; //straight left
      }
      else {
        if (-vx > -2*vy) drive = 2; //straight backward
        //      else if (vy > 2*vx) drive = 10; //diagonal backward right
        else drive = 3; //straight right
      }
    }
    else {
      drive= 0; //All stop
    }

  }
#ifndef NO_MLAB_ROVIO
  startrot = false;
#else
#endif

  sprintf(urlstr,"%s/rev.cgi?Cmd=nav&action=18&drive=%u&speed=%u", host,
          drive, speed);
  pthread_mutex_lock(&curl_mutex);
  CURLcode result = libcurl_sendcommand(general_curl,urlstr, buffer);
  pthread_mutex_unlock(&curl_mutex);
  if (result == CURLE_OK)
    {
      char * offset = strstr(buffer.c_str(), "responses = ");
      unsigned int stat;
      int toknum = sscanf(offset, "responses = %d", &stat);
    }  
  else {
    printf("Failed to send movement command to Rovio\n");
  }
}


void Rovio::
GetPose(double& x, double& y, double& theta) {
  pthread_mutex_lock(&velocity_mutex);
  X += dX*cos(Theta + dTheta/2.0) - dY*sin(Theta + dTheta/2.0);
  Y += dX*sin(Theta + dTheta/2.0) + dY*cos(Theta + dTheta/2.0);
  Theta += dTheta;
  //cout <<"Snapshot: X:"<<X<<" Y: "<<Y<<" Theta:" <<Theta<<endl;
  dX = 0.0;
  dY = 0.0;
  dTheta = 0.0;
  x = X;
  y = Y;
  theta = ROBOT_RAD2DEG(Theta);
  pthread_mutex_unlock(&velocity_mutex);

}
