
#ifndef TRILATERATION
#define TRILATERATION

#include <list>
#include <math.h>

#include "../PoseCalculator/Kalman/mxn_matrix.h"

using namespace std;

static int beaconloc[][3] = {     {0,0,0},
                           {611, 471 ,391 },
                           {613 ,672 ,391 },
                           {258 ,471 ,396 },
                           {778 ,458 ,390 },
                           {274 ,168 ,292 },
                           {419 ,361 ,399 },
                           {835 ,221 ,391 },
                           {594 ,220 ,391 },
                           {903 ,362 ,391 }    };
 
typedef struct beaconrecord {
	double timestamp;
	int beaconid;
	double db;

	void print() {
		printf(" (%lf: B%d,%lf) \n", timestamp, beaconid, db);
	}
};

typedef struct beaconposition {
	double x, y, z;
	beaconposition () {
		x = y = z = 0.0;
	}

	beaconposition (double x, double y, double z) {
		this->x = x;
		this->y = y;
		this->z = z;
	}

	void clear() {
		x = y = z = 0.0;
	}

	double dot(beaconposition bp) {
		return (x*bp.x + y*bp.y + z*bp.z);
	}

	void normalize() {
		double length = sqrt((x*x)+(y*y)+(z*z));
		if (length > 0.0) {
			x = x/length;
			y = y/length;
			z = z/length;
		}
	}

	beaconposition cross(beaconposition bp) {
		beaconposition result;
		result.x = y*bp.z-z*bp.y;
		result.y = z*bp.x-x*bp.z;
		result.z = x*bp.y-y*bp.x;
		return result;
	}

	void print() {
		printf("(%lf,%lf,%lf)\n", x, y, z);
	}
};


class Trilateration {

protected:

  //(* beaconhistory is simply a representation of the only part of the cricketd output that
  //we care about.  For only those cricketd records that contain a valid DB, we store {timestamp, beacon number, and DB value.  
  //The beacon number is the number that follows "BEACON-" *)
  //(* For a continuously running solver, beaconhistory would presumably be trimmed continously at the front end, or it 
  // would grow forever.  It should be kept at least as long as mintimesep, and maybe 10 times longer to allow for
  //a varying mintimesep *)
  list <beaconrecord> beaconhistory;	
  double timestamp;
  beaconposition lastr;
  beaconposition rval;
  int debug;
  bool errordebug;

  // Perform transformation (represented as a matrix) on vector
  beaconposition transformVector(beaconposition bp, mxn_Matrix *matrix);

public:

  Trilateration();

  bool addData(beaconrecord br);
  bool getPosition(beaconposition &bp);
};

#endif
