/**********************************************************************
 **                                                                  **
 **                          FrameGrabber.h                          **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Video framegrabber control for HServer                          **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: frame_grabber.h,v 1.1 2008-07-15 22:33:48 endo Exp $ */

#ifndef FRAME_GRABBER_H
#define FRAME_GRABBER_H

#include <linux/videodev.h>
#include <string>
#include "video.h"

class FrameGrabber : public Video {

protected:
    struct video_audio videoAudio_;
    struct video_capability videoCapability_;
    struct video_channel videoChannel_;
    struct video_mmap videoMmap_;
    struct video_picture videoPicture_;
    string portString_;
    unsigned char *mappedImage_;
    int channel_;
    int shCurrentImageFd_;

    static const int DEFAULT_CHANNEL_;
    static const int DEFAULT_HEIGHT_;
    static const int DEFAULT_WIDTH_;

    void updateStatusBar_(void);
    void setSize_(int height, int width);
    bool initV4Linux_(int width, int height);
    bool readVideoData_(void);

public:
    FrameGrabber(
        Video **a,
        string portString,
        bool showImage);
    ~FrameGrabber(void);
    void control(void);
    void setChannel(int channel);
};

#endif

/**********************************************************************
# $Log: frame_grabber.h,v $
# Revision 1.1  2008-07-15 22:33:48  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.4  2008/04/11 20:22:44  endo
# FrameGrabber fixed.
#
# Revision 1.3  2008/01/21 19:14:15  endo
# Regular backup.
#
# Revision 1.2  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.1  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2006/05/15 06:01:03  endo
# gcc-3.4 upgrade
#
# Revision 1.1.1.1  2005/02/06 22:59:43  endo
# AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
#
# Revision 1.4  2001/03/23 21:32:51  blee
# altered to use a config file
#
# Revision 1.3  2000/09/19 03:51:29  endo
# RCS log added.
#
#
#**********************************************************************/
