/**********************************************************************
 **                                                                  **
 **                         gazebo-video.h                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for Gazebo camera (Player/Stage)                      **
 **                                                                  **
 **  Copyright 2006, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: gazebo-video.h,v 1.1 2008/07/15 22:33:50 endo Exp $ */

#ifndef GAZEBO_VIDEO_H
#define GAZEBO_VIDEO_H

#include <pthread.h>
#include <string>
#include <vector>

#include "module.h"
#include "gazebo.h"
#include "hserver.h"
#include "video.h"

using std::string;
using std::vector;

#define GAZEBO_VIDEO_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}
#define GAZEBO_VIDEO_RAD2DEG(val) ((val)*180.0/M_PI)
#define GAZEBO_VIDEO_DEG2RAD(val) ((val)*M_PI/180.0)

class GazeboVideo : public Video {

protected:
    gz_client_t *gazeboClient_;
    gz_camera_t *gazeboVideo_;
    gz_truth_t *gazeboTruth_;
    string gazeboVideoName_;
    string gazeboTruthName_;
    double lastTimeStamp_;
    int gazeboServerID_;
    int gazeboClientID_;

    static const string DEFAULT_GAZEBO_VIDEO_NAME_;
    static const string DEFAULT_GAZEBO_TRUTH_NAME_;
    static const int DEFAULT_GAZEBO_SERVER_ID_;
    static const int DEFAULT_GAZEBO_CLIENT_ID_;
    static const bool USE_GROUND_TRUTH_;

    void updateStatusBar_(void);
    bool initializeGazebo_(void);
    bool readVideoData_(void);

public:
    GazeboVideo(
        Video **a,
        char *gazeboVideoName,
        char *gazeboTruthName,
        bool showImage);
    ~GazeboVideo(void);
    void control(void);
    void setChannel(int channel);
};    

#endif

/**********************************************************************
# $Log: gazebo-video.h,v $
# Revision 1.1  2008/07/15 22:33:50  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.5  2008/01/21 19:14:15  endo
# Regular backup.
#
# Revision 1.4  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.3  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.2  2006/11/20 22:55:52  endo
# OpenCV added.
#
# Revision 1.1  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
#**********************************************************************/
