/**********************************************************************
 **                                                                  **
 **                         HServerConfig.h                          **
 **                                                                  **
 **                                                                  **
 **  Written by:  Brian Lee                                          **
 **                                                                  **
 **  Copyright 2002 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: HServerConfig.h,v 1.10 2010-05-03 22:12:57 zkira Exp $ */

#ifndef HSERVERCONFIG_H
#define HSERVERCONFIG_H

#include <string>
#include <vector>
#include "HServerConfigSection.h"

using std::string;
using std::vector;

class ClHServerConfig
{
public:
    enum EnCameraValueType
    {
        EnCameraValueType_PORT_STRING = 0,
        EnCameraValueType_COUNT     // this must be declared last
    };

    enum EnLaserValueType
    {
        EnLaserValueType_NAME = 0,
        EnLaserValueType_PORT_STRING,
        EnLaserValueType_ANGLE_OFFSET,
        EnLaserValueType_X_OFFSET,
        EnLaserValueType_Y_OFFSET,
        EnLaserValueType_LISTEN_IPT,
        EnLaserValueType_SEND_IPT,
        EnLaserValueType_STREAM_HOST,
        EnLaserValueType_COUNT     // this must be declared last
    };

    enum EnRobotValueType
    {
        EnRobotValueType_MIN_POINT_TURN_ANGLE = 0,
        EnRobotValueType_MAX_SPEED,
        EnRobotValueType_MAX_STEER_SPEED,
        EnRobotValueType_MAX_STEER_SPEED_ANGLE,
        EnRobotValueType_MAX_ALTITUDE_SPEED,
        EnRobotValueType_MAX_NO_COMMAND_TIME_SEC,
        EnRobotValueType_MIN_STEER_SPEED,
        EnRobotValueType_STOPPING_SPEED_THRESHOLD,
        EnRobotValueType_NUM_VELOCITY_SAMPLES,
        EnRobotValueType_COUNT     // this must be declared last
    };

    enum EnGazeboValueType
    {
        EnGazeboValueType_NAME = 0,
        EnGazeboValueType_TRUTH_NAME,
        EnGazeboValueType_TYPE,
        EnGazeboValueType_ROBOT_CLASS,
        EnGazeboValueType_CLIENT_ID,
        EnGazeboValueType_SERVER_ID,
        EnGazeboValueType_ANGLE_OFFSET,
        EnGazeboValueType_X_OFFSET,
        EnGazeboValueType_Y_OFFSET,
        EnGazeboValueType_LASER_PITCH_ANGLE,
        EnGazeboValueType_SHOW_IMAGE,
        EnGazeboValueType_COUNT     // this must be declared last
    };

    enum EnUSARSimValueType
    {
        EnUSARSimValueType_HOST = 0,
        EnUSARSimValueType_NAME,
        EnUSARSimValueType_ROBOT_CLASS,
        EnUSARSimValueType_PORT,
        EnUSARSimValueType_TYPE,
        EnUSARSimValueType_ANGLE_OFFSET,
        EnUSARSimValueType_POS_X,
        EnUSARSimValueType_POS_Y,
        EnUSARSimValueType_POS_Z,
        EnUSARSimValueType_ROT_YAW,
        EnUSARSimValueType_ROT_PITCH,
        EnUSARSimValueType_ROT_ROLL,
        EnUSARSimValueType_X_OFFSET,
        EnUSARSimValueType_Y_OFFSET,
        EnUSARSimValueType_LASER_PITCH_ANGLE,
        EnUSARSimValueType_SHOW_IMAGE,
        EnUSARSimValueType_COUNT     // this must be declared last
    };

    enum EnRovioValueType
    {
        EnRovioValueType_HOST = 0,
        EnRovioValueType_NAME,
        EnRovioValueType_ROBOT_CLASS,
        EnRovioValueType_PORT,
        EnRovioValueType_TYPE,
        EnRovioValueType_ANGLE_OFFSET,
        EnRovioValueType_POS_X,
        EnRovioValueType_POS_Y,
        EnRovioValueType_POS_Z,
        EnRovioValueType_ROT_YAW,
        EnRovioValueType_ROT_PITCH,
        EnRovioValueType_ROT_ROLL,
        EnRovioValueType_X_OFFSET,
        EnRovioValueType_Y_OFFSET,
        EnRovioValueType_SHOW_IMAGE,
        EnRovioValueType_COUNT     // this must be declared last
    };

    enum EnSkybotixValueType
    {
        EnSkybotixValueType_HOST = 0,
        EnSkybotixValueType_NAME,
        EnSkybotixValueType_ROBOT_CLASS,
        EnSkybotixValueType_PORT,
        EnSkybotixValueType_TYPE,
        EnSkybotixValueType_ANGLE_OFFSET,
        EnSkybotixValueType_POS_X,
        EnSkybotixValueType_POS_Y,
        EnSkybotixValueType_POS_Z,
        EnSkybotixValueType_ROT_YAW,
        EnSkybotixValueType_ROT_PITCH,
        EnSkybotixValueType_ROT_ROLL,
        EnSkybotixValueType_X_OFFSET,
        EnSkybotixValueType_Y_OFFSET,
        EnSkybotixValueType_SHOW_IMAGE,
        EnSkybotixValueType_COUNT     // this must be declared last
    };


    enum EnOpenCVValueType
    {
        EnOpenCVValueType_TRACK_BLOBS = 0,
        EnOpenCVValueType_DETECT_FACE,
        EnOpenCVValueType_DETECT_OBJECTS,
        EnOpenCVValueType_DIRECT_GRAB,
        EnOpenCVValueType_SHOW_IMAGE,
        EnOpenCVValueType_COUNT     // this must be declared last
    };

    enum EnCognachromeValueType
    {
        EnCognachromeValueType_PORT_STRING = 0,
        EnCognachromeValueType_COUNT     // this must be declared last
    };

    enum EnNomadValueType
    {
        EnNomadValueType_TYPE = 0,
        EnNomadValueType_PORT_STRING,
        EnNomadValueType_HOST_NAME,
        EnNomadValueType_COUNT     // this must be declared last
    };

    enum EnPioneerValueType
    {
        EnPioneerValueType_PORT_STRING = 0,
        EnPioneerValueType_COUNT     // this must be declared last
    };

    enum EnPioneer2ValueType
    {
        EnPioneer2ValueType_PORT_STRING = 0,
        EnPioneer2ValueType_COUNT     // this must be declared last
    };

    enum EnAmigoBotValueType
    {
        EnAmigoBotValueType_PORT_STRING = 0,
        EnAmigoBotValueType_COUNT     // this must be declared last
    };

    enum EnFrameGrabberValueType
    {
        EnFrameGrabberValueType_PORT_STRING = 0,
        EnFrameGrabberValueType_SHOW_IMAGE,
        EnFrameGrabberValueType_COUNT     // this must be declared last
    };

    enum EnGpsValueType
    {
        EnGpsValueType_USE_BASE = 0,
        EnGpsValueType_PORT_STRING,
        EnGpsValueType_BASE_LAT,
        EnGpsValueType_BASE_LONG,
        EnGpsValueType_X_DIFF,
        EnGpsValueType_Y_DIFF,
        EnGpsValueType_M_PER_LAT,
        EnGpsValueType_M_PER_LONG,
        EnGpsValueType_COUNT     // this must be declared last
    };

    enum EnCricketValueType
    {
        EnCricketValueType_ID = 0,
        EnCricketValueType_IP,
        EnCricketValueType_COUNT
    };

    enum EnJboxValueType
    {
        EnJboxValueType_JBOX_ID = 0,
        EnJboxValueType_DISABLE_NETWORK_QUERY,
        EnJboxValueType_GPS_SECTION,
        EnJboxValueType_COUNT     // this must be declared last
    };

    enum EnCompassValueType
    {
        EnCompassValueType_TYPE = 0,
        EnCompassValueType_PORT_STRING,
        EnCompassValueType_COUNT     // this must be declared last
    };

    enum EnGyroValueType
    {
        EnGyroValueType_TYPE = 0,
        EnGyroValueType_COUNT     // this must be declared last
    };

    enum EnPoseCalcValueType
    {
        EnPoseCalcValueType_FUSER_TYPE = 0,
        EnPoseCalcValueType_VARIANCE_SCHEME,
        EnPoseCalcValueType_COMPASS2MLAB_HEADING_FACTOR,
        EnPoseCalcValueType_COMPASS2MLAB_HEADING_OFFSET,
        EnPoseCalcValueType_COMPASS_MAX_VALID_ANGSPEED4HEADING,
        EnPoseCalcValueType_GPS2MLAB_HEADING_FACTOR,
        EnPoseCalcValueType_GPS2MLAB_HEADING_OFFSET,
        EnPoseCalcValueType_GPS_MIN_VALID_TRANSSPEED4HEADING,
        EnPoseCalcValueType_GPS_MAX_VALID_ANGSPEED4HEADING,
        EnPoseCalcValueType_COUNT     // this must be declared last
    };

    enum EnHServerConfigSectionType
    {
        EnHServerConfigSectionType_CAMERA = 0,
        EnHServerConfigSectionType_LASER,
        EnHServerConfigSectionType_ROBOT,
        EnHServerConfigSectionType_GAZEBO,
        EnHServerConfigSectionType_USARSIM,
        EnHServerConfigSectionType_ROVIO,
        EnHServerConfigSectionType_SKYBOTIX,
        EnHServerConfigSectionType_OPENCV,
        EnHServerConfigSectionType_COGNACHROME,
        EnHServerConfigSectionType_NOMAD,
        EnHServerConfigSectionType_PIONEER,
        EnHServerConfigSectionType_PIONEER2,
        EnHServerConfigSectionType_AMIGOBOT,
        EnHServerConfigSectionType_FRAMEGRABBER,
        EnHServerConfigSectionType_GPS,
        EnHServerConfigSectionType_CRICKET,
        EnHServerConfigSectionType_JBOX,
        EnHServerConfigSectionType_COMPASS,
        EnHServerConfigSectionType_GYRO,
        EnHServerConfigSectionType_POSECALC,
        EnHServerConfigSectionType_COUNT     // this must be declared last
    };

    enum EnHServerConfigError
    {
        EnHServerConfigError_NONE,
        EnHServerConfigError_BAD_CONFIG_TYPE
    };

    static const int m_aiSECTION_VALUE_COUNTS[];

    ClHServerConfig();
    ~ClHServerConfig();

    bool AddNewSection( int iSectionType, string strName );
    unsigned int GetSectionCount() const;
    int GetSectionType( int iSection ) const;
    bool GetValue( int iSection, int iValueType, string& strValue );
    bool InsertValue( int iSection, int iValueType, const string& strValue );
    bool GetSectionName( int iSection, string& strName ) const;
    bool SetSectionName( int iSection, string strName );

protected:
    vector<ClHServerConfigSection> m_clSectionList;

    bool ValidSection( int iSection ) const;
};

inline unsigned int ClHServerConfig::GetSectionCount() const
{
    return m_clSectionList.size();
}

inline bool ClHServerConfig::ValidSection( int iSection ) const
{
    return ( ( iSection >= 0 ) && ( iSection < (int) m_clSectionList.size() ) );
}

inline int ClHServerConfig::GetSectionType( int iSection ) const
{
    return ( ValidSection( iSection ) ? m_clSectionList[ iSection ].GetType() : -1 );
}

#endif // #ifndef _HSERVERCONFIG_H

/**********************************************************************
# $Log: HServerConfig.h,v $
# Revision 1.10  2010-05-03 22:12:57  zkira
# Addition of rovio & cricket
#
# Revision 1.8  2009/04/17 17:50:05  zkira
# Added max_speed to differentiate it from max_steer_speed
#
# Revision 1.7  2009/04/13 02:58:54  zkira
# max altitude speed & other robot parameters
#
# Revision 1.6  2009/03/02 18:47:07  zkira
# Added ability to change low level robot parameters (e.g. max speed) in the .hserverrc file.
# Added Rovio
#
# Revision 1.5  2009/01/27 20:21:16  zkira
# Added gazebo laser_pitch parameter
#
# Revision 1.4  2008/11/20 03:53:57  ts2883
# The pitch angle information for the laser can now be parsed and read by the
# hserver using .hserverrc file.
#
# Revision 1.3  2008/09/16 16:47:37  zkira
# Added Gazebo robot class, clientID, and serverID to configuration file
#
# Revision 1.2  2008/07/15 22:34:01  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.10  2008/04/13 00:33:42  endo
# Regular backup.
#
# Revision 1.9  2008/04/12 20:17:24  endo
# OpenCV face detection implemented.
#
# Revision 1.8  2008/01/13 00:35:40  endo
# Regular backup.
#
# Revision 1.7  2008/01/09 22:54:03  endo
# USARSimRobot added.
#
# Revision 1.6  2006/12/01 21:14:54  endo
# Regular backup.
#
# Revision 1.5  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.4  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
# Revision 1.3  2006/11/15 23:45:58  endo
# GazeboLaser can now accept the second laser.
#
# Revision 1.2  2006/08/08 21:33:49  endo
# GazeboLaser added.
#
# Revision 1.1.1.1  2006/07/21 17:42:16  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:57  endo
# MissionLab 7.0
#
# Revision 1.2  2006/06/05 21:35:54  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.5  2004/11/22 00:19:22  endo
# Network query of JBox can now be disabled.
#
# Revision 1.4  2004/09/22 04:00:58  endo
# *** empty log message ***
#
# Revision 1.3  2004/09/10 20:05:05  endo
# New PoseCalc integrated.
#
# Revision 1.2  2004/04/24 11:20:54  endo
# 3DM-G added.
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.4  2003/04/26 18:31:06  ebeowulf
# Checked in Pioneer2-dxe module.
#
# Revision 1.3  2003/04/06 13:35:53  endo
# gcc 3.1.1
#
# Revision 1.2  2002/02/18 13:52:47  endo
# AmigoBot added.
#
#
#**********************************************************************/
