/**********************************************************************
 **                                                                  **
 **                       HServerConfigurator.h                      **
 **                                                                  **
 **                                                                  **
 **  Written by:  Brian Lee                                          **
 **                                                                  **
 **  Copyright 2002 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: HServerConfigurator.h,v 1.10 2010-05-03 22:12:57 zkira Exp $ */

#ifndef HSERVERCONFIGURATOR_H
#define HSERVERCONFIGURATOR_H

#include <string>
#include "HServerConfig.h"
#include "nomad.h"

// this holds initialization data for the lasers
struct SuLaserInitData
{
    string strPort;
    bool bSend;
    bool bListen;
    float fAngleOffset;
    float fXOffset;
    float fYOffset;
    string strName;
    string strStreamAddress;

    SuLaserInitData() {}
    SuLaserInitData( const SuLaserInitData& suOriginal ) { *this = suOriginal; }
    inline SuLaserInitData& operator=( const SuLaserInitData& suOriginal );
};

inline SuLaserInitData& SuLaserInitData::operator=( const SuLaserInitData& suOriginal )
{
    if ( this != & suOriginal )
    {
        strPort = suOriginal.strPort;
        bSend = suOriginal.bSend;
        bListen = suOriginal.bListen;
        fAngleOffset = suOriginal.fAngleOffset;
        fXOffset = suOriginal.fXOffset;
        fYOffset = suOriginal.fYOffset;
        strName = suOriginal.strName;
        strStreamAddress = suOriginal.strStreamAddress;
    }

    return *this;
}

// This holds general initialization data for any robot
// It determines mobility type and characteristics (e.g. min point turn angle)
// and other general parameters.
struct SuRobotInitData
{
   double minPointTurnAngle;
   double maxSpeed;
   double maxSteerSpeed;
   double maxSteerSpeedAngle;	
   double maxAltitudeSpeed;
   double maxNoCommandTimeSec;
   double minSteerSpeed;
   double stoppingSpeedThreshold;
   int numVelocitySamples;

   SuRobotInitData() {}
   SuRobotInitData(const SuRobotInitData& suOriginal) {*this = suOriginal;}
   inline SuRobotInitData& operator=(const SuRobotInitData& suOriginal);

};

inline SuRobotInitData& SuRobotInitData::operator=(const SuRobotInitData& suOriginal)
{
    if (this != & suOriginal)
    {
        minPointTurnAngle = suOriginal.minPointTurnAngle;
        maxSpeed = suOriginal.maxSpeed;
        maxSteerSpeed = suOriginal.maxSteerSpeed;
        maxSteerSpeedAngle = suOriginal.maxSteerSpeedAngle;
        maxAltitudeSpeed = suOriginal.maxAltitudeSpeed;
        maxNoCommandTimeSec = suOriginal.maxNoCommandTimeSec;
        minSteerSpeed = suOriginal.minSteerSpeed;
        stoppingSpeedThreshold = suOriginal.stoppingSpeedThreshold;
        numVelocitySamples = suOriginal.numVelocitySamples;
    }

    return *this;
}

// this holds initialization data for Gazebo
struct SuGazeboInitData
{
    string name;
    string truthName;
    int type;
    string robot_class;
    int clientID;
    int serverID;
    float angleOffset;
    float xOffset;
    float yOffset;
    float laserPitchAngle;
    bool showImage;

    SuGazeboInitData() {}
    SuGazeboInitData(const SuGazeboInitData& suOriginal) {*this = suOriginal;}
    inline SuGazeboInitData& operator=(const SuGazeboInitData& suOriginal);
};

inline SuGazeboInitData& SuGazeboInitData::operator=(const SuGazeboInitData& suOriginal)
{
    if (this != & suOriginal)
    {
        name = suOriginal.name;
        truthName = suOriginal.truthName;
        type = suOriginal.type;
        robot_class = suOriginal.robot_class;
        clientID = suOriginal.clientID;
        serverID = suOriginal.serverID;
        angleOffset = suOriginal.angleOffset;
        xOffset = suOriginal.xOffset;
        yOffset = suOriginal.yOffset;
        laserPitchAngle = suOriginal.laserPitchAngle;
        showImage = suOriginal.showImage;
    }

    return *this;
}

// this holds initialization data for USARSim
struct SuUSARSimInitData
{
    string host;
    string name;
    string robot_class;
    int port;
    int type;
    float angleOffset;
    float pos_x;
    float pos_y;
    float pos_z;
    float rot_yaw;
    float rot_pitch;
    float rot_roll;
    float xOffset;
    float yOffset;
    float laserPitchAngle;
    bool showImage;

    SuUSARSimInitData() {}
    SuUSARSimInitData(const SuUSARSimInitData& suOriginal) {*this = suOriginal;}
    inline SuUSARSimInitData& operator=(const SuUSARSimInitData& suOriginal);
};

inline SuUSARSimInitData& SuUSARSimInitData::operator=(const SuUSARSimInitData& suOriginal)
{
    if (this != & suOriginal)
    {
        host = suOriginal.host;
        name = suOriginal.name;
        robot_class = suOriginal.robot_class;
        port = suOriginal.port;
        type = suOriginal.type;
        angleOffset = suOriginal.angleOffset;
        pos_x = suOriginal.pos_x;
        pos_y = suOriginal.pos_y;
        pos_z = suOriginal.pos_z;
        rot_yaw = suOriginal.rot_yaw;
        rot_pitch = suOriginal.rot_pitch;
        rot_roll = suOriginal.rot_roll;
        xOffset = suOriginal.xOffset;
        yOffset = suOriginal.yOffset;
        laserPitchAngle = suOriginal.laserPitchAngle;
        showImage = suOriginal.showImage;
    }

    return *this;
}

// this holds initialization data for Rovio
struct SuRovioInitData
{
    string host;
    string name;
    string robot_class;
    int port;
    int type;
    float angleOffset;
    float pos_x;
    float pos_y;
    float pos_z;
    float rot_yaw;
    float rot_pitch;
    float rot_roll;
    float xOffset;
    float yOffset;
    bool showImage;

    SuRovioInitData() {}
    SuRovioInitData(const SuRovioInitData& suOriginal) {*this = suOriginal;}
    inline SuRovioInitData& operator=(const SuRovioInitData& suOriginal);
};

inline SuRovioInitData& SuRovioInitData::operator=(const SuRovioInitData& suOriginal)
{
    if (this != & suOriginal)
    {
        host = suOriginal.host;
        name = suOriginal.name;
        robot_class = suOriginal.robot_class;
        port = suOriginal.port;
        type = suOriginal.type;
        angleOffset = suOriginal.angleOffset;
        pos_x = suOriginal.pos_x;
        pos_y = suOriginal.pos_y;
        pos_z = suOriginal.pos_z;
        rot_yaw = suOriginal.rot_yaw;
        rot_pitch = suOriginal.rot_pitch;
        rot_roll = suOriginal.rot_roll;
        xOffset = suOriginal.xOffset;
        yOffset = suOriginal.yOffset;
        showImage = suOriginal.showImage;
    }

    return *this;
}



// this holds initialization data for Skybotix
struct SuSkybotixInitData {
    string host;
    string name;
    string robot_class;
    int port;
    int type;
    float angleOffset;
    float pos_x;
    float pos_y;
    float pos_z;
    float rot_yaw;
    float rot_pitch;
    float rot_roll;
    float xOffset;
    float yOffset;
    bool showImage;

    SuSkybotixInitData() {}
    SuSkybotixInitData(const SuSkybotixInitData& suOriginal) {*this = suOriginal;}
    inline SuSkybotixInitData& operator=(const SuSkybotixInitData& suOriginal);
};

inline SuSkybotixInitData& SuSkybotixInitData::operator=(const SuSkybotixInitData& suOriginal)
{
    if (this != & suOriginal)
    {
        host = suOriginal.host;
        name = suOriginal.name;
        robot_class = suOriginal.robot_class;
        port = suOriginal.port;
        type = suOriginal.type;
        angleOffset = suOriginal.angleOffset;
        pos_x = suOriginal.pos_x;
        pos_y = suOriginal.pos_y;
        pos_z = suOriginal.pos_z;
        rot_yaw = suOriginal.rot_yaw;
        rot_pitch = suOriginal.rot_pitch;
        rot_roll = suOriginal.rot_roll;
        xOffset = suOriginal.xOffset;
        yOffset = suOriginal.yOffset;
        showImage = suOriginal.showImage;
    }

    return *this;
}



// this holds initialization data for OpenCV
struct SuOpenCVInitData
{
    bool trackBlobs;
    bool detectFace;
    bool detectObjects;
    bool directGrab;
    bool showImage;

    SuOpenCVInitData() {}
    SuOpenCVInitData(const SuOpenCVInitData& suOriginal) {*this = suOriginal;}
    inline SuOpenCVInitData& operator=(const SuOpenCVInitData& suOriginal);
};

inline SuOpenCVInitData& SuOpenCVInitData::operator=(const SuOpenCVInitData& suOriginal)
{
    if (this != & suOriginal)
    {
        trackBlobs = suOriginal.trackBlobs;
        detectFace = suOriginal.detectFace;
        detectObjects = suOriginal.detectObjects;
        directGrab = suOriginal.directGrab;
        showImage = suOriginal.showImage;
    }

    return *this;
}

struct SuCameraInitData
{
    string strPortString;

    SuCameraInitData() {}
    SuCameraInitData( const SuCameraInitData& suOriginal ) { *this = suOriginal; }
    inline SuCameraInitData& operator=( const SuCameraInitData& suOriginal );
};

inline SuCameraInitData& SuCameraInitData::operator=( const SuCameraInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        strPortString = suOriginal.strPortString;
    }

    return *this;
}

struct SuCognachromeInitData
{
    string strPortString;

    SuCognachromeInitData() {}
    SuCognachromeInitData( const SuCognachromeInitData& suOriginal ) { *this = suOriginal; }
    inline SuCognachromeInitData& operator=( const SuCognachromeInitData& suOriginal );
};

inline SuCognachromeInitData& SuCognachromeInitData::operator=( const SuCognachromeInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        strPortString = suOriginal.strPortString;
    }

    return *this;
}

struct SuNomadInitData
{
    Nomad::EnNomadType enType;
    string strPortString;
    string strHost;

    SuNomadInitData() {}
    SuNomadInitData( const SuNomadInitData& suOriginal ) { *this = suOriginal; }
    inline SuNomadInitData& operator=( const SuNomadInitData& suOriginal );
};

inline SuNomadInitData& SuNomadInitData::operator=( const SuNomadInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        enType = suOriginal.enType;
        strPortString = suOriginal.strPortString;
        strHost = suOriginal.strHost;
    }

    return *this;
}

struct SuPioneerInitData
{
    string strPortString;

    SuPioneerInitData() {}
    SuPioneerInitData( const SuPioneerInitData& suOriginal ) { *this = suOriginal; }
    inline SuPioneerInitData& operator=( const SuPioneerInitData& suOriginal );
};

inline SuPioneerInitData& SuPioneerInitData::operator=( const SuPioneerInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        strPortString = suOriginal.strPortString;
    }

    return *this;
}

struct SuPioneer2InitData
{
    string strPortString;

    SuPioneer2InitData() {}
    SuPioneer2InitData( const SuPioneer2InitData& suOriginal ) { *this = suOriginal; }
    inline SuPioneer2InitData& operator=( const SuPioneer2InitData& suOriginal );
};

inline SuPioneer2InitData& SuPioneer2InitData::operator=( const SuPioneer2InitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        strPortString = suOriginal.strPortString;
    }

    return *this;
}

struct SuAmigoBotInitData
{
    string strPortString;

    SuAmigoBotInitData() {}
    SuAmigoBotInitData( const SuAmigoBotInitData& suOriginal ) { *this = suOriginal; }
    inline SuAmigoBotInitData& operator=( const SuAmigoBotInitData& suOriginal );
};

inline SuAmigoBotInitData& SuAmigoBotInitData::operator=( const SuAmigoBotInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        strPortString = suOriginal.strPortString;
    }

    return *this;
}

struct SuFrameGrabberInitData
{
    string strPortString;
    bool showImage;

    SuFrameGrabberInitData() {}
    SuFrameGrabberInitData( const SuFrameGrabberInitData& suOriginal ) { *this = suOriginal; }
    inline SuFrameGrabberInitData& operator=( const SuFrameGrabberInitData& suOriginal );
};

inline SuFrameGrabberInitData& SuFrameGrabberInitData::operator=( const SuFrameGrabberInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        strPortString = suOriginal.strPortString;
        showImage = suOriginal.showImage;
    }

    return *this;
}

struct SuCricketInitData
{
    int ID;
    string IP;

    SuCricketInitData() {}
    SuCricketInitData( const SuCricketInitData& suOriginal ) { *this = suOriginal; }
    inline SuCricketInitData& operator=( const SuCricketInitData& suOriginal ) {
      if ( this != &suOriginal ) {
        ID = suOriginal.ID;
        IP = suOriginal.IP;
      }

      return *this;
   };
};


struct SuGpsInitData
{
    bool bUseBase;
    string strPortString;
    double dBaseLat;
    double dBaseLong;
    double dXDiff;
    double dYDiff;
    double dMPerLat;
    double dMPerLong;

    SuGpsInitData() {}
    SuGpsInitData( const SuGpsInitData& suOriginal ) { *this = suOriginal; }
    inline SuGpsInitData& operator=( const SuGpsInitData& suOriginal );
};

inline SuGpsInitData& SuGpsInitData::operator=( const SuGpsInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        bUseBase = suOriginal.bUseBase;
        strPortString = suOriginal.strPortString;
        dBaseLat = suOriginal.dBaseLat;
        dBaseLong = suOriginal.dBaseLong;
        dXDiff = suOriginal.dXDiff;
        dYDiff = suOriginal.dYDiff;
        dMPerLat = suOriginal.dMPerLat;
        dMPerLong = suOriginal.dMPerLong;
    }

    return *this;
}

struct SuJboxInitData
{
    int jboxID;
    bool disableNetworkQuery;
    string gpsSection;

    SuJboxInitData() {}
    SuJboxInitData( const SuJboxInitData& suOriginal ) { *this = suOriginal; }
    inline SuJboxInitData& operator=( const SuJboxInitData& suOriginal );
};

inline SuJboxInitData& SuJboxInitData::operator=( const SuJboxInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        jboxID = suOriginal.jboxID;
        disableNetworkQuery = suOriginal.disableNetworkQuery;
        gpsSection = suOriginal.gpsSection;
    }

    return *this;
}

struct SuCompassInitData
{
    int type;
    string portString;

    SuCompassInitData() {}
    SuCompassInitData( const SuCompassInitData& suOriginal ) { *this = suOriginal; }
    inline SuCompassInitData& operator=( const SuCompassInitData& suOriginal );
};

inline SuCompassInitData& SuCompassInitData::operator=( const SuCompassInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        type = suOriginal.type;
        portString = suOriginal.portString;
    }

    return *this;
}

struct SuGyroInitData
{
    int type;
    string portString;

    SuGyroInitData() {}
    SuGyroInitData( const SuGyroInitData& suOriginal ) { *this = suOriginal; }
    inline SuGyroInitData& operator=( const SuGyroInitData& suOriginal );
};

inline SuGyroInitData& SuGyroInitData::operator=( const SuGyroInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        type = suOriginal.type;
        portString = suOriginal.portString;
    }

    return *this;
}

struct SuPoseCalcInitData
{
    int fuserType;
    int varianceScheme;
    double compass2MlabHeadingFacor;
    double compass2MlabHeadingOffset;
    double compassMaxValidAngSpeed4Heading;
    double gps2MlabHeadingFacor;
    double gps2MlabHeadingOffset;
    double gpsMinValidTransSpeed4Heading;
    double gpsMaxValidAngSpeed4Heading;

    SuPoseCalcInitData() {}
    SuPoseCalcInitData( const SuPoseCalcInitData& suOriginal ) { *this = suOriginal; }
    inline SuPoseCalcInitData& operator=( const SuPoseCalcInitData& suOriginal );
};

inline SuPoseCalcInitData& SuPoseCalcInitData::operator=( const SuPoseCalcInitData& suOriginal )
{
    if ( this != &suOriginal )
    {
        fuserType = suOriginal.fuserType;
        varianceScheme = suOriginal.varianceScheme;
        compass2MlabHeadingFacor = suOriginal.compass2MlabHeadingFacor;
        compass2MlabHeadingOffset = suOriginal.compass2MlabHeadingOffset;
        compassMaxValidAngSpeed4Heading = suOriginal.compassMaxValidAngSpeed4Heading;
        gps2MlabHeadingFacor = suOriginal.gps2MlabHeadingFacor;
        gps2MlabHeadingOffset = suOriginal.gps2MlabHeadingOffset;
        gpsMinValidTransSpeed4Heading = suOriginal.gpsMinValidTransSpeed4Heading;
        gpsMaxValidAngSpeed4Heading = suOriginal.gpsMaxValidAngSpeed4Heading;
    }

    return *this;
}

class ClHServerConfigurator
{
public:
    ClHServerConfigurator( ClHServerConfig& clConfig );
    ~ClHServerConfigurator();

    SuCognachromeInitData GetCognachromeOptions( int iSection );
    SuCameraInitData GetCameraOptions( int iSection );
    SuNomadInitData GetNomadOptions( int iSection );
    SuPioneerInitData GetPioneerOptions( int iSection );
    SuPioneer2InitData GetPioneer2Options( int iSection );
    SuAmigoBotInitData GetAmigoBotOptions( int iSection );
    SuFrameGrabberInitData GetFrameGrabberOptions( int iSection );
    SuGpsInitData GetGpsOptions( int iSection );
    SuCricketInitData GetCricketOptions( int iSection );
    SuJboxInitData GetJboxOptions( int iSection );
    SuCompassInitData GetCompassOptions( int iSection );
    SuGyroInitData GetGyroOptions( int iSection );
    SuPoseCalcInitData GetPoseCalcOptions( int iSection );
    SuLaserInitData GetLaserOptions( int iSection );
    SuRobotInitData GetRobotOptions(int iSection);
    SuGazeboInitData GetGazeboOptions(int iSection);
    SuUSARSimInitData GetUSARSimOptions(int iSection);
    SuRovioInitData GetRovioOptions(int iSection);
    SuSkybotixInitData GetSkybotixOptions(int iSection);
    SuOpenCVInitData GetOpenCVOptions(int iSection);

protected:
    ClHServerConfig* m_pclConfig;
    static const double HSCONFIG_HUGE_VAL_;
};

#endif // _HSERVERCONFIGURATOR_H

/**********************************************************************
# $Log: HServerConfigurator.h,v $
# Revision 1.10  2010-05-03 22:12:57  zkira
# Addition of rovio & cricket
#
# Revision 1.8  2009/04/17 17:50:05  zkira
# Added max_speed to differentiate it from max_steer_speed
#
# Revision 1.7  2009/04/13 02:58:54  zkira
# max altitude speed & other robot parameters
#
# Revision 1.6  2009/03/02 18:47:07  zkira
# Added ability to change low level robot parameters (e.g. max speed) in the .hserverrc file.
# Added Rovio
#
# Revision 1.5  2009/01/27 20:21:17  zkira
# Added gazebo laser_pitch parameter
#
# Revision 1.4  2008/11/20 03:53:57  ts2883
# The pitch angle information for the laser can now be parsed and read by the
# hserver using .hserverrc file.
#
# Revision 1.3  2008/09/16 16:47:38  zkira
# Added Gazebo robot class, clientID, and serverID to configuration file
#
# Revision 1.2  2008/07/15 22:34:01  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.10  2008/04/13 00:33:42  endo
# Regular backup.
#
# Revision 1.9  2008/04/12 20:17:24  endo
# OpenCV face detection implemented.
#
# Revision 1.8  2008/01/13 00:35:40  endo
# Regular backup.
#
# Revision 1.7  2008/01/09 22:54:03  endo
# USARSimRobot added.
#
# Revision 1.6  2006/12/01 21:14:54  endo
# Regular backup.
#
# Revision 1.5  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.4  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
# Revision 1.3  2006/11/15 23:45:59  endo
# GazeboLaser can now accept the second laser.
#
# Revision 1.2  2006/08/08 21:33:49  endo
# GazeboLaser added.
#
# Revision 1.1.1.1  2006/07/21 17:42:17  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:57  endo
# MissionLab 7.0
#
# Revision 1.2  2006/06/05 21:35:54  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.1.1.1  2005/02/06 22:59:46  endo
# AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
#
# Revision 1.3  2003/04/26 18:31:06  ebeowulf
# Checked in Pioneer2-dxe module.
#
# Revision 1.2  2002/02/18 13:52:47  endo
# AmigoBot added.
#
#
#**********************************************************************/
