/**********************************************************************
 **                                                                  **
 **                         opencv_handler.c                         **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: opencv_handler.c,v 1.1 2008-07-15 22:33:51 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <vector>

#include "opencv_handler.h"
#include "hserver.h"
#include "message.h"

using std::vector;

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------
OpenCVHandler *gOpenCVHandler = NULL;

//-----------------------------------------------------------------------
// Constatns
//-----------------------------------------------------------------------
const OpenCVHandler::RGBColor255_t OpenCVHandler::FACE_DETECT_DISPLAY_COLOR_ = {255, 0, 0};
const OpenCVHandler::RGBColor255_t OpenCVHandler::OBJECTS_DETECT_DISPLAY_COLOR_[NUM_OBJECTS_DETECT_OBJECT_TYPES] = 
{
    {0, 255, 0}, // Baby (Green)
    {0, 255, 255}, // Construction (Cyan)
    {255, 0, 255}, // High-Voltage (Magenta)
    {255, 255, 0} // Arc-Flash (Yellow)
};
const struct timespec OpenCVHandler::HANDLER_SLEEPTIME_NSEC_ = {0, 10000000}; // 10 msec.
const struct timespec OpenCVHandler::DISPLAY_SLEEPTIME_NSEC_ = {0, 100000000}; // 100 msec.
const string OpenCVHandler::IMAGE_PROCESS_TYPE_NAMES_[NUM_IMAGE_PROCESS_TYPES] =
{
    "BlobTrack",
    "FaceDetect",
    "LetterRecog",
    "ObjectsDetect"
};
const string OpenCVHandler::FG_DETECTOR_NICKNAMES_[NUM_FG_DETECTOR_TYPES] =
{
    "FG_DETECTOR_ACM_MM2003",
    "FG_DETECTOR_ACM_MM2003_SIMPLE",
    "FG_DETECTOR_CVPR1999"
};
const string OpenCVHandler::BLOB_DETECTOR_NICKNAMES_[NUM_BLOB_DETECTOR_TYPES] =
{
    "BLOB_DETECTOR_CC",
    "BLOB_DETECTOR_SIMPLE"
};
const string OpenCVHandler::BLOB_TRACKER_NICKNAMES_[NUM_BLOB_TRACKER_TYPES] =
{
    "BLOB_TRACKER_CCMSPF",
    "BLOB_TRACKER_CC",
    "BLOB_TRACKER_MS",
    "BLOB_TRACKER_MSFG",
    "BLOB_TRACKER_MSPF"
};
const string OpenCVHandler::BLOB_TRAJECTORY_ANALYZER_NICKNAMES_[NUM_BLOB_TRAJECTORY_ANALYZER_TYPES] =
{
    "BLOB_TRAJECTORY_ANALYZER_NONE",
    "BLOB_TRAJECTORY_ANALYZER_HISTPVS",
    "BLOB_TRAJECTORY_ANALYZER_HISTP",
    "BLOB_TRAJECTORY_ANALYZER_HISTPV",
    "BLOB_TRAJECTORY_ANALYZER_HISTSS",
    "BLOB_TRAJECTORY_ANALYZER_TRACKDIST",
    "BLOB_TRAJECTORY_ANALYZER_IOR"
};
//const string OpenCVHandler::FACE_DETECT_DATA_XML_FILE_ = "/usr/local/opencv/opencv-1.0.0/data/haarcascades/haarcascade_frontalface_alt.xml";
const string OpenCVHandler::FACE_DETECT_DATA_XML_FILE_ = "/home/endo/9000/HServer/src/hserver/opencv-data/baby-doll.xml";
const string OpenCVHandler::FACE_DETECT_LABEL_ = "Face";
const string OpenCVHandler::OBJECTS_DETECT_DATA_XML_FILES_[NUM_OBJECTS_DETECT_OBJECT_TYPES] =
{
    "/home/endo/9000/HServer/src/hserver/opencv-data/baby-doll.xml",
    "/home/endo/9000/HServer/src/hserver/opencv-data/construction.xml",
    "/home/endo/9000/HServer/src/hserver/opencv-data/high-voltage.xml",
    "/home/endo/9000/HServer/src/hserver/opencv-data/arc-flash.xml"
};
const string OpenCVHandler::OBJECTS_DETECT_LABELS_[NUM_OBJECTS_DETECT_OBJECT_TYPES] =
{
    "Baby",
    "Construction",
    "High-Voltage",
    "Arc-Flash"
};
const char OpenCVHandler::RGB_CHARS_[NUM_RGB_COLOR_TYPES] = {'R', 'G', 'B'};
const double OpenCVHandler::BLOB_TRACKER_RESET_TIME_SEC_ = 3.0;
const double OpenCVHandler::DEFAULT_FIELD_OF_VIEW_ = 60.0;
const float OpenCVHandler::DEFAULT_FONT_HSCALE_ = 0.85;
const float OpenCVHandler::DEFAULT_FONT_VSCALE_ = 0.85;
const float OpenCVHandler::DEFAULT_FONT_ITALICSCALE_ = 0;
const float OpenCVHandler::DEFAULT_OBJECTS_DETECT_READING_VALUE_ = 1.0;
const float OpenCVHandler::FACE_DETECT_SCALE_FACTOR_ = 1.1;
const float OpenCVHandler::OBJECTS_DETECT_SCALE_FACTORS_[NUM_OBJECTS_DETECT_OBJECT_TYPES] = 
{
    1.10, // Baby
    1.01, // Construction
    1.01, // High-Voltage
    1.10, // Arc-Flash
};
const int OpenCVHandler::SKIP_STATUSBAR_UPDATE_ = 5;
const int OpenCVHandler::WATCHDOG_CHECK_INTERVAL_SEC_ = 10;
const int OpenCVHandler::DEFAULT_DIRECT_GRAB_CAM_INDEX_ = 0;
const int OpenCVHandler::DEFAULT_DIRECT_GRAB_HEIGHT_ = 240;
const int OpenCVHandler::DEFAULT_DIRECT_GRAB_WIDTH_ = 320;
const int OpenCVHandler::DEFAULT_FG_TRAIN_FRAMES_ = 0;
//const int OpenCVHandler::DEFAULT_FG_DETECTOR_TYPE_ = FG_DETECTOR_ACM_MM2003;
const int OpenCVHandler::DEFAULT_FG_DETECTOR_TYPE_ = FG_DETECTOR_ACM_MM2003_SIMPLE;
//const int OpenCVHandler::DEFAULT_BLOB_DETECTOR_TYPE_ = BLOB_DETECTOR_CC;
const int OpenCVHandler::DEFAULT_BLOB_DETECTOR_TYPE_ = BLOB_DETECTOR_SIMPLE;
//const int OpenCVHandler::DEFAULT_BLOB_TRACKER_TYPE_ = BLOB_TRACKER_CCMSPF;
const int OpenCVHandler::DEFAULT_BLOB_TRACKER_TYPE_ = BLOB_TRACKER_CC;
const int OpenCVHandler::DEFAULT_BLOB_TRAJECTORY_ANALYZER_TYPE_ = BLOB_TRAJECTORY_ANALYZER_NONE;
const int OpenCVHandler::DEFAULT_FONT_THICKNESS_ = 1;
const int OpenCVHandler::DEFAULT_FONT_LINETYPE_ = CV_AA;
const int OpenCVHandler::FACE_DETECT_DISPLAY_THICKNESS_ = 2;
const int OpenCVHandler::FACE_DETECT_MIN_NUMBER_ = 2;
const int OpenCVHandler::FACE_DETECT_MIN_HEIGHT_ = 20;
const int OpenCVHandler::FACE_DETECT_MIN_WIDTH_ = 20;
const int OpenCVHandler::OBJECTS_DETECT_DISPLAY_THICKNESS_ = 2;
const int OpenCVHandler::OBJECTS_DETECT_MIN_NUMBER_ = 2;
const int OpenCVHandler::OBJECTS_DETECT_MIN_HEIGHTS_[NUM_OBJECTS_DETECT_OBJECT_TYPES] =
{
    10, // Baby
    40, // Construction
    40, // High-Voltage
    40, // Arc-Flash
};
const int OpenCVHandler::OBJECTS_DETECT_MIN_WIDTHS_[NUM_OBJECTS_DETECT_OBJECT_TYPES] =
{
    10, // Baby
    40, // Construction
    40, // High-Voltage
    40, // Arc-Flash
};
const int OpenCVHandler::MAX_REPORTING_BLOBS_ = 5;
const int OpenCVHandler::MAX_REPORTING_FACES_ = 10;
const int OpenCVHandler::MAX_REPORTING_OBJECTS_[NUM_OBJECTS_DETECT_OBJECT_TYPES] =
{
    2, // Baby
    1, // Construction
    1, // High-Voltage
    1, // Arc-Flash
};
const bool OpenCVHandler::OBJECTS_DETECT_REPORT_SIZE_ = false;
const bool OpenCVHandler::SORT_REPORTING_BLOB_SIZE_ = false;
const bool OpenCVHandler::SORT_REPORTING_FACE_SIZE_ = true;
const bool OpenCVHandler::SORT_REPORTING_0BJECTS_ = true;

//-----------------------------------------------------------------------
// Mutex initialization
//-----------------------------------------------------------------------
pthread_mutex_t OpenCVHandler::outputImagesMutex_[NUM_IMAGE_PROCESS_TYPES] =
{
    PTHREAD_MUTEX_INITIALIZER,
    PTHREAD_MUTEX_INITIALIZER,
    PTHREAD_MUTEX_INITIALIZER
};
pthread_mutex_t OpenCVHandler::sensorDataMutex_[NUM_OPENCV_SENSOR_TYPES] =
{
    PTHREAD_MUTEX_INITIALIZER,
    PTHREAD_MUTEX_INITIALIZER
};

//-----------------------------------------------------------------------
// Constructor
//-----------------------------------------------------------------------
OpenCVHandler::OpenCVHandler(
    OpenCVHandler **a,
    bool trackBlobs,
    bool detectFace,
    bool detectObjects,
    bool directGrab,
    bool showImage) :
    Module((Module**)a, HS_MODULE_NAME_OPENCV),
    rawImage_(NULL),
    blobTracker_(NULL),
    directGrabSource_(NULL),
    faceDetector_(NULL),
    objectDetectors_(NULL),
    faceMemStorage_(NULL),
    fieldOfView_(DEFAULT_FIELD_OF_VIEW_),
    rawImageTimeStamp_(0),
    blobSizeAngles_(NULL),
    blobSizeReadings_(NULL),
    blobColorAngles_(NULL),
    blobColorReadings_(NULL),
    faceSizeAngles_(NULL),
    faceSizeReadings_(NULL),
    directGrab_(directGrab),
    showImage_(showImage),
    handlerThreadIsUp_(false),
    displayThreadIsUp_(false),
    videoIsAvailable_(false)
{   
    double timeStamp;
    int i;
    int width, height, imageSize;
    unsigned char *image = NULL;

    // The image etc.
    for (i = 0; i < NUM_IMAGE_PROCESS_TYPES; i++)
    {
        pthread_mutex_init(&(outputImagesMutex_[i]), NULL);
        shOutputImages_[i] = NULL;
        xwindows_[i] = NULL;
    }

    // Initialize mutexes and the image etc.
    for (i = 0; i < NUM_OPENCV_SENSOR_TYPES; i++)
    {
        pthread_mutex_init(&(sensorDataMutex_[i]), NULL);
        sensors_[i] = NULL;
    }

    // Initialize other variables.
    memset(&blobTrackerParam_, 0x0, sizeof(CvBlobTrackerAutoParam1));
    memset(enableImageProcess_, 0x0, sizeof(bool)*NUM_IMAGE_PROCESS_TYPES);
    memset(resetImageProcess_, 0x0, sizeof(bool)*NUM_IMAGE_PROCESS_TYPES);
    memset(objectAngles_, 0x0, sizeof(float)*NUM_OBJECTS_DETECT_OBJECT_TYPES);
    memset(objectReadings_, 0x0, sizeof(float)*NUM_OBJECTS_DETECT_OBJECT_TYPES);

    // Check the video.
    if (directGrab_)
    {
        directGrabSource_ = cvCaptureFromCAM(DEFAULT_DIRECT_GRAB_CAM_INDEX_);

        if (directGrabSource_ == NULL)
        {
            printfTextWindow("Error: OpenCVHandler::OpenCVHandler(). directGrabSource_ is NULL.");
            delete this;
            a = NULL;
            return;
        }

        // Set the dimension. - Note: OpenCV does not support this yet.
        cvSetCaptureProperty(
            directGrabSource_,
            CV_CAP_PROP_FRAME_HEIGHT,
            (double)DEFAULT_DIRECT_GRAB_HEIGHT_);

        cvSetCaptureProperty(
            directGrabSource_,
            CV_CAP_PROP_FRAME_WIDTH,
            (double)DEFAULT_DIRECT_GRAB_WIDTH_);

        // Get the dimension.
        height = (int)cvGetCaptureProperty(
            directGrabSource_,
            CV_CAP_PROP_FRAME_HEIGHT);

        width = (int)cvGetCaptureProperty(
            directGrabSource_,
            CV_CAP_PROP_FRAME_WIDTH);
    }
    else if (gVideo == NULL)
    {
        printfTextWindow("Error: OpenCVHandler::OpenCVHandler(). gVideo is NULL.");
        delete this;
        a = NULL;
        return;
    }
    else
    {
        // Get the dimension.
        image = gVideo->duplicateCurrentImage(&width, &height, &imageSize, &timeStamp);
        delete [] image;
        image = NULL;
    }

    // Mark video as available.
    videoIsAvailable_ = true;

    // Initialize the raw image.
    initializeRawImage_(width, height);

    // Set the process types.
    enableImageProcess_[IMAGE_PROCESS_TYPE_BLOB_TRACK] = trackBlobs;
    enableImageProcess_[IMAGE_PROCESS_TYPE_FACE_DETECT] = detectFace;
    enableImageProcess_[IMAGE_PROCESS_TYPE_OBJECTS_DETECT] = detectObjects;

    // Update the status bar.
    updateStatusBar_();

    // Add sensors.
    addSensors_();

    // Start the handler thread.
    pthread_create(&handlerThread_, NULL, &startHandlerThread_, (void*)this);

    // Start the watchdog.
    if (gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(
            NAME_,
            WATCHDOG_CHECK_INTERVAL_SEC_,
            getpid(),
            handlerThread_);
    }

    if (showImage_)
    {   
        nanosleep(&DISPLAY_SLEEPTIME_NSEC_, NULL);

        // Start the display thread.
        pthread_create(&displayThread_, NULL, &startDisplayThread_, (void*)this);
    }

    refreshScreen();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
OpenCVHandler::~OpenCVHandler(void)
{
    int i;

    if (displayThreadIsUp_)
    {
        pthread_cancel(displayThread_);
        pthread_join(displayThread_, NULL);
        displayThreadIsUp_ = false;
    }

    if (handlerThreadIsUp_)
    {
        pthread_cancel(handlerThread_);
        pthread_join(handlerThread_, NULL);
        handlerThreadIsUp_ = false;
    }

    for (i = 0; i < NUM_IMAGE_PROCESS_TYPES; i++)
    {
        if (xwindows_[i] != NULL)
        {
            delete xwindows_[i];
            xwindows_[i] = NULL;
        }

        pthread_mutex_lock(&(outputImagesMutex_[i]));

        if (shOutputImages_[i] != NULL)
        {
            cvReleaseImage(&(shOutputImages_[i]));
            shOutputImages_[i] = NULL;
        }

        pthread_mutex_unlock(&(outputImagesMutex_[i]));
    }

    for (i = 0; i < NUM_OPENCV_SENSOR_TYPES; i++)
    {
        pthread_mutex_lock(&(sensorDataMutex_[i]));

        if (sensors_[i] != NULL)
        {
            delete sensors_[i];
            sensors_[i] = NULL;
        }

        pthread_mutex_unlock(&(sensorDataMutex_[i]));
    }

    if (rawImage_ != NULL)
    {
        cvReleaseImage(&rawImage_);
    }

    if(blobTrackerParam_.pBT != NULL)
    {
        cvReleaseBlobTracker(&(blobTrackerParam_.pBT));
    }

    if(blobTrackerParam_.pBD != NULL)
    {
        cvReleaseBlobDetector(&(blobTrackerParam_.pBD));
    }

    if(blobTrackerParam_.pBTGen != NULL)
    {
        cvReleaseBlobTrackGen(&(blobTrackerParam_.pBTGen));
    }

    if(blobTrackerParam_.pBTA != NULL)
    {
        cvReleaseBlobTrackAnalysis(&(blobTrackerParam_.pBTA));
    }

    if(blobTrackerParam_.pFG != NULL)
    {
        cvReleaseFGDetector(&(blobTrackerParam_.pFG));
    }

    if(blobTracker_ != NULL)
    {
        cvReleaseBlobTrackerAuto(&(blobTracker_));
    }

    if (blobSizeAngles_ != NULL)
    {
        delete [] blobSizeAngles_;
        blobSizeAngles_ = NULL;
    }

    if (blobSizeReadings_ != NULL)
    {
        delete [] blobSizeReadings_;
        blobSizeReadings_ = NULL;
    }

    if (blobColorAngles_ != NULL)
    {
        delete [] blobColorAngles_;
        blobColorAngles_ = NULL;
    }

    if (blobColorReadings_ != NULL)
    {
        delete [] blobColorReadings_;
        blobColorReadings_ = NULL;
    }

    if (faceSizeAngles_ != NULL)
    {
        delete [] faceSizeAngles_;
        faceSizeAngles_ = NULL;
    }

    if (faceSizeReadings_ != NULL)
    {
        delete [] faceSizeReadings_;
        faceSizeReadings_ = NULL;
    }

    if (directGrabSource_ != NULL)
    {
        cvReleaseCapture(&directGrabSource_);
        directGrabSource_ = NULL;
    }

    if (faceMemStorage_ != NULL)
    {
        cvReleaseMemStorage(&faceMemStorage_);
        faceMemStorage_ = NULL;
    }

    for (i = 0; i < NUM_OBJECTS_DETECT_OBJECT_TYPES; i++)
    {
        if (objectMemStorages_[i] != NULL)
        {
            cvReleaseMemStorage(&(objectMemStorages_[i]));
            objectMemStorages_[i] = NULL;
        }
    }

    for (i = 0; i < NUM_OBJECTS_DETECT_OBJECT_TYPES; i++)
    {
        if (objectAngles_[i] != NULL)
        {
            delete [] objectAngles_[i];
            objectAngles_[i] = NULL;
        }

        if (objectReadings_[i] != NULL)
        {
            delete [] objectReadings_[i];
            objectReadings_[i] = NULL;
        }
    }

   printfTextWindow("OpenCVHandler exited.");
}

//-----------------------------------------------------------------------
// This function initializes the raw image.
//-----------------------------------------------------------------------
bool OpenCVHandler::initializeRawImage_(int width, int height)
{
    if (rawImage_ != NULL)
    {
        cvReleaseImage(&rawImage_);
    }

    rawImage_ = cvCreateImage(
        cvSize(width, height),
        IPL_DEPTH_8U,
        NUM_RGB_COLOR_TYPES);

	rawImage_->imageData = NULL;

    if (rawImage_ == NULL)
    {
        return false;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function adds the sensors.
//-----------------------------------------------------------------------
void OpenCVHandler::addSensors_(void)
{
    int i, j, ohSensorType, hsSensorType;
    bool validSensor = true;

    for (i = 0; i < NUM_IMAGE_PROCESS_TYPES; i++)
    {
        if (enableImageProcess_[i])
        {
            switch (i) {

            case IMAGE_PROCESS_TYPE_BLOB_TRACK:

                // First, blob size sensor.
                if (blobSizeAngles_ != NULL)
                {
                    delete [] blobSizeAngles_;
                    blobSizeAngles_ = NULL;
                }

                if (blobSizeReadings_ != NULL)
                {
                    delete [] blobSizeReadings_;
                    blobSizeReadings_ = NULL;
                }

                blobSizeAngles_ = new float[MAX_REPORTING_BLOBS_];
                blobSizeReadings_ = new float[MAX_REPORTING_BLOBS_];
                sensorExtraBlobSize_.numBlobSizeTypes = 1;

                sensors_[OPENCV_SENSOR_BLOB_SIZE] = new Sensor(
                    SENSOR_OPENCV_BLOB_SIZE,
                    sizeof(SensorExtraBlobSize_t),
                    (char *)&sensorExtraBlobSize_,
                    0,
                    blobSizeAngles_,
                    0,
                    blobSizeReadings_,
                    &(sensorDataMutex_[OPENCV_SENSOR_BLOB_SIZE]));

                // Next, blob color sensor.
                if (blobColorAngles_ != NULL)
                {
                    delete [] blobColorAngles_;
                    blobColorAngles_ = NULL;
                }

                if (blobColorReadings_ != NULL)
                {
                    delete [] blobColorReadings_;
                    blobColorReadings_ = NULL;
                }

                blobColorAngles_ = new float[(MAX_REPORTING_BLOBS_ * NUM_RGB_COLOR_TYPES)];
                blobColorReadings_ = new float[(MAX_REPORTING_BLOBS_ * NUM_RGB_COLOR_TYPES)];
                sensorExtraBlobColor_.numBlobColorTypes = NUM_RGB_COLOR_TYPES;

                sensors_[OPENCV_SENSOR_BLOB_COLOR] = new Sensor(
                    SENSOR_OPENCV_BLOB_COLOR,
                    sizeof(SensorExtraBlobColor_t),
                    (char *)&sensorExtraBlobColor_,
                    0,
                    blobColorAngles_,
                    0,
                    blobColorReadings_,
                    &(sensorDataMutex_[OPENCV_SENSOR_BLOB_COLOR]));
                break;

            case IMAGE_PROCESS_TYPE_FACE_DETECT:

                if (faceSizeAngles_ != NULL)
                {
                    delete [] faceSizeAngles_;
                    faceSizeAngles_ = NULL;
                }

                if (faceSizeReadings_ != NULL)
                {
                    delete [] faceSizeReadings_;
                    faceSizeReadings_ = NULL;
                }

                faceSizeAngles_ = new float[MAX_REPORTING_FACES_];
                faceSizeReadings_ = new float[MAX_REPORTING_FACES_];
                sensorExtraFaceSize_.numFaceSizeTypes = 1;

                sensors_[OPENCV_SENSOR_FACE_SIZE] = new Sensor(
                    SENSOR_OPENCV_FACE_SIZE,
                    sizeof(SensorExtraFaceSize_t),
                    (char *)&sensorExtraFaceSize_,
                    0,
                    faceSizeAngles_,
                    0,
                    faceSizeReadings_,
                    &(sensorDataMutex_[OPENCV_SENSOR_FACE_SIZE]));
                break;

            case IMAGE_PROCESS_TYPE_LETTER_RECOG:
                break;

            case IMAGE_PROCESS_TYPE_OBJECTS_DETECT:

                for (j = 0; j < NUM_OBJECTS_DETECT_OBJECT_TYPES; j++)
                {
                    if (objectAngles_[j] != NULL)
                    {
                        delete [] objectAngles_[j];
                        objectAngles_[j] = NULL;
                    }

                    if (objectReadings_[j] != NULL)
                    {
                        delete [] objectReadings_[j];
                        objectReadings_[j] = NULL;
                    }

                    objectAngles_[j] = new float[MAX_REPORTING_OBJECTS_[j]];
                    objectReadings_[j] = new float[MAX_REPORTING_OBJECTS_[j]];
                    sensorExtraObject_.numObjectTypes = 1;
                    validSensor = true;

                    switch (j) {

                    case 0:
                        ohSensorType = OPENCV_SENSOR_OBJECT_001;
                        hsSensorType = SENSOR_OPENCV_OBJECT_001;
                        break;

                    case 1:
                        ohSensorType = OPENCV_SENSOR_OBJECT_002;
                        hsSensorType = SENSOR_OPENCV_OBJECT_002;
                        break;

                    case 2:
                        ohSensorType = OPENCV_SENSOR_OBJECT_003;
                        hsSensorType = SENSOR_OPENCV_OBJECT_003;
                        break;

                    case 3:
                        ohSensorType = OPENCV_SENSOR_OBJECT_004;
                        hsSensorType = SENSOR_OPENCV_OBJECT_004;
                        break;

                    default:
                        printfTextWindow("Warning: OpenCVHandler::addSensors_(). Unknown sensor type.");
                        validSensor = false;
                        break;
                    }

                    if (validSensor)
                    {
                        sensors_[ohSensorType] = new Sensor(
                            hsSensorType,
                            sizeof(SensorExtraObject_t),
                            (char *)&sensorExtraObject_,
                            0,
                            objectAngles_[j],
                            0,
                            objectReadings_[j],
                            &(sensorDataMutex_[hsSensorType]));
                    }
                }

                break;
            }
        }
    }
}

//-----------------------------------------------------------------------
// This function starts up the main thread.
//-----------------------------------------------------------------------
void *OpenCVHandler::startHandlerThread_(void *openCVHandlerInstance)
{
    ((OpenCVHandler*)openCVHandlerInstance)->handlerLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function starts up the display thread.
//-----------------------------------------------------------------------
void *OpenCVHandler::startDisplayThread_(void *openCVHandlerInstance)
{
    ((OpenCVHandler*)openCVHandlerInstance)->displayLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to reads data from Robot.
//-----------------------------------------------------------------------
void OpenCVHandler::handlerLoop_(void)
{
    int count = 0;

    handlerThreadIsUp_ = true;

    while(true)
    {
        // Make sure the thread is not exited.
        pthread_testcancel();

        // Process the image.
        processImage_();

        // Compute the frequency.
        compFrequency_();

        // Update the status bar.
        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        // Make sure the thread is not exited.
        pthread_testcancel();

        if (watchdog_ != NULL)
        {
            watchdog_->notifyUpdate();
        }

        // Sleep a bit.
        nanosleep(&HANDLER_SLEEPTIME_NSEC_, NULL);
    }
}

//-----------------------------------------------------------------------
// This function process the iamge.
//-----------------------------------------------------------------------
void OpenCVHandler::processImage_(void)
{
    double timeStamp;
    unsigned char *image = NULL;
    int width, height, imageSize;
    int i;

    if (directGrabSource_ != NULL)
    {
        rawImage_ = cvQueryFrame(directGrabSource_);
        rawImageTimeStamp_ = getCurrentEpochTime();
    }
    else if (gVideo != NULL)
    {
        // Get the latest image.
        image = gVideo->duplicateCurrentImage(&width, &height, &imageSize, &timeStamp);

        if (image == NULL)
        {
            return;
        }
        else if (timeStamp == rawImageTimeStamp_)
        {
            delete [] image;
            image = NULL;
            return;
        }

        rawImageTimeStamp_ = timeStamp;

        if ((rawImage_->imageData) != NULL)
        {
            delete [] rawImage_->imageData;
            rawImage_->imageData = NULL;
        }

        rawImage_->imageData = (char *)image;
    }
    else
    {
        videoIsAvailable_ = false;
        return;
    }

    for (i = 0; i < NUM_IMAGE_PROCESS_TYPES; i++)
    {
        if (!enableImageProcess_[i])
        {
            continue;
        }

        switch (i) {

        case IMAGE_PROCESS_TYPE_BLOB_TRACK:
            processBlobTrack_(rawImage_);
            break;

        case IMAGE_PROCESS_TYPE_FACE_DETECT:
            processFaceDetect_(rawImage_);
            break;

        case IMAGE_PROCESS_TYPE_LETTER_RECOG:
            break;

        case IMAGE_PROCESS_TYPE_OBJECTS_DETECT:
            processObjectsDetect_(rawImage_);
            break;
        }
    }
}

//-----------------------------------------------------------------------
// This function process blob tracking.
//-----------------------------------------------------------------------
void OpenCVHandler::processBlobTrack_(IplImage *rawImage)
{
    CvSize maskSize, blobSize, textSize;
    CvPoint pt;
    CvBlob* blob = NULL;
    CvScalar rgb;
    IplImage *blobTrackMask = NULL, *fgMask = NULL;
    string descString, bufString;
    string::size_type index;
    vector<float> blobSizeAngleVec;
    vector<float> blobSizeReadingVec;
    vector<float> blobColorAngleVec[NUM_RGB_COLOR_TYPES];
    vector<float> blobColorReadingVec[NUM_RGB_COLOR_TYPES];
    vector<float>::iterator iter, bestIter;
    char buf[1024];
    char *desc = NULL;
    double timeStamp;
    float blobSizeAngles[MAX_REPORTING_BLOBS_];
    float blobSizeReadings[MAX_REPORTING_BLOBS_];
    float blobColorAngles[(MAX_REPORTING_BLOBS_ * NUM_RGB_COLOR_TYPES)];
    float blobColorReadings[(MAX_REPORTING_BLOBS_ * NUM_RGB_COLOR_TYPES)];
    float angle, reading;
    float biggestBlobSize;
    int i, j, idx, bestIndex;
    int numBlobs;
    int c;
    const int INDEX = IMAGE_PROCESS_TYPE_BLOB_TRACK;

    if (rawImage == NULL)
    {
        return;
    }

    // Initialize the blob tracker if necessary.
    if ((blobTracker_ == NULL) || (resetImageProcess_[INDEX]))
    {
        if (!initializeBlobTracker_(rawImage))
        {
            enableImageProcess_[INDEX] = false;
        }

        return;
    }

    // Take a time stamp.
    timeStamp = getCurrentEpochTime();

    // Process the image.
    blobTracker_->Process(rawImage, blobTrackMask);

    pthread_mutex_lock(&(outputImagesMutex_[INDEX]));
        
    if (showImage_)
    {
        // Create the blob image.
        fgMask = blobTracker_->GetFGMask();

        if (fgMask != NULL)
        {
            maskSize = cvSize(fgMask->width, fgMask->height);

            if (shOutputImages_[INDEX] == NULL)
            {
                shOutputImages_[INDEX] = cvCreateImage(
                    maskSize,
                    fgMask->depth,
                    NUM_RGB_COLOR_TYPES);
            }

            cvCvtColor(fgMask, shOutputImages_[INDEX], CV_GRAY2BGR);
        }
    }

    // Draw detected blobs
    numBlobs = 0;
    for(i = blobTracker_->GetBlobNum(); i > 0; i--)
    {
        if (numBlobs >= MAX_REPORTING_BLOBS_)
        {
            break;
        }

        numBlobs++;
        idx = i - 1;

        // Get the blob.
        blob = blobTracker_->GetBlob(idx);
        pt = cvPointFrom32f(CV_BLOB_CENTER(blob));
        blobSize = cvSize(
            MAX(1, cvRound(CV_BLOB_RX(blob))),
            MAX(1, cvRound(CV_BLOB_RY(blob))));

        // Compute the angle.
        angle = fieldOfView_ * (0.5 - ((float)(pt.x)/(float)(rawImage->width)));
        //blobSizeAngles[idx] = angle;
        blobSizeAngleVec.push_back(angle);

        // Reading is a blob size with respect to the total image size.
        //blobSizeReadings[idx] =
        reading = 
            (float)((blobSize.width)*(blobSize.height)) /
            (float)((rawImage->width)*(rawImage->height));
        blobSizeReadingVec.push_back(reading);

        // Get the color.
        rgb = getBlobColor_(rawImage, blob);

        for (j = 0; j < NUM_RGB_COLOR_TYPES; j++)
        {
            blobColorAngleVec[j].push_back(angle);

            reading = rgb.val[j]/255.0;
            blobColorReadingVec[j].push_back(reading);
        }

        if ((showImage_) && (fgMask != NULL))
        {
            c = cvRound(255*(blobTracker_->GetState(CV_BLOB_ID(blob))));

            // Draw a filled ellipse with the color of the blob.
            cvEllipse(
                shOutputImages_[INDEX],
                pt,
                blobSize,
                0,
                0,
                360,
                rgb,
                CV_FILLED);

            // Draw a ellipse outline.
            cvEllipse(
                shOutputImages_[INDEX],
                pt,
                blobSize,
                0,
                0,
                360,
                CV_RGB(c,255-c,0),
                cvRound(1+(NUM_RGB_COLOR_TYPES*c)/255));

            // Display ID.
            sprintf(buf, "ID: %03d", CV_BLOB_ID(blob));

            cvGetTextSize(buf, &blobTrackerFont_, &textSize, NULL);
            pt.y -= blobSize.height;

            cvPutText(
                shOutputImages_[INDEX],
                buf,
                pt,
                &blobTrackerFont_,
                CV_RGB(0,255,255));

            // Display color values.
            for (j = 0; j < NUM_RGB_COLOR_TYPES; j++)
            {
                sprintf(buf, "%c: %.2f", RGB_CHARS_[j], rgb.val[j]/255.0);

                pt.y += (textSize.height + 2);

                cvPutText(
                    shOutputImages_[INDEX],
                    buf,
                    pt,
                    &blobTrackerFont_,
                    CV_RGB(0,255,255));
            }

            // Display size.
            sprintf(buf, "Size: %.3f", blobSizeReadingVec.back());

            pt.y += (textSize.height + 2);

            cvPutText(
                shOutputImages_[INDEX],
                buf,
                pt,
                &blobTrackerFont_,
                CV_RGB(0,255,255));

            desc = blobTracker_->GetStateDesc(CV_BLOB_ID(blob));

            if (desc == NULL)
            {
                continue;
            }

            descString = desc;

            while ((int)(descString.size()) > 0)
            {
                index = descString.find('\n');

                if (index != string::npos)
                {
                    bufString = descString.substr(0, index);
                    descString = descString.substr(index+1, string::npos);
                }
                else
                {
                    bufString = descString;
                    descString.clear();
                }
                
                pt.y += (textSize.height + 2);

                cvPutText(
                    shOutputImages_[INDEX],
                    (char *)bufString.c_str(),
                    pt,
                    &blobTrackerFont_,
                    CV_RGB(0,255,255));
            }
        }
    }

    pthread_mutex_unlock(&(outputImagesMutex_[INDEX]));

    idx = 0;

    while (!(blobSizeReadingVec.empty()))
    {
        bestIndex = 0;

        if (SORT_REPORTING_BLOB_SIZE_)
        {
            // Sort the data based on the blob size.
            biggestBlobSize = 0;

            for(iter = blobSizeReadingVec.begin(); iter < blobSizeReadingVec.end(); iter++)
            {
                if (*iter >= biggestBlobSize)
                {
                    biggestBlobSize = *iter;
                    bestIter = iter;
                    bestIndex = bestIter - blobSizeReadingVec.begin();
                }
            }
        }
        else
        {
            bestIter = blobSizeReadingVec.begin();
        }

        // Copy the blob size.
        blobSizeReadings[idx] = *bestIter;
        blobSizeReadingVec.erase(bestIter);

        // Copy the blob size angle.
        bestIter = blobSizeAngleVec.begin() + bestIndex;
        blobSizeAngles[idx] = *bestIter;
        blobSizeAngleVec.erase(bestIter);

        for (j = 0; j < NUM_RGB_COLOR_TYPES; j++)
        {
            // Copy the blob color values.
            bestIter = blobColorReadingVec[j].begin() + bestIndex;
            blobColorReadings[(idx*NUM_RGB_COLOR_TYPES)+j] = *bestIter;
            blobColorReadingVec[j].erase(bestIter);

            // Copy the blob color angles.
            bestIter = blobColorAngleVec[j].begin() + bestIndex;
            blobColorAngles[(idx*NUM_RGB_COLOR_TYPES)+j] = *bestIter;
            blobColorAngleVec[j].erase(bestIter);
        }

        idx++;
    }

    // Copy to the blob size sensor data.
    pthread_mutex_lock(&(sensorDataMutex_[OPENCV_SENSOR_BLOB_SIZE]));

    memcpy(blobSizeAngles_, blobSizeAngles, sizeof(float)*numBlobs);
    memcpy(blobSizeReadings_, blobSizeReadings, sizeof(float)*numBlobs);

    if (sensors_[OPENCV_SENSOR_BLOB_SIZE] != NULL)
    {
        sensors_[OPENCV_SENSOR_BLOB_SIZE]->setAngleInfo(blobSizeAngles_, numBlobs);
        sensors_[OPENCV_SENSOR_BLOB_SIZE]->setReadingInfo(blobSizeReadings_, numBlobs);
    }

    pthread_mutex_unlock(&(sensorDataMutex_[OPENCV_SENSOR_BLOB_SIZE]));

    // Copy to the blob color sensor data.
    pthread_mutex_lock(&(sensorDataMutex_[OPENCV_SENSOR_BLOB_COLOR]));

    memcpy(
        blobColorAngles_,
        blobColorAngles,
        sizeof(float)*numBlobs*NUM_RGB_COLOR_TYPES);
    memcpy(
        blobColorReadings_,
        blobColorReadings,
        sizeof(float)*numBlobs*NUM_RGB_COLOR_TYPES);

    if (sensors_[OPENCV_SENSOR_BLOB_COLOR] != NULL)
    {
        sensors_[OPENCV_SENSOR_BLOB_COLOR]->setAngleInfo(
            blobColorAngles_,
            numBlobs*NUM_RGB_COLOR_TYPES);
        sensors_[OPENCV_SENSOR_BLOB_COLOR]->setReadingInfo(
            blobColorReadings_,
            numBlobs*NUM_RGB_COLOR_TYPES);
    }

    pthread_mutex_unlock(&(sensorDataMutex_[OPENCV_SENSOR_BLOB_COLOR]));

    if ((getCurrentEpochTime() - timeStamp) > BLOB_TRACKER_RESET_TIME_SEC_)
    {
        // It tool too long to process the image. Tracker needs to be reset.
        resetImageProcess_[INDEX] = true;
    }
}

//-----------------------------------------------------------------------
// This function process face detection.
//-----------------------------------------------------------------------
void OpenCVHandler::processFaceDetect_(IplImage *rawImage)
{
    CvPoint pt, pt1, pt2;
    CvRect *rect = NULL;
    CvSeq *faces = NULL;
    CvSize textSize;
    string descString, bufString;
    vector<CvPoint> pts1, pts2;
    vector<float> faceSizeAngleVec;
    vector<float> faceSizeReadingVec;
    vector<float>::iterator iter, bestIter;
    char buf[1024];
    double timeStamp;
    float faceSizeAngles[MAX_REPORTING_FACES_];
    float faceSizeReadings[MAX_REPORTING_FACES_];
    float angle, reading;
    float biggestFaceSize;
    int i, idx, bestIndex, numFaces = 0;
    const int INDEX = IMAGE_PROCESS_TYPE_FACE_DETECT;

    if (rawImage == NULL)
    {
        return;
    }

    // Initialize the faceDetector if necessary.
    if ((faceDetector_ == NULL) || (resetImageProcess_[INDEX]))
    {
        if (!initializeFaceDetector_(rawImage))
        {
            enableImageProcess_[INDEX] = false;
        }

        return;
    }

    // Take a time stamp.
    timeStamp = getCurrentEpochTime();

 	// "Resets" the memory but does not deallocate it.
	cvClearMemStorage(faceMemStorage_);

	// Run the main object recognition function.  The arguments are: 
	// 1. the image to use
	// 2. the pre-trained Haar classifier cascade data
	// 3. memory storage for rectangles around recognized objects
	// 4. a scale factor "by which the search window is scaled between the 
	//    subsequent scans, for example, 1.1 means increasing window by 10%"
	// 5. the "minimum number (minus 1) of neighbor rectangles that makes up 
	//    an object. All the groups of a smaller number of rectangles than 
	//    min_neighbors-1 are rejected. If min_neighbors is 0, the function 
	//    does not any grouping at all and returns all the detected candidate 
	//    rectangles, which may be useful if the user wants to apply a 
	//    customized grouping procedure."
	// 6. flags which determine the mode of operation
	// 7. the minimum object size (if possible, increasing this will 
	//    really speed up the process)
    faces = cvHaarDetectObjects(
        rawImage,
        faceDetector_,
        faceMemStorage_,
        FACE_DETECT_SCALE_FACTOR_, 
		FACE_DETECT_MIN_NUMBER_,
        CV_HAAR_DO_CANNY_PRUNING,
        cvSize(FACE_DETECT_MIN_WIDTH_, FACE_DETECT_MIN_HEIGHT_));

    // If any faces were detected, draw rectangles around them.
    if (faces != NULL)
    {
        numFaces = faces->total;

        for(i = 0; i < numFaces; i++)
        {
            // Setup two points that define the extremes of the rectangle, 
            // then draw it to the image..
            rect = (CvRect*)cvGetSeqElem(faces, i);
            pt1.x = rect->x;
            pt2.x = (rect->x + rect->width);
            pt1.y = rect->y;
            pt2.y = (rect->y + rect->height);

            // Remember.
            pts1.push_back(pt1);
            pts2.push_back(pt2);

            // Find the center.
            pt.x = (pt1.x + pt2.x)/2;
            pt.y = (pt1.y + pt2.y)/2;

            // Compute the angle.
            angle = fieldOfView_ * (0.5 - ((float)(pt.x)/(float)(rawImage->width)));
            faceSizeAngleVec.push_back(angle);

            // Reading is a face size with respect to the total image size.
            reading = 
                (float)((rect->width)*(rect->height)) /
                (float)((rawImage->width)*(rawImage->height));
            faceSizeReadingVec.push_back(reading);
        }
    }

    pthread_mutex_lock(&(outputImagesMutex_[INDEX]));
        
    if (showImage_)
    {
        cvCopyImage(rawImage, shOutputImages_[INDEX]);

        // If any faces were detected, draw rectangles around them.
        if (numFaces > 0)
        {
            for(i = 0; i < numFaces; i++)
            {
                // Setup two points that define the extremes of the rectangle, 
                // then draw it to the image..
                cvRectangle(
                    shOutputImages_[INDEX],
                    pts1[i],
                    pts2[i],
                    CV_RGB(255,0,0),
                    FACE_DETECT_DISPLAY_THICKNESS_);

                    // Label it.
                    pt.x = (pts1[i].x + pts2[i].x)/2;
                    pt.y = pts1[i].y;

                    sprintf(buf, "%s %d", FACE_DETECT_LABEL_.c_str(), (i+1));
                    cvGetTextSize(buf, &faceDetectorFont_, &textSize, NULL);
                    pt.x -= (textSize.width/2);
                    pt.y -= textSize.height;

                    cvPutText(
                        shOutputImages_[INDEX],
                        buf,
                        pt,
                        &faceDetectorFont_,
                        CV_RGB(
                            FACE_DETECT_DISPLAY_COLOR_.red,
                            FACE_DETECT_DISPLAY_COLOR_.green,
                            FACE_DETECT_DISPLAY_COLOR_.blue));
            }
        }
    }

    pthread_mutex_unlock(&(outputImagesMutex_[INDEX]));

    idx = 0;

    while (!(faceSizeReadingVec.empty()))
    {
        bestIndex = 0;

        if (SORT_REPORTING_FACE_SIZE_)
        {
            // Sort the data based on the face size.
            biggestFaceSize = 0;

            for(iter = faceSizeReadingVec.begin(); iter < faceSizeReadingVec.end(); iter++)
            {
                if (*iter >= biggestFaceSize)
                {
                    biggestFaceSize = *iter;
                    bestIter = iter;
                    bestIndex = bestIter - faceSizeReadingVec.begin();
                }
            }
        }
        else
        {
            bestIter = faceSizeReadingVec.begin();
        }

        // Copy the face size.
        faceSizeReadings[idx] = *bestIter;
        faceSizeReadingVec.erase(bestIter);

        // Copy the face size angle.
        bestIter = faceSizeAngleVec.begin() + bestIndex;
        faceSizeAngles[idx] = *bestIter;
        faceSizeAngleVec.erase(bestIter);

        idx++;

        if (idx >= MAX_REPORTING_FACES_)
        {
            numFaces = idx;
            break;
        }
    }

    // Copy to the face size sensor data.
    pthread_mutex_lock(&(sensorDataMutex_[OPENCV_SENSOR_FACE_SIZE]));

    memcpy(faceSizeAngles_, faceSizeAngles, sizeof(float)*numFaces);
    memcpy(faceSizeReadings_, faceSizeReadings, sizeof(float)*numFaces);

    if (sensors_[OPENCV_SENSOR_FACE_SIZE] != NULL)
    {
        sensors_[OPENCV_SENSOR_FACE_SIZE]->setAngleInfo(faceSizeAngles_, numFaces);
        sensors_[OPENCV_SENSOR_FACE_SIZE]->setReadingInfo(faceSizeReadings_, numFaces);
    }

    pthread_mutex_unlock(&(sensorDataMutex_[OPENCV_SENSOR_FACE_SIZE]));
}

//-----------------------------------------------------------------------
// This function process objects detection.
//-----------------------------------------------------------------------
void OpenCVHandler::processObjectsDetect_(IplImage *rawImage)
{
    CvPoint pt, pt1, pt2;
    CvRect *rect = NULL;
    CvSeq *objectsList[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    CvSize textSize;
    string descString, bufString;
    vector<CvPoint> pts1[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    vector<CvPoint> pts2[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    vector<float> objectAngleVec[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    vector<float> objectReadingVec[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    vector<float>::iterator iter, bestIter;
    char buf[1024];
    double timeStamp;
    float *objectAngles = NULL;
    float *objectReadings = NULL;
    float angle, reading;
    float biggestObject;
    int i, j, k, idx, bestIndex;
    int ohSensorType;
    int numObjectsList[NUM_OBJECTS_DETECT_OBJECT_TYPES];
    bool validSensor = true, validPoint = false;
    const int INDEX = IMAGE_PROCESS_TYPE_OBJECTS_DETECT;

    if (rawImage == NULL)
    {
        return;
    }

    // Initialize the objectsDetector if necessary.
    if ((objectDetectors_ == NULL) || (resetImageProcess_[INDEX]))
    {
        if (!initializeObjectDetectors_(rawImage))
        {
            enableImageProcess_[INDEX] = false;
        }

        return;
    }

    // Take a time stamp.
    timeStamp = getCurrentEpochTime();

    // Initialize some variables.
    for (i = 0; i < NUM_OBJECTS_DETECT_OBJECT_TYPES; i++)
    {
        // Reset the number of objects.
        numObjectsList[i] = 0;

        // "Resets" the memory but does not deallocate it.
        cvClearMemStorage(objectMemStorages_[i]);

        // Run the main object recognition function.  The arguments are: 
        // 1. the image to use
        // 2. the pre-trained Haar classifier cascade data
        // 3. memory storage for rectangles around recognized objects
        // 4. a scale factor "by which the search window is scaled between the 
        //    subsequent scans, for example, 1.1 means increasing window by 10%"
        // 5. the "minimum number (minus 1) of neighbor rectangles that makes up 
        //    an object. All the groups of a smaller number of rectangles than 
        //    min_neighbors-1 are rejected. If min_neighbors is 0, the function 
        //    does not any grouping at all and returns all the detected candidate 
        //    rectangles, which may be useful if the user wants to apply a 
        //    customized grouping procedure."
        // 6. flags which determine the mode of operation
        // 7. the minimum object size (if possible, increasing this will 
        //    really speed up the process)
        objectsList[i] = cvHaarDetectObjects(
            rawImage,
            objectDetectors_[i],
            objectMemStorages_[i],
            OBJECTS_DETECT_SCALE_FACTORS_[i], 
            OBJECTS_DETECT_MIN_NUMBER_,
            CV_HAAR_DO_CANNY_PRUNING,
            cvSize(OBJECTS_DETECT_MIN_WIDTHS_[i], OBJECTS_DETECT_MIN_HEIGHTS_[i]));

        if (objectsList[i] != NULL)
        {
            numObjectsList[i] = objectsList[i]->total;
            validPoint = true;

            for(j = 0; j < objectsList[i]->total; j++)
            {
                // Setup two points that define the extremes of the rectangle, 
                // then draw it to the image..
                rect = (CvRect*)cvGetSeqElem(objectsList[i], j);
                pt1.x = rect->x;
                pt2.x = (rect->x + rect->width);
                pt1.y = rect->y;
                pt2.y = (rect->y + rect->height);

                // Find the center.
                pt.x = (pt1.x + pt2.x)/2;
                pt.y = (pt1.y + pt2.y)/2;

                // Check occlusion.
                for (k = 0; k < (int)(pts1[i].size()); k++)
                {
                    if ((((pt.x) >= (pts1[i][k].x)) && ((pt.x) <= (pts2[i][k].x))) &&
                        (((pt.y) >= (pts1[i][k].y)) && ((pt.y) <= (pts2[i][k].y))))
                    {
                        // Occlusion. Exclude it from the list.
                        numObjectsList[i]--;
                        validPoint = false;
                        break;
                    }
                }

                if (validPoint)
                {
                    // Remember.
                    pts1[i].push_back(pt1);
                    pts2[i].push_back(pt2);

                    // Compute the angle.
                    angle = fieldOfView_ * (0.5 - ((float)(pt.x)/(float)(rawImage->width)));
                    objectAngleVec[i].push_back(angle);

                    if (OBJECTS_DETECT_REPORT_SIZE_)
                    {
                        // Reading is a object size with respect to the total image size.
                        reading = 
                            (float)((rect->width)*(rect->height)) /
                            (float)((rawImage->width)*(rawImage->height));
                        objectReadingVec[i].push_back(reading);
                    }
                    else
                    {
                        objectReadingVec[i].push_back(
                            DEFAULT_OBJECTS_DETECT_READING_VALUE_);
                    }
                }
            }
        }
    }

    pthread_mutex_lock(&(outputImagesMutex_[INDEX]));
        
    if (showImage_)
    {
        cvCopyImage(rawImage, shOutputImages_[INDEX]);

        // If any object were detected, draw rectangles around them.
        for (i = 0; i < NUM_OBJECTS_DETECT_OBJECT_TYPES; i++)
        {
            if (numObjectsList[i] > 0)
            {
                for(j = 0; j < numObjectsList[i]; j++)
                {
                    // Setup two points that define the extremes of the rectangle, 
                    // then draw it to the image..
                    cvRectangle(
                        shOutputImages_[INDEX],
                        pts1[i][j],
                        pts2[i][j],
                        CV_RGB(
                            OBJECTS_DETECT_DISPLAY_COLOR_[i].red,
                            OBJECTS_DETECT_DISPLAY_COLOR_[i].green,
                            OBJECTS_DETECT_DISPLAY_COLOR_[i].blue),
                        OBJECTS_DETECT_DISPLAY_THICKNESS_);

                    // Label it.
                    pt.x = (pts1[i][j].x + pts2[i][j].x)/2;
                    pt.y = pts1[i][j].y;

                    sprintf(buf, "%s %d", OBJECTS_DETECT_LABELS_[i].c_str(), (j+1));
                    cvGetTextSize(buf, &objectDetectorsFont_, &textSize, NULL);
                    pt.x -= (textSize.width/2);
                    pt.y -= textSize.height;

                    cvPutText(
                        shOutputImages_[INDEX],
                        buf,
                        pt,
                        &objectDetectorsFont_,
                        CV_RGB(
                            OBJECTS_DETECT_DISPLAY_COLOR_[i].red,
                            OBJECTS_DETECT_DISPLAY_COLOR_[i].green,
                            OBJECTS_DETECT_DISPLAY_COLOR_[i].blue));
                }
            }
        }
    }

    pthread_mutex_unlock(&(outputImagesMutex_[INDEX]));

    for (i = 0; i < NUM_OBJECTS_DETECT_OBJECT_TYPES; i++)
    {
        idx = 0;
        objectAngles = new float[MAX_REPORTING_OBJECTS_[i]];
        objectReadings = new float[MAX_REPORTING_OBJECTS_[i]];

        while (!(objectReadingVec[i].empty()))
        {
            bestIndex = 0;

            if (SORT_REPORTING_0BJECTS_)
            {
                // Sort the data based on the object size.
                biggestObject = 0;

                for(iter = objectReadingVec[i].begin(); iter < objectReadingVec[i].end(); iter++)
                {
                    if (*iter >= biggestObject)
                    {
                        biggestObject = *iter;
                        bestIter = iter;
                        bestIndex = bestIter - objectReadingVec[i].begin();
                    }
                }
            }
            else
            {
                bestIter = objectReadingVec[i].begin();
            }

            // Copy the object size.
            objectReadings[idx] = *bestIter;
            objectReadingVec[i].erase(bestIter);

            // Copy the object angle.
            bestIter = objectAngleVec[i].begin() + bestIndex;
            objectAngles[idx] = *bestIter;
            objectAngleVec[i].erase(bestIter);

            idx++;

            if (idx >= MAX_REPORTING_OBJECTS_[i])
            {
                numObjectsList[i] = idx;
                break;
            }
        }

        switch (i) {

        case 0:
            ohSensorType = OPENCV_SENSOR_OBJECT_001;
            break;

        case 1:
            ohSensorType = OPENCV_SENSOR_OBJECT_002;
            break;

        case 2:
            ohSensorType = OPENCV_SENSOR_OBJECT_003;
            break;

        case 3:
            ohSensorType = OPENCV_SENSOR_OBJECT_004;
            break;

        default:
            printfTextWindow("Warning: OpenCVHandler::processObjectsDetect_(). Unknown sensor type.");
            validSensor = false;
            break;
        }

        if (validSensor)
        {
            // Copy to the object sensor data.
            pthread_mutex_lock(&(sensorDataMutex_[ohSensorType]));

            memcpy(objectAngles_[i], objectAngles, sizeof(float)*numObjectsList[i]);
            memcpy(objectReadings_[i], objectReadings, sizeof(float)*numObjectsList[i]);

            if (sensors_[ohSensorType] != NULL)
            {
                sensors_[ohSensorType]->setAngleInfo(objectAngles_[i], numObjectsList[i]);
                sensors_[ohSensorType]->setReadingInfo(objectReadings_[i], numObjectsList[i]);
            }

            pthread_mutex_unlock(&(sensorDataMutex_[ohSensorType]));
        }

        if (objectAngles != NULL)
        {
            delete [] objectAngles;
            objectAngles = NULL;
        }

        if (objectReadings != NULL)
        {
            delete [] objectReadings;
            objectReadings = NULL;
        }
    }
}

//-----------------------------------------------------------------------
// This function returns the average color of the blob.
//-----------------------------------------------------------------------
CvScalar OpenCVHandler::getBlobColor_(const IplImage *rawImage, const CvBlob *blob)
{
    IplImage *maskImage = NULL;
    CvScalar rgb;

    // Create the blob mask.
    maskImage = cvCreateImage(
        cvSize(rawImage->width, rawImage->height),
        rawImage->depth,
        1);
    cvSetZero(maskImage);

    cvEllipse(
        maskImage,
        cvPointFrom32f(CV_BLOB_CENTER(blob)),
        cvSize(MAX(1, cvRound(CV_BLOB_RX(blob))), MAX(1, cvRound(CV_BLOB_RY(blob)))),
        0,
        0,
        360,
        CV_RGB(255, 255, 255),
        CV_FILLED);


    rgb = cvAvg(rawImage, maskImage);

    // Free the memory.
    cvReleaseImage(&maskImage);

    return rgb;
}

//-----------------------------------------------------------------------
// This function initalizes blob tracking.
//-----------------------------------------------------------------------
bool OpenCVHandler::initializeBlobTracker_(IplImage *rawImage)
{
    CvSize textSize;
    CvPoint pt;
    char buf[1024];
    const int INDEX = IMAGE_PROCESS_TYPE_BLOB_TRACK;

    resetImageProcess_[INDEX] = false;

    if (showImage_)
    {
        // Initialize the font.
        cvInitFont(
            &blobTrackerFont_,
            CV_FONT_HERSHEY_PLAIN,
            DEFAULT_FONT_HSCALE_,
            DEFAULT_FONT_VSCALE_,
            DEFAULT_FONT_ITALICSCALE_,
            DEFAULT_FONT_THICKNESS_,
            DEFAULT_FONT_LINETYPE_);

        pthread_mutex_lock(&(outputImagesMutex_[INDEX]));

        if (shOutputImages_[INDEX] != NULL)
        {
            if (((shOutputImages_[INDEX])->imageData) != NULL)
            {
                cvReleaseData(shOutputImages_[INDEX]);
                cvCreateData(shOutputImages_[INDEX]);
            }
        }
        else
        {
            shOutputImages_[INDEX] = cvCreateImage(
                cvSize(rawImage->width, rawImage->height),
                IPL_DEPTH_8U,
                NUM_RGB_COLOR_TYPES);
        }

        // Show the initialization message at the center.
        sprintf(buf, "Initializing...");
        cvGetTextSize(buf, &blobTrackerFont_, &textSize, NULL);

        pt.x = (((shOutputImages_[INDEX])->width) - (textSize.width))/2;
        pt.y = ((shOutputImages_[INDEX])->height)/2;
        
        cvPutText(
            shOutputImages_[INDEX],
            buf,
            pt,
            &blobTrackerFont_,
            CV_RGB(200,200,255));

        pthread_mutex_unlock(&(outputImagesMutex_[INDEX]));
    }

    if(blobTrackerParam_.pBT != NULL)
    {
        cvReleaseBlobTracker(&(blobTrackerParam_.pBT));
    }

    if(blobTrackerParam_.pBD != NULL)
    {
        cvReleaseBlobDetector(&(blobTrackerParam_.pBD));
    }

    if(blobTrackerParam_.pBTGen != NULL)
    {
        cvReleaseBlobTrackGen(&(blobTrackerParam_.pBTGen));
    }

    if(blobTrackerParam_.pBTA != NULL)
    {
        cvReleaseBlobTrackAnalysis(&(blobTrackerParam_.pBTA));
    }

    if(blobTrackerParam_.pFG != NULL)
    {
        cvReleaseFGDetector(&(blobTrackerParam_.pFG));
    }

    if (blobTracker_ != NULL)
    {
        cvReleaseBlobTrackerAuto(&blobTracker_);
    }

    memset(&blobTrackerParam_, 0x0, sizeof(CvBlobTrackerAutoParam1));

    // Create autotracker module and its components
    blobTrackerParam_.FGTrainFrames = 0;

    // Create FG Detection module
    blobTrackerParam_.pFG = createOpenCVFGDetector_(DEFAULT_FG_DETECTOR_TYPE_);

    if(blobTrackerParam_.pFG == NULL)
    {
        printfTextWindow("Error: OpenCVHandler::processBlobTrack_(). FG detector module could not be created.");
        return false;
    }

    // Set the nick name.
    blobTrackerParam_.pFG->SetNickName(
        (char *)(FG_DETECTOR_NICKNAMES_[DEFAULT_FG_DETECTOR_TYPE_].c_str()));

    // Create Blob Entrance Detection module
    blobTrackerParam_.pBD = createOpenCVBlobDetector_(DEFAULT_BLOB_DETECTOR_TYPE_);

    if(blobTrackerParam_.pBD == NULL)
    {
        printfTextWindow("Error: OpenCVHandler::processBlobTrack_(). Blob detector module could not be created.");
        return false;
    }

    // Set the nick name.
    blobTrackerParam_.pBD->SetNickName(
        (char *)(BLOB_DETECTOR_NICKNAMES_[DEFAULT_BLOB_DETECTOR_TYPE_].c_str()));

    // Create blob tracker module
    blobTrackerParam_.pBT = createOpenCVBlobTracker_(DEFAULT_BLOB_TRACKER_TYPE_);

    if(blobTrackerParam_.pBT == NULL)
    {
        printfTextWindow("Error: OpenCVHandler::processBlobTrack_(). Blob tracker module could not be created.");
        return false;
    }

    // Set the nick name.
    blobTrackerParam_.pBT->SetNickName(
        (char *)(BLOB_TRACKER_NICKNAMES_[DEFAULT_BLOB_TRACKER_TYPE_].c_str()));;

    // create blob trajectory generation module
    blobTrackerParam_.pBTGen = NULL;
 
    // create blob trajectory post processing module
    blobTrackerParam_.pBTPP = NULL;

    //blobTrackerParam_.UsePPData = (bt_corr && MY_STRICMP(bt_corr,"PostProcRes")==0);
    blobTrackerParam_.UsePPData = false;

    // create blob trajectory analysis module
    blobTrackerParam_.pBTA = createBlobTrajectoryAnalyzer_(DEFAULT_BLOB_TRAJECTORY_ANALYZER_TYPE_);
 
    if(blobTrackerParam_.pBTA != NULL)
    {
        // Set the nick name.
        blobTrackerParam_.pBTA->SetNickName(
            (char *)(BLOB_TRAJECTORY_ANALYZER_NICKNAMES_[DEFAULT_BLOB_TRAJECTORY_ANALYZER_TYPE_].c_str()));;
    }

    // Finally create the tracker.
    blobTracker_ = cvCreateBlobTrackerAuto1(&blobTrackerParam_);

    // Check if successfully created.
    if (blobTracker_ == NULL)
    {
        printfTextWindow("Error: OpenCVHandler::processBlobTrack_(). Blob tracker could not be created.");
        return false;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function initializes face detector.
//-----------------------------------------------------------------------
bool OpenCVHandler::initializeFaceDetector_(IplImage *rawImage)
{
    CvSize textSize;
    CvPoint pt;
    char buf[1024];
    const int INDEX = IMAGE_PROCESS_TYPE_FACE_DETECT;

    resetImageProcess_[INDEX] = false;

    if (showImage_)
    {
        // Initialize the font.
        cvInitFont(
            &faceDetectorFont_,
            CV_FONT_HERSHEY_PLAIN,
            DEFAULT_FONT_HSCALE_,
            DEFAULT_FONT_VSCALE_,
            DEFAULT_FONT_ITALICSCALE_,
            DEFAULT_FONT_THICKNESS_,
            DEFAULT_FONT_LINETYPE_);

        pthread_mutex_lock(&(outputImagesMutex_[INDEX]));

        if (shOutputImages_[INDEX] != NULL)
        {
            if (((shOutputImages_[INDEX])->imageData) != NULL)
            {
                cvReleaseData(shOutputImages_[INDEX]);
                cvCreateData(shOutputImages_[INDEX]);
            }
        }
        else
        {
            shOutputImages_[INDEX] = cvCreateImage(
                cvSize(rawImage->width, rawImage->height),
                IPL_DEPTH_8U,
                NUM_RGB_COLOR_TYPES);

            cvCopyImage(rawImage, shOutputImages_[INDEX]);
        }

        // Show the initialization message at the center.
        sprintf(buf, "Initializing...");
        cvGetTextSize(buf, &faceDetectorFont_, &textSize, NULL);

        pt.x = (((shOutputImages_[INDEX])->width) - (textSize.width))/2;
        pt.y = ((shOutputImages_[INDEX])->height)/2;
        
        cvPutText(
            shOutputImages_[INDEX],
            buf,
            pt,
            &faceDetectorFont_,
            CV_RGB(200,200,255));

        pthread_mutex_unlock(&(outputImagesMutex_[INDEX]));
    }

 	// Load the pre-trained Haar classifier data.
	faceDetector_ = (CvHaarClassifierCascade*)cvLoad(
		FACE_DETECT_DATA_XML_FILE_.c_str(),
        0,
        0,
        0);
 
    // Check if successfully created.
    if (faceDetector_ == NULL)
    {
        printfTextWindow(
            "Error: OpenCVHandler::%s():\n Face detector could not be created.",
            __FUNCTION__);

        return false;
    }

	// Create a CvMemStorage object for use by the face detection function.
	faceMemStorage_ = cvCreateMemStorage(0);

    // Check if successfully created.
    if (faceMemStorage_ == NULL)
    {
        printfTextWindow(
            "Error: OpenCVHandler::%s().\n CvMemStorage object could not be created.",
            __FUNCTION__);

        return false;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function initializes object detectors.
//-----------------------------------------------------------------------
bool OpenCVHandler::initializeObjectDetectors_(IplImage *rawImage)
{
    CvSize textSize;
    CvPoint pt;
    char buf[1024];
    int i;
    const int INDEX = IMAGE_PROCESS_TYPE_OBJECTS_DETECT;

    resetImageProcess_[INDEX] = false;

    if (showImage_)
    {
        // Initialize the font.
        cvInitFont(
            &objectDetectorsFont_,
            CV_FONT_HERSHEY_PLAIN,
            DEFAULT_FONT_HSCALE_,
            DEFAULT_FONT_VSCALE_,
            DEFAULT_FONT_ITALICSCALE_,
            DEFAULT_FONT_THICKNESS_,
            DEFAULT_FONT_LINETYPE_);

        pthread_mutex_lock(&(outputImagesMutex_[INDEX]));

        if (shOutputImages_[INDEX] != NULL)
        {
            if (((shOutputImages_[INDEX])->imageData) != NULL)
            {
                cvReleaseData(shOutputImages_[INDEX]);
                cvCreateData(shOutputImages_[INDEX]);
            }
        }
        else
        {
            shOutputImages_[INDEX] = cvCreateImage(
                cvSize(rawImage->width, rawImage->height),
                IPL_DEPTH_8U,
                NUM_RGB_COLOR_TYPES);

            cvCopyImage(rawImage, shOutputImages_[INDEX]);
        }

        // Show the initialization message at the center.
        sprintf(buf, "Initializing...");
        cvGetTextSize(buf, &objectDetectorsFont_, &textSize, NULL);

        pt.x = (((shOutputImages_[INDEX])->width) - (textSize.width))/2;
        pt.y = ((shOutputImages_[INDEX])->height)/2;
        
        cvPutText(
            shOutputImages_[INDEX],
            buf,
            pt,
            &objectDetectorsFont_,
            CV_RGB(200,200,255));

        pthread_mutex_unlock(&(outputImagesMutex_[INDEX]));
    }

    objectDetectors_ = new CvHaarClassifierCascade *[NUM_OBJECTS_DETECT_OBJECT_TYPES];

 	// Load the pre-trained Haar classifier data.
    for (i = 0; i < NUM_OBJECTS_DETECT_OBJECT_TYPES; i++)
    {
        objectDetectors_[i] = (CvHaarClassifierCascade*)cvLoad(
            OBJECTS_DETECT_DATA_XML_FILES_[i].c_str(),
            0,
            0,
            0);
 
        // Check if successfully created.
        if (objectDetectors_[i] == NULL)
        {
            printfTextWindow(
                "Error: OpenCVHandler::%s():\n Objects detector (%d) could not be created.",
                __FUNCTION__,
                i);

            return false;
        }

        // Create a CvMemStorage object for use by the objects detection function.
        objectMemStorages_[i] = cvCreateMemStorage(0);
    }

    // Check if successfully created.
    if (objectMemStorages_ == NULL)
    {
        printfTextWindow(
            "Error: OpenCVHandler::%s().\n CvMemStorage object could not be created.",
            __FUNCTION__);

        return false;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function creates OpenCV's forground obect detection module.
//-----------------------------------------------------------------------
CvFGDetector *OpenCVHandler::createOpenCVFGDetector_(int fgDetectorType)
{
    CvFGDetector *fgDetector = NULL;

    switch (fgDetectorType) {

    case FG_DETECTOR_ACM_MM2003:
        // Foreground Object Detection from Videos Containing Complex Background. ACM MM2003.
        fgDetector = cvCreateFGDetectorBase(CV_BG_MODEL_FGD, NULL);
        break;

    case FG_DETECTOR_ACM_MM2003_SIMPLE:
        // Simplyfied version of FG_DETECTOR_ACM_MM2003.
        fgDetector = cvCreateFGDetectorBase(CV_BG_MODEL_FGD_SIMPLE, NULL); 
        break;

    case FG_DETECTOR_CVPR1999:
        // Adaptive background mixture models for real-time tracking. CVPR1999.
        fgDetector = cvCreateFGDetectorBase(CV_BG_MODEL_MOG, NULL);
        break;

    default:
        printfTextWindow("Error: OpenCVHandler::createFGDetector_(). Unknown FG detector type.");
        break;
    }

    return fgDetector;
}

//-----------------------------------------------------------------------
// This function creates OpenCV's blob detection module.
//-----------------------------------------------------------------------
CvBlobDetector *OpenCVHandler::createOpenCVBlobDetector_(int blobDetectorType)
{
    CvBlobDetector *blobDetector = NULL;

    switch (blobDetectorType) {

    case BLOB_DETECTOR_CC:
        // Detect new blob by tracking CC of FG mask.
        blobDetector = cvCreateBlobDetectorCC();
        break;

    case BLOB_DETECTOR_SIMPLE:
        // Detect new blob by uniform moving of connected components of FG mask.
        blobDetector = cvCreateBlobDetectorSimple();
        break;

    default:
        printfTextWindow("Error: OpenCVHandler::createBlobDetector_(). Unknown blob detector type.");
        break;
    }

    return blobDetector;
}

//-----------------------------------------------------------------------
// This function creates OpenCV's blob tracker module.
//-----------------------------------------------------------------------
CvBlobTracker *OpenCVHandler::createOpenCVBlobTracker_(int blobTrackerType)
{
    CvBlobTracker *blobTracker = NULL;

    switch (blobTrackerType) {

    case BLOB_TRACKER_CCMSPF:
        // Connected component tracking and MSPF resolver for collision.
        blobTracker = cvCreateBlobTrackerCCMSPF();
        break;

    case BLOB_TRACKER_CC:
        // Simple connected component tracking.
        blobTracker = cvCreateBlobTrackerCC();
        break;

    case BLOB_TRACKER_MS:
        // Mean shift algorithm.
        blobTracker = cvCreateBlobTrackerMS();
        break;

    case BLOB_TRACKER_MSFG:
        // Mean shift algorithm with FG mask using.
        blobTracker = cvCreateBlobTrackerMSFG();
        break;

    case BLOB_TRACKER_MSPF:
        // Particle filtering based on MS weight.
        blobTracker = cvCreateBlobTrackerMSPF();
        break;

    default:
        printfTextWindow("Error: OpenCVHandler::createBlobTracker_(). Unknown blob tracker type.");
        break;
    }

    return blobTracker;
}

//-----------------------------------------------------------------------
// This function creates OpenCV's blob trajectory analyzer.
//-----------------------------------------------------------------------
CvBlobTrackAnalysis *OpenCVHandler::createBlobTrajectoryAnalyzer_(int blobTrajectoryAnalyzerType)
{
    CvBlobTrackAnalysis *blobTrajectoryAnalyzer = NULL;

    switch (blobTrajectoryAnalyzerType) {

    case BLOB_TRAJECTORY_ANALYZER_NONE:
        // No analyzer.
        blobTrajectoryAnalyzer = NULL;
        break;

    case BLOB_TRAJECTORY_ANALYZER_HISTPVS:
        // Histogramm of 5D feture vector analysis (x,y,vx,vy,state)
        blobTrajectoryAnalyzer = cvCreateModuleBlobTrackAnalysisHistPVS();
        break;

    case BLOB_TRAJECTORY_ANALYZER_HISTP:
        // Histogramm of 2D feture vector analysis (x,y).
        blobTrajectoryAnalyzer = cvCreateModuleBlobTrackAnalysisHistP();
        break;

    case BLOB_TRAJECTORY_ANALYZER_HISTPV:
        // Histogramm of 4D feture vector analysis (x,y,vx,vy)
        blobTrajectoryAnalyzer = cvCreateModuleBlobTrackAnalysisHistPV();
        break;

    case BLOB_TRAJECTORY_ANALYZER_HISTSS:
        // Histogramm of 4D feture vector analysis (startpos,endpos).
        blobTrajectoryAnalyzer = cvCreateModuleBlobTrackAnalysisHistSS();
        break;

    case BLOB_TRAJECTORY_ANALYZER_TRACKDIST:
        // Compare tracks directly.
        blobTrajectoryAnalyzer = cvCreateModuleBlobTrackAnalysisTrackDist();
        break;

    case BLOB_TRAJECTORY_ANALYZER_IOR:
        // Integrator (by OR operation) of several analysers.
        blobTrajectoryAnalyzer = cvCreateModuleBlobTrackAnalysisIOR();
        break;

    default:
        printfTextWindow("Error: OpenCVHandler::createBlobTrajectoryAnalyzer_(). Unknown blob trajectory analyzer type.");
        break;
    }

    return blobTrajectoryAnalyzer;
}

//-----------------------------------------------------------------------
// This function loops around to display the output image.
//-----------------------------------------------------------------------
void OpenCVHandler::displayLoop_(void)
{
    displayThreadIsUp_ = true;
    
    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Show the current images.
        showOutputImages_();

        // Check for the pthread termination again.
        pthread_testcancel();

        nanosleep(&DISPLAY_SLEEPTIME_NSEC_, NULL);
    }
}

//-----------------------------------------------------------------------
// This function hides the output images by closing the xwindow.
//-----------------------------------------------------------------------
void OpenCVHandler::hideOutputImages_(void)
{
    int i;

    for (i = 0; i < NUM_IMAGE_PROCESS_TYPES; i++)
    {
        pthread_mutex_lock(&(outputImagesMutex_[i]));

        if (xwindows_[i] != NULL)
        {
            delete xwindows_[i];
            xwindows_[i] = NULL;
        }

        pthread_mutex_unlock(&(outputImagesMutex_[i]));
    }
}

//-----------------------------------------------------------------------
// This function displays the output image in the xwindow.
//-----------------------------------------------------------------------
void OpenCVHandler::showOutputImages_(void)
{
    string displayName;
    int width, height;
    int i;
    unsigned int w, h;
    bool createXwindow = false;

    if (!videoIsAvailable_)
    {
        return;
    }

    for (i = 0; i < NUM_IMAGE_PROCESS_TYPES; i++)
    {
        if (!enableImageProcess_[i])
        {
            continue;
        }

        pthread_mutex_lock(&(outputImagesMutex_[i]));

        if ((shOutputImages_[i] != NULL) &&
            (((shOutputImages_[i])->imageData) != NULL) && 
            (((shOutputImages_[i])->imageSize) > 0))
        {
            width = (shOutputImages_[i])->width;
            height = (shOutputImages_[i])->height;

            // Check Xwindow.
            createXwindow = false;

            if (xwindows_[i] == NULL)
            {
                createXwindow = true;
            }
            else
            {
                (xwindows_[i])->getDisplayDimension(&w, &h);

                if (((int)w != width) || ((int)h != height))
                {
                    createXwindow = true;

                    delete xwindows_[i];
                    xwindows_[i] = NULL;
                }
            }

            if (createXwindow)
            {
                // Just create the image this time.
                displayName = NAME_ + ": " + IMAGE_PROCESS_TYPE_NAMES_[i];

                new Xwindow(
                    &(xwindows_[i]),
                    (char *)(displayName.c_str()),
                    0,
                    0,
                    width,
                    height);
            }
            else
            {
                // Display the image.
                (xwindows_[i])->displayRGBImage(
                    0,
                    0,
                    width,
                    height,
                    (unsigned char *)((shOutputImages_[i])->imageData));
            }
        }

        pthread_mutex_unlock(&(outputImagesMutex_[i]));

        pthread_testcancel();
    }
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void OpenCVHandler::updateStatusBar_(void)
{
    char buf[100];
    double frequency = 0;
    int i;
    bool processEnabled = false;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[0] = 0;

    if (!videoIsAvailable_)
    {
        sprintf(buf, "OpenCV: %c No Video", statusbarSpinner_->getStatus());
    }
    else
    {
        if (handlerThreadIsUp_)
        {
            sprintf(buf, "OpenCV: %c Processing", statusbarSpinner_->getStatus());

            for (i = 0; i < NUM_IMAGE_PROCESS_TYPES; i++)
            {
                if (enableImageProcess_[i])
                {
                    if (!processEnabled)
                    {
                        sprintf(buf, "%s %s", buf, IMAGE_PROCESS_TYPE_NAMES_[i].c_str());
                        processEnabled = true;
                    }
                    else
                    {
                        sprintf(buf, "%s, %s", buf, IMAGE_PROCESS_TYPE_NAMES_[i].c_str());
                    }
                }
            }

            if (DISPLAY_FREQUENCY)
            {
                getFrequency_(frequency);
                sprintf(buf, "%s : %.2f Hz", buf, frequency);
            }
        }
        else
        {
            sprintf(buf, "OpenCV: %c Initializing...", statusbarSpinner_->getStatus());
        }
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

//-----------------------------------------------------------------------
// This function sets up the user interface for OpenCV.
//-----------------------------------------------------------------------
void OpenCVHandler::control(void)
{
    bool done = false;
    int msgData[2];
    int c;

    msgData[0] = 0;
    msgData[1] = 0;

    messageDrawWindow(
        EnMessageType_OPENCV_CONTROL,
        EnMessageErrType_NONE,
        msgData);

    do
    {
        c = getch();

        switch(c) {

        case 'd':
            messageHide();
            delete this;
            return;
            break;

        case 'r':
            memset(
                resetImageProcess_,
                (unsigned char)1,
                sizeof(bool)*NUM_IMAGE_PROCESS_TYPES);
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;

        }
    } while (!done);

    messageHide();
}

/**********************************************************************
# $Log: opencv_handler.c,v $
# Revision 1.1  2008-07-15 22:33:51  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.33  2008/04/29 22:23:48  endo
# Regular backup.
#
# Revision 1.32  2008/04/27 02:41:10  endo
# Regular backup.
#
# Revision 1.31  2008/04/23 18:01:12  endo
# Regular backup.
#
# Revision 1.30  2008/04/20 00:41:38  endo
# Regular backup.
#
# Revision 1.29  2008/04/18 21:50:29  endo
# Regular backup.
#
# Revision 1.28  2008/04/18 21:46:36  endo
# Regular backup.
#
# Revision 1.27  2008/04/17 02:39:09  endo
# Regular backup.
#
# Revision 1.26  2008/04/17 01:29:06  endo
# Pioneer2 cleaned up a bit.
#
# Revision 1.25  2008/04/15 00:07:20  endo
# Regular backup.
#
# Revision 1.24  2008/04/14 21:33:29  endo
# Regular backup.
#
# Revision 1.23  2008/04/13 00:33:42  endo
# Regular backup.
#
# Revision 1.22  2008/04/12 20:39:52  endo
# Regular backup.
#
# Revision 1.21  2008/04/12 20:17:24  endo
# OpenCV face detection implemented.
#
# Revision 1.20  2007/02/06 20:17:44  endo
# Regular backup.
#
# Revision 1.19  2007/02/02 22:43:32  endo
# Regular backup.
#
# Revision 1.18  2007/01/19 21:07:54  endo
# Regular backup.
#
# Revision 1.17  2007/01/19 15:50:50  endo
# Regular backup.
#
# Revision 1.16  2007/01/17 17:27:21  endo
# Regular backup.
#
# Revision 1.15  2007/01/12 23:24:37  endo
# Regular backup.
#
# Revision 1.14  2007/01/06 18:57:28  endo
# Regular backup.
#
# Revision 1.13  2007/01/03 22:58:23  endo
# Regular backup.
#
# Revision 1.12  2006/12/17 04:42:53  endo
# SENSOR_OPENCV_BLOB_COLOR added.
#
# Revision 1.11  2006/12/13 17:05:38  endo
# Regular backup.
#
# Revision 1.10  2006/12/12 20:45:15  endo
# Regular update.
#
# Revision 1.9  2006/12/07 20:30:23  endo
# Regular backup.
#
# Revision 1.8  2006/12/01 21:14:54  endo
# Regular backup.
#
# Revision 1.7  2006/11/30 17:51:55  endo
# Regular update.
#
# Revision 1.6  2006/11/29 23:13:28  endo
# OpenCV improved.
#
# Revision 1.5  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.4  2006/11/24 23:17:14  endo
# OpenCV BlobTrack improved.
#
# Revision 1.3  2006/11/22 22:24:28  endo
# OpenCV BlobTrack implemented.
#
# Revision 1.2  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.1  2006/11/20 22:55:52  endo
# OpenCV added.
#
#**********************************************************************/
