
#include "skybotix_interface.h"
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 
#include <unistd.h>

#ifdef NO_FREEIMAGE
#else 
#include <FreeImage.h>
#endif

#include <iostream>

#ifndef NO_OPENCV
#include <highgui.h>
#include <cv.h>
#endif

typedef unsigned char BYTE;

using namespace std;
SkybotixInterface::SkybotixInterface(const char* ip_addr) {

#ifndef NO_OPENCV
  m_pLastImage = NULL;
#endif

  X = Y = Z = Roll = Pitch = Yaw = 0.0;
 
  host = strdup(ip_addr);
  errorBuffer = new char[CURL_ERROR_SIZE];
   pthread_mutex_init(&curl_mutex, NULL);
 
 // Create our curl handle  
  curl_setup(general_curl);
  curl_setup(mutexed_curl);

  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  
  pthread_create(&mySubThread, &attr, thunk2<SkybotixInterface, &SkybotixInterface::PeriodicUpdate>, this);

  // Initialize variables for grabbing camera images
  CurFrame.DataSize = 0;
  CurFrame.Data = NULL;
  NextFrameTime = 0;

}

void SkybotixInterface::sockerror(const char *msg) {
    perror(msg);
    exit(0);
}

static int writer(char *data, size_t size, size_t nmemb, std::string *buffer) {
  // What we will return  
  int result = 0;  
  
  // Is there anything in the buffer?  
  if (buffer != NULL)  
  {  
    // Append the data to the buffer  
    buffer->append(data, size * nmemb);  
  
    // How much did we write?  
    result = size * nmemb;  
  }  
  //  printf("Wrote %d bytes\n", result);
  return result;  
}  



void SkybotixInterface::curl_setup(CURL*& curl) {
  curl = curl_easy_init();  
  
  if (curl) {  
    // Now set up all of the curl options  
    curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, errorBuffer);  
    curl_easy_setopt(curl, CURLOPT_HEADER, 0);  
    curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1);  
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writer);  
  }
  else {
    printf("CURL initialization failed\n");
    exit(0);
  }
}

void SkybotixInterface::PeriodicUpdate() {

  while (true) {
    // get pose
    // libcurl_getenc(host);
    usleep(1000);
  }
}


SkybotixInterface::~SkybotixInterface(void) {
  // Always clean up  
  curl_easy_cleanup(general_curl);  
  curl_easy_cleanup(mutexed_curl);
  pthread_cancel(mySubThread);
  delete errorBuffer;
}

CURLcode SkybotixInterface::libcurl_sendcommand(CURL* curl, char* urlstr, string& buffer) {
  string url(urlstr);  
  
  curl_easy_setopt(curl, CURLOPT_URL, urlstr);  

  CURLcode result = CURLE_FAILED_INIT;  
  
  curl_easy_setopt(curl, CURLOPT_URL, urlstr);  
  curl_easy_setopt(curl, CURLOPT_WRITEDATA, &buffer);  
    
  // Attempt to retrieve the remote page  
  result = curl_easy_perform(curl);  
    
  return result;
}		    




void SkybotixInterface::libcurl_sendcommand_noresp(char * urlstr) {
  string buffer;
  pthread_mutex_lock(&curl_mutex);
  libcurl_sendcommand(general_curl, urlstr, buffer);
  pthread_mutex_unlock(&curl_mutex);
}



void SkybotixInterface::libcurl_grabframe(char* base) {
  string buffer;
  char urlstr[100];
   
  pthread_mutex_lock(&curl_mutex);
  
  // reset the curframe data
  CurFrame.Data = NULL;
  CurFrame.DataSize = 0;

  sprintf(urlstr,"%s/?action=snapshot", base);
  
  //printfTextWindow("Curling!");
  CURLcode result = libcurl_sendcommand(general_curl, urlstr, buffer);
    //printfTextWindow("Curl %s", urlstr);

  if (result == CURLE_OK) {  

    // make a buffer to hold the image data
    BYTE* membuff = new BYTE[buffer.size()];
    memcpy(membuff, buffer.c_str(), buffer.size());
     
    CurFrame.Data = (char *)membuff;
    CurFrame.DataSize = buffer.size();

    //printfTextWindow("Buffer size: %d\n", buffer.size());

  }  
  else  {  
    //printfTextWindow("Curl error");;
    cout << "Error: [" << result << "] - " << errorBuffer<<endl;  
  }
  
  pthread_mutex_unlock(&curl_mutex);
}

FrameData SkybotixInterface::GetCurFrame(void) {
  FrameData fd;

  pthread_mutex_lock(&curl_mutex);
  fd = CurFrame;
  pthread_mutex_unlock(&curl_mutex);

  return fd;
}


// we want to grab a frame every 1 sec otherwise just return the current frame
void SkybotixInterface::GetImage() {

  timeval curtime;

  // grab the frame once a sec	
  gettimeofday(&curtime, NULL);
  if (curtime.tv_sec >= NextFrameTime) {
    NextFrameTime = curtime.tv_sec + 1;
    
    if (CurFrame.Data != NULL) {
      delete CurFrame.Data;
    }
    libcurl_grabframe(host);
  }
  
  return;
}

