/**********************************************************************
 **                                                                  **
 **                            statusbar.h                           **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Status bar updates for HServer                                  **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: statusbar.h,v 1.2 2008-07-15 22:33:53 endo Exp $ */

#ifndef STATUS_BAR_H
#define STATUS_BAR_H

#include <panel.h>
#include <pthread.h>

#define STATUSBAR_MAX_LINES 25

class Statusbar {

protected:
    WINDOW *window_;
    PANEL  *panel_;
    int shNumLines_;
    int index[STATUSBAR_MAX_LINES];
    char *lines[STATUSBAR_MAX_LINES];
    char *empty_line;

    static pthread_mutex_t numLinesMutex_;

    const int ORIGINAL_BAR_SIZE_;

public:
    void resize();
    void draw(void);
    void update(int line_idx);
    int add(char *line);
    void remove(int line_idx);
    Statusbar(int size);
};

class StatusbarSpinner {

protected:
    int counter_;

public:
    StatusbarSpinner(void);
    ~StatusbarSpinner(void);
    char getStatus(void);
};


extern Statusbar *statusbar;

#endif

/**********************************************************************
# $Log: statusbar.h,v $
# Revision 1.2  2008-07-15 22:33:53  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.2  2006/11/29 23:13:28  endo
# OpenCV improved.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2004/09/10 19:43:28  endo
# New PoseCalc integrated.
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.2  2000/09/19 03:49:46  endo
# RCS log added.
#
#
#**********************************************************************/
