/**********************************************************************
 **                                                                  **
 **                         usarsim-laser.h                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for USARSim laser.                                    **
 **                                                                  **
 **  Copyright 2008, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: usarsim-laser.h,v 1.5 2008-11-20 03:51:46 ts2883 Exp $ */

#ifndef USARSIM_LASER_H
#define USARSIM_LASER_H

#include <pthread.h>
#include <string>
#include <vector>

#include "module.h"
#include "sensors.h"
#include "hserver.h"
#include "robot_config.h"
#include "usarsim-robot.h"

using std::string;
using std::vector;

#define USARSIM_LASER_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}
#define USARSIM_LASER_RAD2DEG(val) ((val)*180.0/M_PI)
#define USARSIM_LASER_DEG2RAD(val) ((val)*M_PI/180.0)

class USARSimLaser : public Module {

protected:
    USARSimRobot **usarsimRobot_;
    pthread_t readerThread_;
    extra_laser_t extraInfo_;
    double USARSimLaserFOV_;
    double USARSimLaserRange_;
    string usarsimLaserName_;
    float *shAngles_;
    float shPitchAngle_;
    float *shReadings_;
    float *shRawAngles_;
    float *shRawReadings_;
    float angleOffset_;
    float xOffset_;
    float yOffset_;
    int shMinReadingIndex_;
    int numReadings_;
    bool readerThreadIsUp_;

    RobotPartLaser* m_pclPartLaser;
    pthread_mutex_t readingsMutex_;
    Sensor *sensorLaser_;
    vector<USARSimLaser*> laserList_;
    float *allAngles_;
    float *allReadings_;
    float *allRawReadings_;
    int totalReadings_;

    static int globalLaserNumber;
    int myLaserNumber;

    static const struct timespec READER_SLEEPTIME_NSEC_;
    static const string DEFAULT_USARSIM_LASER_NAME_;
    static const int SKIP_STATUSBAR_UPDATE_;
    static const int WATCHDOG_CHECK_INTERVAL_SEC_;
    static const double DEFAULT_USARSIM_LASER_FOV_;
    static const double DEFAULT_USARSIM_LASER_RANGE_;
  
    void updateStatusBar_(void);
    void readerLoop_(void);
    void readLaserData_(void);
    void addSensor_(void);
    void addToOverallBuffer_(void);
    void removeFromOverallBuffer_(void);
    void logData_(double logTime);
    void accountForXYOffset_(float &reading, float &angle);
    bool initializeUSARSim_(void);

    static void* startReaderThread_(void *usarsimInstance);

public:
    USARSimLaser(
        USARSimLaser **a,
        char *usarsimLaserName,
        float angleOffset, 
        float xOffset,
        float yOffset,
        float laserPitchAngle);
    ~USARSimLaser(void);
    void control(void);
};    

extern USARSimLaser *gUSARSimLaser;

#endif

/**********************************************************************
# $Log: usarsim-laser.h,v $
# Revision 1.5  2008-11-20 03:51:46  ts2883
# Finished adding the pitch information for the laser.
#
# Revision 1.4  2008/11/05 22:14:20  zkira
# In the process of adding pitch information for laser
#
# Revision 1.3  2008/10/30 22:50:51  zkira
# Added support for up to 4 laser sensors
#
# Revision 1.2  2008/10/28 23:23:29  zkira
# Fixed PartLaser (sensor error in hclient)
#
# Revision 1.1  2008/07/15 22:33:54  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.1  2008/02/01 23:18:26  endo
# USARSimLaser added.
#
#**********************************************************************/
