/**********************************************************************
 **                                                                  **
 **                         usarsim-video.h                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for USARSim video (via ImageSrv).                     **
 **                                                                  **
 **  Copyright 2008, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: usarsim-video.h,v 1.1 2008-07-15 22:33:55 endo Exp $ */

#ifndef USARSIM_VIDEO_H
#define USARSIM_VIDEO_H

#include <pthread.h>
#include <string>
#include <vector>

#include "module.h"
#include "hserver.h"
#include "video.h"

using std::string;
using std::vector;

#define USARSIM_VIDEO_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}
#define USARSIM_VIDEO_RAD2DEG(val) ((val)*180.0/M_PI)
#define USARSIM_VIDEO_DEG2RAD(val) ((val)*M_PI/180.0)

class USARSimVideo : public Video {

    typedef enum InitiConnectionStatus_t
    {
        INIT_CONNECTION_STATUS_FAILURE,
        INIT_CONNECTION_STATUS_NOT_READY,
        INIT_CONNECTION_STATUS_SUCCESS
    };

    typedef enum USARSimImageType_t
    {
        USARSIM_IMAGETYPE_RAW = 0,
        USARSIM_IMAGETYPE_JPEG_QUALITYSUPERB = 1,
        USARSIM_IMAGETYPE_JPEG_QUALITYGOOD = 2,
        USARSIM_IMAGETYPE_JPEG_QUALITYNORMAL = 3,
        USARSIM_IMAGETYPE_JPEG_QUALITYAVERAGE = 4,
        USARSIM_IMAGETYPE_JPEG_QUALITYBAD = 5
    };

    typedef struct ImageData_t
    {
        int imageSize;
        unsigned char *image;
    };

protected:
    string serverHostMachine_;
    string usarsimVideoName_;
    vector<ImageData_t> imageBuffer_;
    double lastTimeStamp_;
    int shSocketFd_;
    int serverPort_;
    int imageBufferIndex_;

    static pthread_mutex_t socketMutex_;

    static const string DEFAULT_USARSIM_SERVER_HOST_MACHINE_;
    static const string DEFAULT_USARSIM_VIDEO_NAME_;
    static const string USARSIM_ACK_STRING_;
    static const int DEFAULT_USARSIM_SERVER_PORT_;
    static const int IMAGE_BUFFER_SIZE_;
    static const int INVALID_SOCKET_FD_;
    static const unsigned int MAX_READ_BLOCK_SIZE_;

    int connectToServer_(string hostMacine, int port);
    void updateStatusBar_(void);
    bool initializeUSARSim_(void);
    bool readNBytes_(const int fd, unsigned char *buf, const int n);
    bool readVideoData_(void);

public:
    USARSimVideo(
        Video **a,
        char *serverHostMahine,
        char *usarsimVideoName,
        int serverPort,
        bool showImage);
    ~USARSimVideo(void);
    void control(void);
    void setChannel(int channel);
};    

#endif

/**********************************************************************
# $Log: usarsim-video.h,v $
# Revision 1.1  2008-07-15 22:33:55  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.2  2008/01/21 22:23:14  endo
# USARSimVideo implemented.
#
# Revision 1.1  2008/01/21 19:14:16  endo
# Regular backup.
#
#**********************************************************************/
