/**********************************************************************
 **                                                                  **
 **                              video.h                             **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Video control for HServer.                                      **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: video.h,v 1.2 2008-07-15 22:33:55 endo Exp $ */

#ifndef VIDEO_H
#define VIDEO_H

#include <stdio.h>
#include <pthread.h>
#include <string>

#include "module.h"
#include "robot_config.h"
#include "xwindow.h"

using std::string;

class Video : public Module {

protected:
    Xwindow *xwindow_;
    pthread_t readerThread_;
    pthread_t displayThread_;
    double frameRate_;
    double lastCapturedTime_;
    double shCurrentImageTimeStamp_;
    int videoType_;
    int shCurrentImageWidth_;
    int shCurrentImageHeight_;
    int shCurrentImageSize_;
    unsigned char *shCurrentImage_;
    bool showImage_;
    bool readerThreadIsUp_;
    bool displayThreadIsUp_;

    static pthread_mutex_t currentImageMutex_;

    static const int READER_THREAD_USLEEP_;
    static const int DISPLAY_THREAD_USLEEP_;
    static const int SKIP_STATUSBAR_UPDATE_;
    static const int WATCHDOG_CHECK_INTERVAL_SEC_;

    void showCurrentImage_(void);
    void readerLoop_(void);
    void displayLoop_(void);

    static void *startReaderThread_(void *videoInstance);
    static void *startDisplayThread_(void *videoInstance);
    static bool bgr2rgb_(
        int w,
        int h,
        const unsigned char *bgrImage,
        unsigned char **rgbImage);
    static bool flipY_(
        int w,
        int h,
        unsigned char **image);

    virtual void updateStatusBar_(void) = 0;
    virtual bool readVideoData_(void) = 0;

public:
    Video(Video** a, int videoType, string name, bool showImage);
    unsigned char *duplicateCurrentImage(
        int *width,
        int *height,
        int *imageSize,
        double *timeStamp);
    void dumpCurrentImage(FILE *ppmFile);

    virtual ~Video(void);
    virtual void control() = 0;
    virtual void setChannel(int channel) = 0;
};

extern Video *gVideo;

#endif

/**********************************************************************
# $Log: video.h,v $
# Revision 1.2  2008-07-15 22:33:55  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.6  2008/01/21 22:23:14  endo
# USARSimVideo implemented.
#
# Revision 1.5  2008/01/21 19:14:17  endo
# Regular backup.
#
# Revision 1.4  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.3  2006/11/22 22:24:28  endo
# OpenCV BlobTrack implemented.
#
# Revision 1.2  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2006/05/15 06:01:03  endo
# gcc-3.4 upgrade
#
# Revision 1.1.1.1  2005/02/06 22:59:43  endo
# AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
#
# Revision 1.4  2001/03/23 21:32:51  blee
# altered to use a config file
#
# Revision 1.3  2000/09/19 03:51:29  endo
# RCS log added.
#
#
#**********************************************************************/
