/**********************************************************************
 **                                                                  **
 **                              xwindow.h                           **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  X11 Interface for HServer                                       **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: xwindow.h,v 1.2 2008-07-15 22:33:55 endo Exp $ */

#ifndef XWINDOW_H
#define XWINDOW_H

#include <string>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include "module.h"

using std::string;

/*
void xwindowShowWindow();
void xwindowConnect();
void xwindowPutImage(XImage *ximg,int x,int y,int w,int h);
void xwindowPutRGB(int x,int y,int w,int h, unsigned char *data);
void xwindowDrawRect(int x,int y,int w,int h,int c);
void xwindowDrawPoint(int x,int y,int c);
*/

#define XWINDOW_MAX_COLORS 10
#define XWINDOW_NUM_COLORS 5
#define COL_BLACK    0
#define COL_WHITE    1
#define COL_RED      2
#define COL_GREEN    3
#define COL_BLUE     4

class Xwindow : public Module {

    typedef struct ImageData_t {
        double timeStamp;
        int x;
        int y;
        int width;
        int height;
        int imageSize;
        unsigned char *rgbImage;
    };

    typedef enum ImageColorOrder_t {
        IMAGE_COLOR_ORDER_RGB,
        IMAGE_COLOR_ORDER_BGR
    };

protected:
    Display *display_;
    Window window_;
    XSizeHints *sizeHints_;
    pthread_t displayThread_;
    GC gc_;
    ImageData_t shCurrentImageData_;
    string displayName_;
    double lastImageDataTimeStamp_;
    int screenNum_;
    int screenDepth_;
    int numColors_;
    int displayX_;
    int displayY_;
    unsigned int displayWidth_;
    unsigned int displayHeight_;
    unsigned int colors_[XWINDOW_MAX_COLORS];
    bool displayThreadIsUp_;

    static pthread_mutex_t currentImageDataMutex_;

    static const string DEFAULT_WINDOW_NAME_;
    static const string DEFAULT_COLOR_FILE_NAME_;
    static const int READER_THREAD_USLEEP_;
    static const int SKIP_STATUSBAR_UPDATE_;
    static const int WATCHDOG_CHECK_INTERVAL_SEC_;
    static const int DEFAULT_DISPLAY_X_;
    static const int DEFAULT_DISPLAY_Y_;
    static const unsigned int DEFAULT_DISPLAY_WIDTH_;
    static const unsigned int DEFAULT_DISPLAY_HEIGHT_;

    unsigned int rgb2Color_(int r, int g, int b);
    void updateDisplay_(void);
    void checkXEvent_(void);
    void handleKeyPress_(XKeyEvent *report);
    void createWindow_(void);
    void drawRect(int x, int y, int w, int h, int c);
    void drawPoint(int x,int y,int c);
    void updateStatusBar_(void);
    void displayLoop_(void);
    void mapWindow_();
    void saveImage_(
        int x,
        int y,
        int w,
        int h,
        const unsigned char* image,
        int imageColorOrder);
    bool rgb2xImage_(
        int w,
        int h,
        const unsigned char* rgbImage,
        XImage *xImage);
    bool bgr2rgb_(
        int w,
        int h,
        const unsigned char *bgrImage,
        unsigned char **rgbImage);
    bool bgr2rgb_(int w, int h, unsigned char **data);
    bool initializeWindow_(void);
    bool loadColors_(void);

    static void *startDisplayThread_(void *xwindowInstance);

public:
    void displayRGBImage(
        int x,
        int y,
        int w,
        int h,
        unsigned char *rgbImage);
    void displayBGRImage(
        int x,
        int y,
        int w,
        int h,
        unsigned char *bgrImage);
    void control();
    void getDisplayDimension(unsigned int *width, unsigned int *height);
    Xwindow(
        Xwindow **a,
        char *displayName,
        int displayX,
        int displayY,
        int displayWidth,
        int displayHeight);
    ~Xwindow();
};

inline void Xwindow::getDisplayDimension(unsigned int *width, unsigned int *height)
{
    *width = displayWidth_;
    *height = displayHeight_;
}

extern Xwindow *gXwindow;

#endif

/**********************************************************************
# $Log: xwindow.h,v $
# Revision 1.2  2008-07-15 22:33:55  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.5  2006/11/24 23:17:14  endo
# OpenCV BlobTrack improved.
#
# Revision 1.4  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.3  2006/11/20 16:43:38  endo
# Xwindow improved.
#
# Revision 1.2  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.2  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.3  2004/09/10 19:43:28  endo
# New PoseCalc integrated.
#
# Revision 1.2  2004/04/24 11:20:53  endo
# 3DM-G added.
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.3  2000/09/19 03:51:29  endo
# RCS log added.
#
#
#**********************************************************************/
